/** @file indexset.h @brief Classes IndexSet, TaIndexSet */


/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Copyright (C) 2007  Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */



#ifndef FAUDES_INDEXSET_H
#define FAUDES_INDEXSET_H

#include "definitions.h"
#include "attributes.h"
#include "baseset.h"
#include "symboltable.h"
#include <set>
#include <map>
#include <algorithm>


namespace faudes {



/** @addtogroup ContainerClasses */
/** @{*/

// Forward declaration for the attributed version of indexset
template<class Attr> class TaIndexSet;

/**
 * Set of indices. This class is built on top of the faudes version
 * TBaseSet of STL sets. It provides the essentials of the STL interface
 * and inherits the standard token IO interface from Type, so you may 
 * use Read and Write for file/string/console IO.
 *
 * IndexSet is used as common base for TaIndexSet (indices with attributes),
 * NameSet (indices with symbolic names) and TaNameSet (indices with
 * attributes and symbolic names).
 *
 * The index of value 0 is reserved to indicate the invalid index. If the macro FAUDES_CHECKED
 * is defined, the attempt to insert an index 0 triggers an exception (id 61). Invalid
 * iterators throw an exception (id 62) when used as an argument to a IndexSet function. 
 *
 * Technically, file IO is done by TaIndexSet functions. This requires a copy and for
 * that reason may be reimplemented in a future versions.
 * The format is demonstrated by the following example
 * of a set with name "MySet" consisting of indices 17, 25, 26, 27, 28, 40:
 * \code
 * <Myset> 
 * 17 
 * <Consecutive> 25 4 </Consecutive> 
 * 40 
 * </MySet>  
 * \endcode
 * Theere is a minimum number of consecutive indices for the use of the consecutive
 * section defined in definitions.h.
 *
 */

class IndexSet : public TBaseSet<Idx> {

  /** 
   * We implement "protected privacy for template classes" by friendship.
   * This is used for the pragmatic implemention conversion constructors.
   */
  template<class Attr> friend class TaIndexSet;

public:

  /**
   * Constructor. 
   */
  IndexSet(void);

  /**
   * Copy-constructor. 
   */
  IndexSet(const TBaseSet<Idx>& rOtherSet);

  /**
   * Construct from file. Uses the Read(TokenReader&, const std::string&) function to read.
   * a IndexSet from section rLabel in file rFilename.
   *
   * @param rFilename
   *   Name of file to read
   * @param rLabel
   *   Section to read
   *
   * @exception Exception
   *   - IO Errors (id 1)
   *   - token musmatch (id 50, 51, 52)
   */
  IndexSet(const std::string& rFilename, const std::string& rLabel = "IndexSet");

  /**
   * Construct on heap with same type.
   *
   * @return 
   *   Pointer to new TBaseSet
   */
  virtual IndexSet* New(void) const; 

  /**
   * Virtual destructor
   */
  virtual ~IndexSet(void) {};

  /**
   * Copy indexset. Performs a fake copy, see TBaseSet.
   * This function maintains attributes if types can be casted.
   *
   * @param rOtherSet 
   *    Destination to copy to
   */
  virtual void Copy(IndexSet& rOtherSet) const;

  /** Assignment operator (fake copy)
   *
   * @param rSrc 
   *    Source from which to assign
   * @return
   *    Ref to this set
   */
  virtual IndexSet& operator=(const IndexSet& rSrc) { rSrc.Copy(*this); return *this;};

  /**
   * Copy indexset. Performs a fake copy, see TBaseSet.
   * This function sets attributes to default.
   *
   * @param rOtherSet 
   *    Destination to copy to
   */
  virtual void CopyWithoutAttributes(IndexSet& rOtherSet) const;

  /**
   * Get maximum index used in this set (0 for emptyset)
   *
   * @return 
   *   max  indices 
   */
  Idx MaxIndex(void) const;

  /**
   * Test whether index is not 0
   * @param index
   *    Index to test
   * @return
   *    True if idx!=0
   */
   bool Valid(Idx index) const;
   
  /** 
   * Iterators on indexset. 
   */
  typedef TBaseSet<Idx>::Iterator Iterator;

  /**
   * Insert new index to set
   *
   * @return
   *   New index 
   */
  Idx Insert(void);

  /** 
   * Insert specified index
   *
   * @param idx
   *    Index to insert
   * @return
   *   True if index was new to set 
   */
  bool Insert(Idx idx);

  /**
   * Compute an Idx type signature for a Set. The value is computed by 
   * summing up the product of every index with an integer starting at
   * 1 (for the first index) and ending at Size() (for the last index).
   *
   * @return
   *   Idx type set signature
   */
  Idx Signature(void) const;

  /**
   * Return pretty printable index. 
   *
   * @param index
   *   Index to print
   *
   * @return
   *   String
   */
  std::string Str(Idx index) const {return ToStringInteger(index); };

 protected:

  /** 
   * Write to TokenWriter, see Type::Write for public wrappers.
   * Appends the IndexSet to 
   * TokenWriter. This will write a section with label as specified that holds
   * integer tokens to represent the IndexSet. For non-default attribute values,
   * the respective index is followed by the attribute value. The latter may 
   * consist of sections (or tokens different from integer or string) to allow
   * for transparent reading; see AttributeVoid and AttributeFlags. Example for
   * a set with name "MySet" consisting of indices 17, 25, 40, where index 25 has a
   * nontrivial attribute attached:
   * \code
   * <Myset> 
   * 17 
   * 25 <AValue> "Some Value" </AValue> 
   * 40 
   * </MySet>  
   * \endcode
   *
   * @param tw
   *   Reference to TokenWriter
   * @param rLabel
   *   Label of section to write, defaults to name of set or "IndexSet"
   * @param pContext
   *   Write context to provide contextual information
   *
   * @exception Exception
   *   - IO errors (id 2)
   */
  virtual void DoWrite(TokenWriter& tw, const std::string& rLabel="", const Type* pContext=0) const;

  /**
   * Read from TokenReader, see Type::Read for public wrappers. 
   * This method invokes TokenReader::SeekBegin() to seek the specified
   * section, reads subsequent integer tokens as indexes, calls matching TokenReader::SeekEnd(). 
   * When used by a derived class with attributes, attributes are read, too. If no section is
   * specified, the section "IndexSet" is read.
   *
   * @param rTr
   *   Reference to tokenreader
   * @param rLabel
   *   Section to read, defaults to "IndexSet"
   * @param pContext
   *   Read context to provide contextual information
   *
   * @exception Exception
   *   - IO errors (id 1)
   *   - token mismatch (id 50, 51, 52)
   */
  virtual void DoRead(TokenReader& rTr, const std::string& rLabel="", const Type* pContext=0);


};

/* convenience typedef for stateset */
typedef IndexSet StateSet;



/**
 * Set of indices with attributes. A TaIndexSet<Attr> is a template derived from IndexSet,
 * such that  each set element has an Attribute of class Attr.
 * The template parameter Attr must provide basic functions for file IO and must define a default
 * value, see AttributeVoid and AttributeFlags for an examples.
 * Attributes are maintained by an STL map to associate indices
 * with attributes different from the default value. Thus, attributes with default value 
 * do not take up any memeory.
 *
 * @param mAttributeMap
 *   STL map of type std::map<Idx,Attr> to hold attributes
 *
 * The copy constructor copies attributes; the methods InsertSet(), EraseSet() and
 * RestrictSet() maintain attributes; all other set operations derived from IndexSet  
 * either return a IndexSet (no attributes) or set all attributes to the dafault value.
 * To set or get an attribute of an index that does not exist in the set is considered
 * as an error and triggers an exception (id 60) when the macro FAUDES_CHECKED is defined. 
 *
 * The format for token IO is demonstrated by the following example
 * of a set with name "MySet" consisting of indices 17, 25, 40, where index 25 has a
 * nontrivial attribute attached:
 * \code
 * <Myset> 
 * 17 
 * 25 <AValue> "Some Value" </AValue> 
 * 40 
 * </MySet>  
 * \endcode
 * Note that attributes may be either subsections or tokens of type different from integer and string.
 * This is to allow reading a token stream to a class with incompatible (or no) attributes. 
 */



template<class Attr>
class TaIndexSet : public virtual IndexSet {

  /** 
   * We implement "protected privacy for template classes" by friendship.
   * This is used for the pragmatic implemention conversion constructors.
   */
  friend class IndexSet;

 public:
  /**
   * Constructor. 
   */
  TaIndexSet(void);

  /**
   * Copy-constructor (from TaIndexSet, incl attributes) 
   */
  TaIndexSet(const TaIndexSet& rOtherSet);

  /**
   * Copy-constructor (from IndexSet, sets attributes to default) 
   */
  TaIndexSet(const IndexSet& rOtherSet);

  /**
   * Construct from file. 
   * This constructor uses the Read(TokenReader&, const std::string&) function to read.
   *
   * @param rFilename
   *   Name of file
   * @param rLabel
   *   Section for the set in file
   *
   * @exception Exception
   *   - IO errors (id 1)
   *   - token mismatch (id 50, 51, 52)
   */
  TaIndexSet(const std::string& rFilename, const std::string& rLabel = "IndexSet");

  /**
   * Construct on heap with same type.
   *
   * @return 
   *   Pointer to new TaIndexSet
   */
  virtual TaIndexSet* New(void) const; 

  /**
   * Virtual destructor
   */
  virtual ~TaIndexSet(void) {};

  /** Assignment operator (uses Copy)
   *
   * @param rSrc 
   *    Source from which to assign
   * @return
   *    Ref to this set
   */
  virtual TaIndexSet& operator=(const TaIndexSet& rSrc) { rSrc.Copy(*this); return *this;};

  /** Assignment operator (uses Copy)
   *
   * @param rSrc 
   *    Source from which to assign
   * @return
   *    Ref to this set
   */
  virtual TaIndexSet& operator=(const IndexSet& rSrc) { rSrc.Copy(*this); return *this;};

  /**
   * Copy to other index with same attribute type.
   *
   * @param
   *   rSet
   */
   virtual void Copy(TaIndexSet<Attr>& rSet) const;

  /**
   * Copy to other index set. 
   * This function maintains attributes
   * if types can be casted.
   *
   * @param
   *   rSet
   */
   virtual void Copy(IndexSet& rSet) const;

  /**
   * Get number of explicit (aka non-default) attributes.
   *
   * @return 
   *   Number of entries in mAttributeMap
   */
  virtual Idx AttributesSize(void) const;

  /**
   * Attribute typeinfo.
   *
   * @return
   *    Pointer to some attribute of this sets attribute type.
   */
  virtual const Attr* Attributep(void) const { static Attr attr; return &attr; };

  /**
   * Attribute typeinfo
   *
   * @return
   *    Reference to some attribute of this sets attribute type
   */
  virtual const AttributeVoid& Attribute(void) const { static Attr attr; return attr;};

  /** 
   * Attribute typeinfo.
   * This set accepts all attributes that we can cast to our attribute type.
   * @param rAttr
   *     Attribute type to test.
   * @return True, if attribute type is accepted.
   */
  virtual bool AttributeTry(const Type& rAttr) const {return dynamic_cast< const Attr* >(&rAttr)!=0;};

  /** 
   * Get attribute reference by index. Note that in order to produce a non-const 
   * reference this method will insert an explicit default attribute if necessary. 
   * If a const reference is sufficient, you should use Attribute(Idx) const instead.
   *
   * @param idx
   *    Index to specify element
   * @exception Exception
   *   - Element does not exist (63)
   * @return 
   *    Pointer to attribute
   */
  virtual Attr* Attributep(const Idx& idx);

  /** 
   * Get attribute by index. This function returns a const reference to the attribute
   * of the specified index. 
   * @param idx
   *    Index to specify element
   * @exception Exception
   *   - Element does not exist (63)
   * @return 
   *    Reference to attribute
   */
  virtual const Attr& Attribute(const Idx& idx) const;

  /** 
   * Set attribute. Sets the attribute of the sepcified element to the given value. 
   *
   * @param idx
   *    Index to specify element
   * @param attr
   *    Attribute value.
   * @exception Exception
   *   - Element does not exist (63)
   */
  virtual void Attribute(const Idx& idx, const Attr& attr);

  /** 
   * Set attribute. Provided that the attribute can be casted to the
   * appropriate type, this method sets the attribute of the sepcified element to the given value. 
   *
   * @param idx
   *    Index to specify element
   * @param attr
   *    Attribute value.
   * @exception Exception
   *   - Element does not exist (63)
   *   - Cannot cast attribute type (63)
   */
  virtual void Attribute(const Idx& idx, const Type& attr);

  /** 
   * Set attribute. Provided that the attribute can be casted to the
   * appropriate type, this method sets the attribute of the sepcified element to the given value. 
   *
   * @param idx
   *    Index to specify element
   * @param attr
   *    Attribute value.
   * @exception Exception
   *   - Element does not exist (63)
   */
  virtual void AttributeTry(const Idx& idx, const Type& attr);

  /**
   * Set attributes. Provided that rOtherSet has attributes that can be
   * casted to the appropriate type, attributes are copied per element from rOtherSet. 
   * Elements of this set which are not in rOtherSet maintain their attribute. 
   *
   * @param rOtherSet
   *   Other IndexSet
   * @exception Exception
   *   - Element does not exist (63)
   *   - Cannot cast attribute type (63)
   */
   virtual void Attributes(const IndexSet& rOtherSet);

  /**
   * Set attributes. Attributes are copied per element from rOtherSet. 
   * Elements of this set which are not in rOtherSet maintain their attribute. 
   *
   * @param rOtherSet
   *   Other IndexSet
   */
   virtual void Attributes(const TaIndexSet& rOtherSet);

  /** 
   * Clear attribute to default value. 
   *
   * @param idx
   *    index to specify element
   */
  virtual void ClrAttribute(const Idx& idx);

  /** 
   * Clear all attributes to default value.
   *
   */
  virtual void ClearAttributes(void);


  /** 
   * Erase Element (incl its attribute) 
   *
   * @param idx
   *    Index to specify element
   * @return
   *    True if element used to exist
   */
  bool Erase(Idx idx);


  /** 
   * Erase element by iterator (incl attribute) 
   *
   * @param pos
   *    Iterator to specify element
   * @return 
   *    Iterator to next element or End()
   */
  Iterator Erase(const Iterator& pos); 


  /** 
   * Erase elements given by other set. This function
   * ignores the attributes of the other set and maintains the attributes 
   * of the remaining elements in this set.
   *
   * @param rOtherSet 
   *    Elements to erase
   */
   void EraseSet(const IndexSet& rOtherSet);

  /** 
   * Restrict to specified subset. Erases any elements no in
   * the specified set. This function
   * ignores the attributes of the other set and maintains the attributes 
   * of the remaining elements in this set.
   *
   * @param rOtherSet 
   *    Elements to erase
   */
   void RestrictSet(const IndexSet& rOtherSet);


  /**
   * Insert new index to set using default attribute.
   *
   * @return
   *   New index 
   */
  Idx Insert(void);

  /** 
   * Insert new index with attribute. 
   * If the index allready exists, the
   * attribute is overwritten by rAttr.
   *
   * @param rAttr
   *   Specify attribute of new element
   * @return
   *   new index 
   */
  Idx Insert(const Attr& rAttr);

  /** 
   * Insert element. If the element exists, the attribute is maintained.
   * If the element does not exist, it is inserted with default attribute.
   *
   * @param idx
   *   Index to specify element
   * @return
   *   True if element was new to set 
   */
  bool Insert(Idx idx);

  /** 
   * Insert element with attribute. 
   *
   * @param idx
   *   Index to specify element
   * @param attr
   *   Specify attribute of (new) element
   * @return
   *   True if element was new to set 
   */
  bool Insert(Idx idx, const Attr& attr);

  /**
   * Inserts elements of rOtherSet.
   * Attributes of this set are maintained, newly inserted elements have default attribute.
   *
   *
   * @param rOtherSet
   *   Other IndexSet
   */
   void InsertSet(const IndexSet& rOtherSet);

  /**
   * Inserts elements of rOtherSet.
   * Attributes of this set are maintained, new elements are inserted with attribute.
   *
   * @param rOtherSet
   *   Other IndexSet
   */
   void InsertSet(const TaIndexSet& rOtherSet);

  /** 
   * Clear all set.
   */
  virtual void Clear(void);
      
  /**
   * Set union. The union is wrt the set of indices, the result is accumulated in this set.
   * Attributes are set to default. See also InsertSet(const TaIndexSet&).
   *
   * @param rOtherSet
   *   Other TaIndexSet
   */
  void SetUnion(const TaIndexSet& rOtherSet);

  /**
   * Set union. The union is wrt the set of indices, the result is accumulated in this set.
   * Attributes are set to default. See also InsertSet(const IndexSet&).
   *
   * @param rOtherSet
   *   Other IndexSet
   */
  void SetUnion(const IndexSet& rOtherSet);

  /**
   * Set intersection. The intersection is  wrt set of indices, the result is stored in this set.
   * Attributes are set to default.
   * 
   * @param rOtherSet
   *   Other IndexSet
   */
  void SetIntersection(const TaIndexSet& rOtherSet);

  /**
   * Set intersection. The intersection is  wrt set of indices, the result is stored in this set.
   * Attributes are set to default.
   * 
   * @param rOtherSet
   *   Other IndexSet
   */
  void SetIntersection(const IndexSet& rOtherSet);


 protected:
  /** Stored attributes */
  std::map<Idx,Attr> mAttributeMap;

  /** default attribute */ 
  static Attr mDefAttribute; 


};

/** Statics per template parameter */
template<class Attr> Attr TaIndexSet<Attr>::mDefAttribute;

/** Convenience Macro */
//template<class Attr> typedef TaStateSet<class Attr> TaIndexSet<class Attr>
#define TaStateSet TaIndexSet


/** @} doxygen group */


/*
*************************************************************************************
*************************************************************************************
 Implementation
*************************************************************************************
*************************************************************************************
*/

// more includes
//#include "vgenerator.h"

// TaIndexSet()
template<class Attr>
TaIndexSet<Attr>::TaIndexSet(void) : 
  IndexSet()
{
  FD_DC("TaIndexSet(" << this << ")::TaIndexSet()");
  //IndexSet::Name("IndexSet");
}

// TaIndexSet(rOtherSet)
template<class Attr>
TaIndexSet<Attr>::TaIndexSet(const TaIndexSet& rOtherSet) : 
  IndexSet(rOtherSet) 
{
  FD_DC("TaIndexSet(" << this << ")::TaIndexSet(rOtherSet " << &rOtherSet << ")");
  // copy my members
  mAttributeMap=rOtherSet.mAttributeMap;
}

// TaIndexSet(rOtherSet)
template<class Attr>
TaIndexSet<Attr>::TaIndexSet(const IndexSet& rOtherSet) :
  IndexSet() 
{
  FD_DC("TaIndexSet(" << this << ")::TaIndexSet(rOtherSet " << &rOtherSet << ")");
  // call base copy
  rOtherSet.Copy(*this);
}

// pointer constructor
template<class Attr>
TaIndexSet<Attr>* TaIndexSet<Attr>::New(void) const { 
  return new TaIndexSet<Attr>(); 
}

// File constructor
template<class Attr>
TaIndexSet<Attr>::TaIndexSet(const std::string& rFilename, const std::string& rLabel) {
  FD_DC("TaIndexSet(" << this << ")::TaIndexSet(" << rFilename << ")");
  Read(rFilename, rLabel);
}


// Copy(rSet) attributes known and match
template<class Attr>
void TaIndexSet<Attr>::Copy(TaIndexSet<Attr>& rSet) const {
  FD_DC("TaIndexSet(" << this << ")::Copy(" << &rSet<<")");
  // call baseset method (avoid direct base)
  TBaseSet<Idx>::Copy(rSet);
  // copy my members (aka attributes)
  rSet.mAttributeMap=mAttributeMap; 
}  

// Copy(rSet)
template<class Attr>
void TaIndexSet<Attr>::Copy(IndexSet& rSet) const {
  FD_DC("TaIndexSet(" << this << ")::Copy([v] " << &rSet<<")");
  FD_DC("TaIndexSet(" << this << ")::Copy(): dst type " << typeid(rSet).name());
  FD_DC("TaIndexSet(" << this << ")::Copy(): src type " << typeid(*this).name());
  // call attribute smart base
  IndexSet::Copy(rSet);
}

// AttributesSize()
template<class Attr>
Idx TaIndexSet<Attr>::AttributesSize(void) const {
  return (Idx)mAttributeMap.size();
}

// Attributep(idx)
template<class Attr>
Attr* TaIndexSet<Attr>::Attributep(const Idx& idx) {
#ifdef FAUDES_CHECKED
  if(!Exists(idx)) {
    std::stringstream errstr;
    errstr << "index " << idx << " not member of set" << std::endl;
    throw Exception("TaIndexSet::Attributep(idx)", errstr.str(), 60);
  }
#endif
  typename std::map<Idx,Attr>::iterator ait;
  ait=mAttributeMap.find(idx);
  if( ait!=mAttributeMap.end() ) 
    return & (ait->second);
  // insert explicit default attribute
  typename std::map<Idx,Attr>::value_type entry(idx,mDefAttribute);
  ait=mAttributeMap.insert(entry).first;
  return & (ait->second);
}

// Attribute(idx)
template<class Attr>
const Attr& TaIndexSet<Attr>::Attribute(const Idx& idx) const{
#ifdef FAUDES_CHECKED
  if (!Exists(idx)) {
    std::stringstream errstr;
    errstr << "index " << idx << " not member of set" << std::endl;
    throw Exception("TaIndexSet::Attribute(idx)", errstr.str(), 60);
  }
#endif
  typename std::map<Idx,Attr>::const_iterator ait;
  ait=mAttributeMap.find(idx);
  if( ait!=mAttributeMap.end() ) {
    return ait->second;
  } 
  return mDefAttribute;
}

// Attribute(idx,attr)
template<class Attr>
void TaIndexSet<Attr>::Attribute(const Idx& idx, const Attr& attr) {
#ifdef FAUDES_CHECKED
  if (!Exists(idx)) {
    std::stringstream errstr;
    errstr << "index " << idx << " not member of set" << std::endl;
    throw Exception("TaIndexSet::Attribute(idx,attr)", errstr.str(), 60);
  }
#endif
  if(attr.IsDefault()) {
    mAttributeMap.erase(idx);
  } else {
    mAttributeMap[idx]=attr;
  }
}

// Attribute(idx,attr)
template<class Attr>
void TaIndexSet<Attr>::Attribute(const Idx& idx, const Type& rAttr) {
  // cast
  const Attr* attrp = dynamic_cast<const Attr*>(&rAttr);
  if(!attrp) {
    std::stringstream errstr;
    errstr << "cannot cast attribute type " << typeid(rAttr).name();
    errstr << " for set " << Name() << " with type " << typeid(*this).name();
    throw Exception("TaIndexSet::Attribute", errstr.str(), 63);
  }
  // set
  Attribute(idx, *attrp);
}

// AttributeTry(idx,attr)
template<class Attr>
void TaIndexSet<Attr>::AttributeTry(const Idx& idx, const Type& rAttr) {
  FD_DC("TaIndexSet(" << this << ")::AttributeTry(" << idx << ", attr): try to set " << 
     rAttr.ToString() << " with type " << typeid(rAttr).name());
  // cast
  const Attr* attrp = dynamic_cast<const Attr*>(&rAttr);
  if(!attrp) return;
  // set
  Attribute(idx, *attrp);
}

// Attributes(set)
template<class Attr>
void TaIndexSet<Attr>::Attributes(const TaIndexSet& rOtherSet) {
  FD_DC("TaIndexSet(" << this << ")::Attributes(set) with type " << typeid(rOtherSet.Attribute()).name());
  Detach();
  iterator it1 = pSet->begin();
  iterator it2 = rOtherSet.pSet->begin();
  while ((it1 != pSet->end()) && (it2 != rOtherSet.pSet->end())) {
    if (*it1 < *it2) {
      ++it1;
    }
    else if (*it1 == *it2) {
      Attribute(*it1,rOtherSet.Attribute(*it2)); // todo: optimize 
      ++it1;
      ++it2;
    }
    else { // (*it1 > *it2)
      ++it2;
    }
  }
}

// Attributes(set)
template<class Attr>
void TaIndexSet<Attr>::Attributes(const IndexSet& rOtherSet) {
  FD_DC("TaIndexSet(" << this << ")::Attributes(set) with type " << typeid(rOtherSet.Attribute()).name());
  Detach();
  iterator it1 = pSet->begin();
  iterator it2 = rOtherSet.pSet->begin();
  while ((it1 != pSet->end()) && (it2 != rOtherSet.pSet->end())) {
    if (*it1 < *it2) {
      ++it1;
    }
    else if (*it1 == *it2) {
      Attribute(*it1,rOtherSet.Attribute(*it2)); // todo: optimize 
      ++it1;
      ++it2;
    }
    else { // (*it1 > *it2)
      ++it2;
    }
  }
}

//ClearAttributes()
template<class Attr>
void TaIndexSet<Attr>::ClearAttributes(void) {
  mAttributeMap.clear();
}

//ClrAttribute(idx)
template<class Attr>
void TaIndexSet<Attr>::ClrAttribute(const Idx& idx) {
  mAttributeMap.erase(idx);
}

//Clear
template<class Attr>
void TaIndexSet<Attr>::Clear(void) {
  FD_DC("TaIndexSet(" << this << ")::Clear()");
  IndexSet::Clear();
  mAttributeMap.clear();
}  

// Insert()
template<class Attr>
Idx TaIndexSet<Attr>::Insert(void) {
  FD_DC("TaIndexSet(" << this << ")::Insert()");
  Idx index=MaxIndex()+1;
  Insert(index);
  return index;
}

//Insert(idx)
template<class Attr>
bool TaIndexSet<Attr>::Insert(Idx idx) {
  FD_DC("TaIndexSet(" << this << ")::Insert("<< idx <<")");
  bool ret=TBaseSet<Idx>::Insert(idx);
  //mAttributeMap.erase(idx);
  return ret;
}

// Insert(attr)
template<class Attr>
Idx TaIndexSet<Attr>::Insert(const Attr& attr) {
  FD_DC("TaIndexSet(" << this << ")::Insert(" << attr.ToString() << ")");
  Idx index = Insert();
  if(!attr.IsDefault()) {
    mAttributeMap[index]=attr;
  }
  return index;
}

//Insert(idx,attr)
template<class Attr>
bool TaIndexSet<Attr>::Insert(Idx idx, const Attr& attr) {
  FD_DC("TaIndexSet(" << this << ")::Insert("<< idx <<",attr)");
  bool ret=TBaseSet<Idx>::Insert(idx);
  if(attr.IsDefault())
    mAttributeMap.erase(idx);
  else
    mAttributeMap[idx]=attr;
  return ret;
}


// InsertSet(set)
template<class Attr>
void TaIndexSet<Attr>::InsertSet(const TaIndexSet& rOtherSet) {
  Detach();
  iterator it1 = pSet->begin();
  iterator it2 = rOtherSet.pSet->begin();
  while ((it1 != pSet->end()) && (it2 != rOtherSet.pSet->end())) {
    if (*it1 < *it2) {
      ++it1;
    }
    else if (*it1 == *it2) {
      ++it1;
      ++it2;
    }
    else { // (*it1 > *it2)
      Insert(*it2,rOtherSet.Attribute(*it2)); // todo: optimize 
      ++it2;
    }
  }
  while (it2 != rOtherSet.pSet->end()) {
    Insert(*it2,rOtherSet.Attribute(*it2)); // todo: optimize 
    ++it2;
  }
}

// InsertSet(set)
template<class Attr>
void TaIndexSet<Attr>::InsertSet(const IndexSet& rOtherSet) {
  IndexSet::InsertSet(rOtherSet);
}

//Erase(idx)
template<class Attr>
bool TaIndexSet<Attr>::Erase(Idx idx) {
  bool ret=TBaseSet<Idx>::Erase(idx);
  mAttributeMap.erase(idx);
  return ret;
}


//Erase(pos)
template<class Attr>
typename TaIndexSet<Attr>::Iterator TaIndexSet<Attr>::Erase(const Iterator& pos) { 
#ifdef FAUDES_CHECKED
  if (pos == End()) {
    std::stringstream errstr;
    errstr << "iterator out of range " << std::endl;
    throw Exception("TaIndexSet::Erase", errstr.str(), 60);
  }
#endif
  mAttributeMap.erase(*pos);
  return TBaseSet<Idx>::Erase(pos); 
}


//EraseSet(set)
template<class Attr>
void TaIndexSet<Attr>::EraseSet(const IndexSet& rOtherSet) {
  FD_DC("TaIndexSet(" << this << ")::EraseSet(" << rOtherSet.ToString() << ")");
  Detach();
  // todo: test and optimize 
  iterator tmpit;
  iterator it = pSet->begin();
  iterator oit = rOtherSet.pSet->begin();
  while ((it != pSet->end()) && (oit != rOtherSet.pSet->end())) {
    if (*it < *oit) {
      it=pSet->lower_bound(*oit); // alt: ++it;
    }
    else if (*it == *oit) { 
      mAttributeMap.erase(*it);
      tmpit=it;
      ++it;
      ++oit;
      pSet->erase(tmpit);
    }
    else { // (*it > *oit)
      oit=rOtherSet.pSet->lower_bound(*it); // ++it2;
    }
  }
}


//RestrictSet(set)
template<class Attr>
void TaIndexSet<Attr>::RestrictSet(const IndexSet& rOtherSet) {
  FD_DC("TaIndexSet(" << this << ")::RestrictSet(" << rOtherSet.ToString() << ")");
  Detach();
  // todo: test and optimize 
  iterator tmpit;
  iterator it = pSet->begin();
  iterator oit = rOtherSet.pSet->begin();
  while ((it != pSet->end()) && (oit != rOtherSet.pSet->end())) {
    if (*it < *oit) {
      mAttributeMap.erase(*it);
      tmpit=it;
      ++it;
      pSet->erase(tmpit);
    }
    else if (*it == *oit) { 
      ++it;
      ++oit;
    }
    else { // (*it > *oit)
      ++oit;
    }
  }
}

// SetUnion(set)
template<class Attr>
void TaIndexSet<Attr>::SetUnion(const IndexSet& rOtherSet) {
  IndexSet::SetUnion(rOtherSet);
  mAttributeMap.clear();
}

// SetUnion(set)
template<class Attr>
void TaIndexSet<Attr>::SetUnion(const TaIndexSet& rOtherSet) {
  IndexSet::SetUnion(rOtherSet);
  mAttributeMap.clear();
}

// SetIntersection(set)
template<class Attr>
void TaIndexSet<Attr>::SetIntersection(const IndexSet& rOtherSet) {
  IndexSet::SetIntersection(rOtherSet);
  mAttributeMap.clear();
}

// SetIntersection(set)
template<class Attr>
void TaIndexSet<Attr>::SetIntersection(const TaIndexSet& rOtherSet) {
  IndexSet::SetIntersection(rOtherSet);
  mAttributeMap.clear();
}



} // namespace faudes

#endif 
