/** @file rtiwrapper.h Runtime interface, wrapper from core functions */

    /* FAU Discrete Event Systems Library (libfaudes)

    Copyright (C) 2009 Ruediger Berndt
    Copyright (C) 2009 Thomas Moor

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef FAUDES_RTIWRAPPER_H
#define FAUDES_RTIWRAPPER_H

#include "rtifncts.h"
#include "corefaudes.h"


namespace faudes {


/*
********************************************************************
********************************************************************
********************************************************************

Helper to define Function derivates

********************************************************************
********************************************************************
********************************************************************
*/


/**
 * A TSimpeFunction derives a class from Function to invoke a
 * c-function with parameters of specified type. It must be a simple 
 * function in that it 
 *  - has a unique signature, and
 *  - has all but the last parameter const,
 *  - has no more than three parameters.
 * Technically, a TSimpleFunction is a template
 * that emulates variadic parameters in a rather crude way.
 *
 */

template<class FTYPE0, void (*CFNCT)(FTYPE0&)>
class TSimpleFunction1 : public Function {	
public:
 
  // constructor
  TSimpleFunction1(const FunctionDefinition* fdef) : Function(fdef) {}; 

  // new on heap (factory)
  virtual Function* New() const {return new TSimpleFunction1<FTYPE0,CFNCT>(pFuncDef); };

protected: 

  // typed parameters
  FTYPE0* pp0;

  // type check (variant and parameter number known to be ok)
  virtual void DoTypeCheck(void) {
    FD_DRTI("TSimpleFunction1::DoTypeCheck(): " << pFuncDef->Name() << " #" << pVariant->Size());
    DoTypeCast<FTYPE0>(0,pp0);
    FD_DRTI("TSimpleFunction1::DoTypeCheck(): " << pFuncDef->Name() << " done");
  };

  // execute (parameter cast ok)
  virtual void DoExecute(void) {
    FD_DRTI("TSimpleFunction1::DoExecute(): " << pFuncDef->Name() << " #" << pVariant->Size());
    CFNCT(*pp0);
    FD_DRTI("TSimpleFunction1::DoExecute(): " << pFuncDef->Name() << " done");
  };
};


template<class FTYPE0, class FTYPE1, void (*CFNCT)(const FTYPE0&, FTYPE1&)>
class TSimpleFunction2 : public Function {	
public: 

  // constructor
  TSimpleFunction2(const FunctionDefinition* fdef) : Function(fdef) {}; 

  // new on heap (factory)
  virtual Function* New() const {return new TSimpleFunction2<FTYPE0,FTYPE1,CFNCT>(pFuncDef); };

protected: 

  // typed parameters
  FTYPE0* pp0;
  FTYPE1* pp1;

  // type check (variant and parameter number known to be ok)
  virtual void DoTypeCheck(void) {
    FD_DRTI("TSimpleFunction2::DoTypeCheck(): " << pFuncDef->Name() << " #" << pVariant->Size());
    DoTypeCast<FTYPE0>(0,pp0);
    FD_DRTI("TSimpleFunction2::DoTypeCheck(): " << pFuncDef->Name() << " done");
  };

  // execute (parameter cast ok)
  virtual void DoExecute(void) {
    FD_DRTI("TSimpleFunction2::DoExecute(): " << pFuncDef->Name() << " #" << pVariant->Size());
    CFNCT(*pp0,*pp1);
    FD_DRTI("TSimpleFunction2::DoExecute(): " << pFuncDef->Name() << " done");
  };
};

template<class FTYPE0, class FTYPE1, class FTYPE2, void (*CFNCT)(const FTYPE0&, const FTYPE1&, FTYPE2&)>
class TSimpleFunction3 : public Function {	
public: 

  // constructor
  TSimpleFunction3(const FunctionDefinition* fdef) : Function(fdef) {}; 

  // new on heap (factory)
  virtual Function* New() const {return new TSimpleFunction3<FTYPE0,FTYPE1,FTYPE2,CFNCT>(pFuncDef); };

protected: 

  // typed parameters
  FTYPE0* pp0;
  FTYPE1* pp1;
  FTYPE2* pp2;

  // type check (variant and parameter number known to be ok)
  virtual void DoTypeCheck(void) {
    FD_DRTI("TSimpleFunction3::DoTypeCheck(): " << pFuncDef->Name() << " #" << pVariant->Size());
    DoTypeCast<FTYPE0>(0,pp0);
    DoTypeCast<FTYPE0>(1,pp1);
    DoTypeCast<FTYPE0>(2,pp2);
    FD_DRTI("TSimpleFunction3::DoTypeCheck(): " << pFuncDef->Name() << " done");
  };

  // execute (parameter cast ok)
  virtual void DoExecute(void) {
    FD_DRTI("TSimpleFunction3::DoExecute(): " << pFuncDef->Name() << " #" << pVariant->Size());
    CFNCT(*pp0,*pp1,*pp2);
    FD_DRTI("TSimpleFunction3::DoExecute(): " << pFuncDef->Name() << " done");
  };
};



/**
 * A TNestedFunction derives a class from Function to invoke a
 * c-function repeatavely. Technicalls, it is a template class and
 * it takes the argument type and the c-function as template parameters.
 * The c-function must take three parameters, the first two const.
 *
 */

template<class FTYPE, void (*CFNCT)(const FTYPE&, const FTYPE&, FTYPE&)> 
class TNestedFunction : public Function {	

public: 

  // constructor
  TNestedFunction(const FunctionDefinition* fdef) : Function(fdef) {}; 

  // new on heap (factory)
  virtual Function* New() const {return new TNestedFunction<FTYPE,CFNCT>(pFuncDef); };

protected: 

  // typed parameters
  std::vector<FTYPE*> pvec;

  // type check (variant and parameter number known to be ok)
  virtual void DoTypeCheck(void) {
    FD_DRTI("TNestedFunction::DoTypeCheck(): " << pFuncDef->Name() << " #" << pVariant->Size());
    // allocate typed references
    pvec.resize(pVariant->Size(),NULL);
    // its all integers for this function
    for(int i=0; i < pVariant->Size(); i++) 
      DoTypeCast<FTYPE>(i,pvec.at(i));
    // done
    FD_DRTI("TNestedFunction::DoTypeCheck(): " << pFuncDef->Name() << " done");
  };

  // execute (parameter cast ok)
  virtual void DoExecute(void) {
    FD_DRTI("TNestedFunction::DoExecute(): " << pFuncDef->Name() << " #" << pVariant->Size());
    // signature by numer of args
    int args=pVariant->Size()-1;
    int res =pVariant->Size()-1;
    // cases: #1
    if(args==1) {
      *(pvec[res])=*pvec.at(0);
    }
    // cases: #2+
    if(args>=2) {
      CFNCT(*pvec.at(0),*pvec.at(1),*(pvec[res]));
      // loop sum
      for(int i=2; i < args; i++) 
        CFNCT(*pvec.at(i),*(pvec[res]),*pvec.at(res));
    }; 
    FD_DRTI("TNestedFunction::DoExecute(): " << pFuncDef->Name() << " done");
  }

};



/*
********************************************************************
********************************************************************
********************************************************************

c-Functions to have uniform api

********************************************************************
********************************************************************
********************************************************************
*/



} // namespace


#endif // header
