/** @file mtc_attributes.h

Implements the color representation for single states in multitasking
automata as state attributes

*/

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2008  Matthias Singer
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef MTC_ATTRIBUTES_H

#include "tokenreader.h"
#include "tokenwriter.h"
#include "attributes.h"
#include "mtc_colorset.h"

namespace faudes {

/**
 * 
 * State attributes for multitasking automata
 *
 * @ingroup MultitaskingPlugin
 */
class AttributeColoredState : public AttributeFlags {

  public:

    /** Default constructor */
    AttributeColoredState(void) {
      mpColorSymbolTable=mColors.SymbolTablep(); };

    ~AttributeColoredState(void) {}

   /**
    * Construct on heap. 
    * Technically not a constructor, this function creates an object with the
    * same type AttributeVoid. 
    * @return 
    *   Pointer to new aoid object
    */
    virtual AttributeColoredState* New(void) const { return new AttributeColoredState(); };

   /**
    * Copy attribute. Virtual interface to copy members using a
    * dynamic cast to figure the actual type of the destination. 
    * If the latter is known at compiletime, performance is better
    * via the assignment operator.
    *
    * @param rOtherAttr 
    *    Destination to copy to
    */
   virtual void Copy(AttributeVoid& rOtherAttr) const;

    /**
    * Test for default value
    *
    * @return
    *   True for defaultvalue
    */
    bool  IsDefault(void) const {return (mColors.Empty() && (mFlags==mDefFlags)); };

   /**
    * Iterator for first entry in mColors
    */
    NameSet::Iterator Begin();

   /**
    * Iterator for last entry in mColors 
    */
    NameSet::Iterator End();

    /**
    * Set pointer to color symbol table
    *
    * @param pSymTab
    *   New color symbol table
    */
    void ColorSymTab(SymbolTable *pSymTab);

    /**
    * Get pointer to color symbol table
    *
    * @return
    *   Pointer to current color symbol table
    */
    SymbolTable *ColorSymTabp(void) const;

    /** Colorset */
    ColorSet mColors;

protected:

   /** Pointer to colorsymboltable */
    SymbolTable* mpColorSymbolTable;

   /**
    * Copy attribute members
    *
    * @param rOtherAttr 
    *    Destination to copy to
    */
    void DoCopy(AttributeColoredState& rOtherAttr) const; 


   /**
    * Write attributes to tokenwriter
    *
    * @param rTw
    *   Tokenreader to write to
    * @param rLabel
    *   Section to write
    * @param pContext
    *   Write context to provide contextual information
    *
    * @exception Exception
    *   I/O errors
    */
    virtual void DoWrite(TokenWriter& rTw, const std::string& rLabel="", const Type* pContext=0) const;

    /**
    * Read attributes from tokenreader
    *
    * @param rTr
    *   Tokenreader to read from
    * @param rLabel
    *   Section to read
    * @param pContext
    *   Read context to provide contextual information
    *
    * @exception Exception
    *   Thrown on file format errors (ids 50, 51)
    */
    virtual void DoRead(TokenReader& rTr,const std::string& rLabel="", const Type* pContext=0);



}; // class AttributeColoredState

} // namespace faudes

#define MTC_ATTRIBUTES_H
#endif
