/** @file sp_simconditionset.cpp  Set of named simulation conditions */


/* 
   Copyright (C) 2008  Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt
*/

#include "sp_simconditionset.h"
#include "sp_lpexecutor.h"

namespace faudes {

/*
************************************************
************************************************
************************************************

implementation of attributes

************************************************
************************************************
************************************************
*/



//DoWrite(rTw);
void SimConditionAttribute::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) rLabel;
  Token token;

  // report context
  FD_DC("SimConditionAttribute::DoWrite(rTr," << rLabel <<", " << pContext << "): cast " <<
     dynamic_cast<const ParallelExecutor*>(pContext));

  // write event condition
  if(mEventCondition) {
    rTw.WriteBegin("EventCondition");
    if(mEnabled) rTw.WriteOption("Enabled");
    if(!mEnabled) rTw.WriteOption("Disabled");
    if(mBreakCondition) rTw.WriteOption("Break");
    mEventConditionAttribute.mStart.Write(rTw,"StartEvents");
    mEventConditionAttribute.mStop.Write(rTw,"StopEvents");
    rTw.WriteEnd("EventCondition");
  }

  // write state condition
  if(mStateCondition) {
    rTw.WriteBegin("StateCondition");
    if(mEnabled) rTw.WriteOption("Enabled");
    if(!mEnabled) rTw.WriteOption("Disabled");
    if(mBreakCondition) rTw.WriteOption("Break");
    if(mStateConditionAttribute.mAllFlag) rTw.WriteOption("Conjunction");
    if(!mStateConditionAttribute.mAllFlag) rTw.WriteOption("Disjunction");
    // get context
    const ParallelExecutor* executor= dynamic_cast<const ParallelExecutor*>(pContext);
    Idx n=mStateConditionAttribute.mStateSets.size();
    // test context
    if(executor)
      if(executor->Size()!=n) {
        std::stringstream errstr;
        errstr << "Invalid ParalleExecutor context (dimension mismatch)";
        throw Exception("SimConditionAttribute::DoWrite", errstr.str(), 501);
      }
    // write the sets
    for(Idx i=0; i<n; i++) {
      const StateSet& set=mStateConditionAttribute.mStateSets.at(i);
      if(executor) executor->At(i).Generator().WriteStateSet(rTw,set);
      else set.Write(rTw,"StateSet");
    }
    rTw.WriteEnd("StateCondition");
  }
}

//DoRead(rTr)
void SimConditionAttribute::DoRead(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  (void) rLabel;
  bool err;
  Token token;

  // report context
  FD_DC("SimConditionAttribute::DoRead(rTr," << rLabel <<", " << pContext << "): cast " <<
     dynamic_cast<const ParallelExecutor*>(pContext));

  // clear myself
  err=false;
  mEventCondition=false;
  mStateCondition=false;
  mEnabled=true;
  mBreakCondition=false;
  mEventConditionAttribute.mStart.Clear();
  mEventConditionAttribute.mStop.Clear();
  mStateConditionAttribute.mStateSets.clear();
  mStateConditionAttribute.mAllFlag=false;

  // switch by toke
  rTr.Peek(token);

  // read event condition
  if(token.Type()==Token::Begin && token.StringValue()=="EventCondition") {
    rTr.ReadBegin("EventCondition");
    mEventCondition=true;
    while(!rTr.Eos("EventCondition")) {
      rTr.Peek(token);
      if((token.Type()==Token::Option) && (token.StringValue()=="Enabled")) {
	mEnabled=true;
	rTr.Get(token);
        continue;
      }
      if((token.Type()==Token::Option) && (token.StringValue()=="Disabled")) {
	mEnabled=false;
	rTr.Get(token);
        continue;
      }
      if((token.Type()==Token::Option) && (token.StringValue()=="Break")) {
	mBreakCondition=true;
	rTr.Get(token);
        continue;
      }
      if((token.Type()==Token::Begin) && (token.StringValue()=="StartEvents")) {
        mEventConditionAttribute.mStart.Read(rTr,"StartEvents");
        continue;
      }
      if((token.Type()==Token::Begin) && (token.StringValue()=="StopEvents")) {
        mEventConditionAttribute.mStop.Read(rTr,"StopEvents");
        continue;
      }
      // ignore errors
      rTr.Get(token);
    }
    rTr.ReadEnd("EventCondition");
  }

  // read state condition
  if(token.Type()==Token::Begin && token.StringValue()=="StateCondition") {
    rTr.ReadBegin("StateCondition");
    mStateCondition=true;
    while(!rTr.Eos("StateCondition")) {
      rTr.Peek(token);
      if((token.Type()==Token::Option) && (token.StringValue()=="Enabled")) {
	mEnabled=true;
	rTr.Get(token);
        continue;
      }
      if((token.Type()==Token::Option) && (token.StringValue()=="Disabled")) {
	mEnabled=false;
	rTr.Get(token);
        continue;
      }
      if((token.Type()==Token::Option) && (token.StringValue()=="Break")) {
	mBreakCondition=true;
	rTr.Get(token);
        continue;
      }
      if((token.Type()==Token::Option) && (token.StringValue()=="Conjunction")) {
	mStateConditionAttribute.mAllFlag=true;
	rTr.Get(token);
        continue;
      }
      if((token.Type()==Token::Begin) && (token.StringValue()=="StateSet")) {
        // get context
        const ParallelExecutor* executor= dynamic_cast<const ParallelExecutor*>(pContext);
        Idx i=mStateConditionAttribute.mStateSets.size();
        // test context
        if(executor)
          if(executor->Size()<=i) {
            std::stringstream errstr;
            errstr << "Invalid ParalleExecutor context (dimension mismatch)";
            throw Exception("SimConditionAttribute::DoRead", errstr.str(), 501);
          }
        // read the set
        StateSet set;
        if(executor) executor->At(i).Generator().ReadStateSet(rTr,"StateSet",set);
        else set.Read(rTr,"StateSet");
        set.Name("StateSet");
        mStateConditionAttribute.mStateSets.push_back(set);
        continue;
      }
      // ignore errors
      rTr.Get(token);
    }
    rTr.ReadEnd("StateCondition");
  }
  // report error
  if(err) {
    std::stringstream errstr;
    errstr << "invalid simulation condition" << rTr.FileLine();
    throw Exception("SimConditionAttribute::Read", errstr.str(), 52);
  }
}

// condition state: sample
void SimConditionAttribute::Satisfied(bool on, tpTime::Type now) {
  if(on==mSatisfied) return;
  mSatisfied=on;
  // newly activated
  if(mSatisfied) {
    if(mActivationTime>0) mSamplesPeriod.Sample(now-mActivationTime); 
    mActivationTime=now;
  }
  // newly deactivated
  if(!mSatisfied) {
    mSamplesDuration.Sample(now-mActivationTime); 
  }
}


/*
************************************************
************************************************
************************************************

implementation of condition set

************************************************
************************************************
************************************************
*/

// get enabled set
SimConditionSet SimConditionSet::EnabledConditions(void) {
  SimConditionSet result;
  for(Iterator cit=Begin(); cit != End(); ++cit) 
    if(Enabled(*cit)) result.Insert(*cit,Attribute(*cit));
  return result;
}

// condition state: reset alls
void SimConditionSet::Reset(void) {
  for(Iterator cit=Begin(); cit != End(); ++cit) {
    SimConditionAttribute* pattr=Attributep(*cit);
    pattr->Reset();
    pattr->mSamplesPeriod.Name(SymbolicName(*cit)+" - Period");
    pattr->mSamplesDuration.Name(SymbolicName(*cit)+" - Duration");
  }
}



} // namespace faudes

