/** @file sp_simconditionset.h  Set of named simulation conditions */

/*
   Copyright (C) 2008  Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt
*/


#include "tokenreader.h"
#include "tokenwriter.h"
#include "nameset.h"
#include "attributes.h"
#include "sp_densityfnct.h"


#ifndef FAUDES_SP_SIMCONDITIONSET_H
#define FAUDES_SP_SIMCONDITIONSET_H

namespace faudes {

/** 

@defgroup SimConditionAttributes Simulation Condition Attributes 

@ingroup SimulatorPlugin 

Attributes to define simulation conditions used for statistic analysis
of simulation runs.

@copydoc faudes::SimConditionAttribute

@{ 

*/

/**  
 * Defining data of event based simulation condition
 *
 * An event based condition has a set of start events and a set of
 * stop events. It becomes satisfied when a start event occurs and it 
 * becomes unsatisfied when a stop event occurs. Thus, when start or stop
 * events fail to alternate, the first event of the respective type is relevant.
 *
 */
class SimEventConditionAttribute {
 public:
  EventSet mStart;
  EventSet mStop;
};

/**  
 * Defining data of state based simulation condition
 *
 * A state based condition has a set of states per generator an is satisfied
 * while the components of the current discrete state vector are within the
 * respective set. A flag indicates whether each or some component 
 * must comply with the state set to satisfy the condition.
 * 
 * Note: Since state indices are local to generators, a SimStateConditionAttribute 
 * is only meaningful in the context of a particular ParallelExecutor. 
 *
 */

// state based condition
class  SimStateConditionAttribute {
public:
  bool mAllFlag;
  std::vector<StateSet> mStateSets;
};


/**
 * Attribute for a simulation condition
 *
 * In order to extract statistical data from a simulation by a LoggingExecutor
 * or some derived class, so called simulation conditions
 * are defined. At any instance of time, a condition is satisfied or
 * dissatisfied. Statistical data can then be requested regarding 
 * the period and duration of each condition. Currently, two
 * types of conditions are available:
 *
 * - SimStateConditionAttribute addresses conditions of the current state
 * - SimEventConditionAttribute addresses the external event sequence.
 *
 * A Condition may be flagged as a break condition to halt simulation
 * when satisfied. A condition may be enabled for tracking or not.
 *
 * The class SimConditionAttribute summarizes all data to represent
 * faudes simulation conditions. It also holds some state of the condition
 * wrt execution and provides an interface for sampling. The latter 
 * may be seperated to a different class in a future revision. The class
 * SimConditionAttribute does, however, not implement any test whether
 * or not a condition is satisfied. This is done by the LoggingExecutor.
 *
 * As a faudes attribute, conditions can be referenced by 
 * names the via std faudes container TaNameSet. For token io, a
 * ParallelExecutor should be provided as context to access symbolic
 * state names. The file format of a set of
 * events equipped with condition attributes is illustrated by the below example
 * to monitor some performance aspects of one simple machine:
 * 
 * @code
 *
 * <Conditions>
 *
 * % monitor when the machine is idle
 * "IdleCond"      
 * <StateCondition>
 * <StateSet> "idle" </StateSet>
 * </StateCondition>
 *
 * % halt simulation  when the machine is down
 * % however, the condition is disabled
 * "DownCond"      
 * +Break+     
 * +Disabled+     
 * <StateCondition>
 * <StateSet> "down" </StateSet>
 * </StateCondition>
 *
 * % monitor how long the prozessing one work piece takes
 * "OperationCond"     
 * <EventCondition>
 * <StartEvents>  "alpha" </StartEvents>
 * <StopEvents>   "beta"  </StopEvents>
 * </EventCondition>
 *
 * </Conditions>
 *
 * @endcode
 */

class SimConditionAttribute : public AttributeVoid {

  public:

  /** Convenience typedef */
  typedef std::vector<StateSet>::iterator Iterator;
  typedef std::vector<StateSet>::const_iterator CIterator;

  /** 
   * Default constructor. Cinstructs a SimConditionAttribute of 
   * with no type ie neither state condition nor event condition.
   */
  SimConditionAttribute(void) : AttributeVoid() { 
    mEventCondition=false; mStateCondition=false; mEnabled=true; mBreakCondition=false; Reset();};

 /**
  * Construct on heap. 
  * Technically not a constructor, this function creates an object with the
  * same type. It is the callers reponsabilty to delete the object 
  * when no longer needed. 
  *
  * @return 
  *   Pointer to new attribute object
  */
  virtual SimConditionAttribute* New(void) const { return new SimConditionAttribute(); };

  /** 
   * Test for default value 
   * 
   * @return
   *   True, if this attribute has its default value
   */
  virtual bool IsDefault(void) const {
    return !mEventCondition && !mStateCondition && mEnabled && !mBreakCondition; };


  /** 
   * Test for state condition.
   *
   * @return
   *   True, if this attribute defines a state condition
   */
  bool IsStateCondition(void) const {return mStateCondition; };

  /** 
   * Test for event condition.
   *
   * @return
   *   True, if this attribute defines an event condition
   */
  bool IsEventCondition(void) const {return mEventCondition; };

  /** 
   * Test for break condition.
   *
   * @return
   *   True, if this condition halts simulation.
   */
  bool IsBreakCondition(void) const {return mBreakCondition; };

  /** 
   * Test whether condition is enabled.
   *
   * @return
   *   True, if this condition is monitored during simulation.
   */
  bool IsEnabled(void) const {return mEnabled; };

  /** 
   * Set state condition attribute. Define this attribute to
   * represent the specified state condition.
   *
   * @param rStateConditionAttribute
   *     Define state condition
   */
  void StateCondition(const SimStateConditionAttribute&  rStateConditionAttribute) 
    {mStateConditionAttribute=rStateConditionAttribute; mStateCondition=true; mEventCondition=false;
    for(Idx i=0; i<mStateConditionAttribute.mStateSets.size(); i++) 
      mStateConditionAttribute.mStateSets.at(i).Name("StateSet");
  };

  /** 
   * Set event condition attribute. Define this attribute to
   * represent the specified event condition.
   *
   * @param rEventConditionAttribute
   *     Define event condition
   */
  void EventCondition(const SimEventConditionAttribute&  rEventConditionAttribute) 
    {mEventConditionAttribute=rEventConditionAttribute; mEventCondition=true; mStateCondition=false;};

  /** 
   * Set break flag.
   * 
   * @param on
   *   True, to indicate that this condition halts simulation.
   */
  void BreakCondition(bool on) {mBreakCondition=on; };

  /** 
   * Set enabled flag.
   * 
   * @param on
   *   True, to indicate that this condition is to be monitored during simulation.
   */
  void Enabled(bool on) {mEnabled=on; };

  /**  
   * Get event condition. Note that the attribute can only return
   * meaningful data if it actually is an event condition.  
   *
   *  @return
   *    Defining data of this attribute's event condition.
   *
   */
  const SimEventConditionAttribute& EventCondition(void) const {return mEventConditionAttribute; };

  /**  
   * Get state condition. Note that the attribute can only return
   * meaningful data if it actually is an state condition.  
   *
   *  @return
   *    Defining data of this attribute's state condition.
   *
   */
  const SimStateConditionAttribute& StateCondition(void) const {return mStateConditionAttribute; };

  /**
   *  Reset conditions execution state. The execution state of a condition
   *  consists of all data accumulated during simulation eg statistical data
   *  and whether or not the condition is currently satisfied. The execution state
   *  resides in the attribute for pragmatic reasons only.
   */
  void Reset(void) 
  { mSatisfied=false; mActivationTime=-1; mSamplesPeriod.Clear(); mSamplesDuration.Clear(); };

  /** 
   * Test whether the condition is currently satisfied. This is part
   * of the condition execution state. 
   *
   * @return
   *   True, if the conditions is considered satisfied
   */
  bool Satisfied(void) const { return mSatisfied; };

  /** 
   * Set the condition to be satisfied. This is part
   * of the condition execution state. Since it is the executor
   * that determines whether a condition is satisfied, and since
   * the condition state resides in the attribute, the executor
   * is meant to notify state changes.
   *
   * @param on
   *   True, if the conditions is considered satisfied
   * @param now
   *   Time at which the state change occures
   */
  void Satisfied(bool on, tpTime::Type now);
   
  /** 
   * Sampled period, at which this condition becomes satisfied. 
   */
  SampledDensityFunction mSamplesPeriod;

  /** 
   * Sampled durations, for which this condition remains satisfied. 
   */
  SampledDensityFunction mSamplesDuration;

protected:
 
  /** Indicate precense of a event condition */
  bool mEventCondition;

  /** Indicate precense of a state condition */
  bool mStateCondition;

  /** Indicate that we halt simulation when this condition is satisfied */
  bool mBreakCondition;

  /** Indicate that this condition should be considere at all */
  bool mEnabled;

  /** Event based condotion data */
  SimEventConditionAttribute mEventConditionAttribute;

  /** State based condotion data */
  SimStateConditionAttribute mStateConditionAttribute;

  /** Condotion state: recorded as satisfied */
  bool mSatisfied;

  /** Condition state: when last satisfied became true */
  tpTime::Type mActivationTime;
   

 protected:

  /**
   * Reads the attribute from TokenReader, see AttributeVoid for public wrappers.
   * 
   * If the current token indicates a condition section, the method reads the  
   * condition data from that section. Else it does nothing. Exceptions may only be thrown
   * on invalid data within the condition section. The label argiment is ignored, we 
   * use hardcoded labled "EventCondition" and "StateCondition" to figure the type of 
   * condition. When a ParallelExecutor is provided as context, it is used to interpret
   * symbolc state names of a state condition.
   *
   * @param rTr
   *   TokenReader to read from
   * @param rLabel
   *   Section to read
   * @param pContext
   *   Read context to provide contextual information
   *
   * @exception Exception
   *   - IO error (id 1)
   */
  virtual void DoRead(TokenReader& rTr, const std::string& rLabel="", const Type* pContext=0);
 
  /**
   * Writes the attribute to TokenWriter, see AttributeVoid for public wrappers.
   *
   * Writes a condition section to include data on state- or event-condition. The label argument
   * is ignored, we use hardcoded keywords "EventCondition" and StateCondition". When a 
   * ParallelExecutor is provided as context, it state conditions are written with 
   * symbolic state names.
   *
   * @param rTw
   *   TokenWriter to write to
   * @param rLabel
   *   Section to write
   * @param pContext
   *   Read context to provide contextual information
   *
   * @exception Exception
   *   - IO error (id 2)
   */
  virtual void DoWrite(TokenWriter& rTw,const std::string& rLabel="", const Type* pContext=0) const;



}; // class SimConditionAttribute


/**
 * Set of simulation named conditions. 
 *
 * Note: we currently share a symboltabel with the global
 * event set. This will definitely change in a future revision.
 */

class SimConditionSet : public TaNameSet<SimConditionAttribute> {

 public:

  /** Default constructor */
  SimConditionSet(void) : TaNameSet<SimConditionAttribute>() {};

  /** Copy constructor */
  SimConditionSet(const SimConditionSet& rOtherSet) : IndexSet(), TaNameSet<SimConditionAttribute>(rOtherSet) {};

  /**  Virtual destructor */
  virtual ~SimConditionSet(void) {};

  /** Assignment operator */
  virtual SimConditionSet& operator=(const SimConditionSet& rSrc) { rSrc.Copy(*this); return *this;};

  /** Test condition for enabled */
  bool Enabled(Idx cond) const { return Attribute(cond).IsEnabled(); };

  /** Set condition enabled */
  void Enabled(Idx cond, bool on) { Attributep(cond)->Enabled(on); };

  /** Get set of enabled conditions */
  SimConditionSet EnabledConditions(void);

  /** Reset all condition states */
  void Reset(void);

};



/** @} doxygen group */

} // namespace

#endif
