/** @file helper.cpp   Helper functions */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "helper.h"

/* Include extra defs for mswindows */
#ifdef FAUDES_WINEXTRA
#include "winextra.h"
#endif


namespace faudes {

// ToStringInteger(number)
std::string ToStringInteger(long int number) { 
  if(number>= std::numeric_limits<long int>::max()) return "inf";
  if(number<= std::numeric_limits<long int>::min()+1) return "-inf";
  std::string res;
  std::stringstream sstr;
  sstr << number;
  sstr >> res;
  return res;
}

// ToStringInteger16(number)
std::string ToStringInteger16(long int number) { 
   std::string res;
   std::stringstream sstr;
   sstr << "0x" << std::setbase(16) << number;
   sstr >> res;
   return res;
}

// ToStringFloat(number)
// todo: check range, prevent sci notation
std::string ToStringFloat(double number) { 
   if(number == (long int) number) 
     return(ToStringInteger((long int) number));
   std::string res;
   std::stringstream sstr;
   sstr << std::fixed;
   sstr << number;
   sstr >> res;
   return res;
}

// ExpandString(rString, len) 
std::string ExpandString(const std::string& rString, unsigned int len) {
  std::string res;
  res	= rString;
  std::string::size_type xtra = (std::string::size_type) len - rString.length();
  if ((xtra > 0) && (xtra < 10000)) {
	  res.append(xtra, ' ');
	}
  return res;
} 

// CollapseString(rString, len) 
std::string  CollapsString(const std::string& rString, unsigned int len) {
  if(len <=1) return rString;
  if(rString.length() <= len) return rString;
  int chead = len/2;
  int ctail = len-chead;
  return rString.substr(0,chead) + "..." + rString.substr(rString.length()-ctail,ctail);
} 

// ToIdx(rString)
Idx ToIdx(const std::string& rString) {
	char * end;
	unsigned long ul = strtoul (rString.c_str(), &end, 0);
	unsigned long idxmax = std::numeric_limits<Idx>::max();
	if (ul > idxmax) {
		throw Exception("atoidx", "Idx overflow", 600);
	}
	else {
		return (Idx) ul;
	}
}

// String Substitution
std::string StringSubstitute(const std::string& rString,const std::string& rFrom,const std::string& rTo) {
  // prep result
  std::string res;
  std::size_t pos=0; 
  // loop over occurences of "from"
  while(pos<rString.length()) {
    std::size_t next=rString.find(rFrom,pos);
    if(next==std::string::npos) break;
    res.append(rString.substr(pos, next));
    res.append(rTo);
    pos=next+rFrom.length();
  }
  // get end
  if(pos<rString.length()) 
    res.append(rString.substr(pos));
  // done
  return res;
}

// FDVersionString()
std::string FDVersionString() {
	return std::string(FAUDES_VERSION);
}

// FDPluginsString()
std::string FDPluginsString() {
  return std::string(FAUDES_PLUGINS);
}

// FDContributorsString()
std::string FDContributorsString() {
  return 
    "Berndt, Breindel, Doerr, Duevel, Franchi, Hellenschmidt, Mohr, Moor, Musunoi, "
    "Opitz, Petzoldt, Perk, Rempel, Ritter, Schlein, Schmidt, Singer, Wittmann, Zaddach, et al";
}


// ProcessDot(infile,outfile,format)
void ProcessDot(const std::string& rDotFile, 
  const std::string& rOutFile, const std::string& rOutFormat, const std::string& rDotExec)
{
  std::string format=rOutFormat;
  // guess format from filename
  if(format=="") {
    if(rOutFile.rfind('.')+1 < rOutFile.size()) {
       format=rOutFile.substr(rOutFile.rfind('.')+1);
    }
  }  
  // test format
  if (format == "canon"); 
  else if (format == "dot"); 
  else if (format == "xdot"); 
  else if (format == "cmap");
  else if (format == "dia");
  else if (format == "fig"); 
  else if (format == "gd"); 
  else if (format == "gd2"); 
  else if (format == "gif"); 
  else if (format == "hpgl"); 
  else if (format == "imap"); 
  else if (format == "cmapx");
  else if (format == "ismap"); 
  else if (format == "jpg"); 
  else if (format == "jpeg"); 
  else if (format == "mif"); 
  else if (format == "mp"); 
  else if (format == "pcl"); 
  else if (format == "pic"); 
  else if (format == "plain"); 
  else if (format == "plain-ext"); 
  else if (format == "png"); 
  else if (format == "ps"); 
  else if (format == "ps2"); 
  else if (format == "svg"); 
  else if (format == "svgz"); 
  else if (format == "vrml"); 
  else if (format == "vtx"); 
  else if (format == "wbmp"); 
  else {
    std::stringstream errstr;
    errstr << "Dot output format \"" << format << "\" unknown";
    throw Exception("faudes::ProcessDot", errstr.str(), 3);
  }
  std::string dotcommand = rDotExec + " -T"+format+" \""+rDotFile+"\" -o \""+rOutFile+"\"";
  if (system(dotcommand.c_str()) != 0) {
    throw Exception("faudes::ProcessDot", 
        "Error in running " + dotcommand, 3);
  }
}


// CreateTempFile(void)
// todo: sys dependant, report, investigate threads
std::string CreateTempFile(void) {
  char filename[]= "faudes_temp_XXXXXX";
  int filedes = mkstemp(filename);
  if(filedes==-1) {
    FD_DF("faudes::CreateTempFile(): error");
    return "";
  }
  close(filedes);
  std::string res(filename);
  FD_DF("faudes::CreateTempFile(): " << res);
  return(res);
}


// RemoveFile(void)
// todo: sys dependant *
bool RemoveFile(const std::string& rFileName) {
  return std::remove(rFileName.c_str()) == 0;
}


#ifndef FAUDES_DIR_SEPARATORS 
#define FAUDES_DIR_SEPARATORS "/"
#endif

// ExtractPath
std::string ExtractPath(const std::string& rFullName) {
  std::string res="";
  unsigned int seppos = rFullName.find_last_of(FAUDES_DIR_SEPARATORS);
  if(seppos==std::string::npos) return res;
  res=rFullName.substr(0,seppos+1);
  return res;
}

// PrependPath
std::string PrependPath(const std::string& rPathName, const std::string& rFileName) {
  std::string res=rPathName;
  char sepchar=std::string(FAUDES_DIR_SEPARATORS).at(0);
  if(res!="")
  if(res.at(res.length()-1)!=sepchar)
    res.append(1,sepchar);
  res.append(rFileName);
  return res;
}

// Debug/Error Logging
std::ostream* pDebugStream=NULL; 
std::ostream* DebugStream(void) { return pDebugStream ? pDebugStream : &std::cerr;}
void DebugToStream(std::ostream* pStream) {
  pDebugStream = pStream;
}
void DebugToFile(const std::string& rFilename) {
  if(pDebugStream) delete pDebugStream;
  std::ofstream* fdebug = new std::ofstream();
  fdebug->open(rFilename.c_str());
  DebugToStream(fdebug);
}

} // namespace faudes
