/** rti2hmtl.cpp  Utility to generate a html documents from rti files */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2009 Ruediger Berndt
Copyright (C) 2009 Thomas Moor

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */
	 

#include <string>
#include <cctype>
#include <iostream>
#include <fstream>
#include "corefaudes.h"


using namespace faudes;

// ******************************************************************
// error exit
// ******************************************************************

void Usage(const std::string& rMessage="") {
  // ui hints
  std::cerr << rMessage << std::endl;
  std::cerr << "utility to generate an html snippets from an libFAUDES" << std::endl;
  std::cerr << "run-time interface definition file" << std::endl;
  std::cerr << std::endl << "usage: " << std::endl;
  std::cerr << " rti2html <rti input file> -index <html output file>" << std::endl;
  std::cerr << " rti2html <rti input file> -plugin <plugin name> <html output file>" << std::endl;
  std::cerr << " rti2html <rti input file> -function <function name> <html output file>" << std::endl;
  std::cerr << " rti2html <rti input file> -type <type name> <html output file>" << std::endl;
  std::cerr << " rti2html <rti input file> -signature <function name> <html output file>" << std::endl;
  std::cerr << " rti2html <rti input file> -shortdoc <function/type name> <html output file>" << std::endl;
  std::cerr << std::endl << "note: use \"-\" as output file for console output" << std::endl;
  exit(1);
}


// ******************************************************************
// helper: name of plugin by keyword (first word, fix case)
// ******************************************************************

std::string PluginName(const std::string& key) {
  std::string pname = key;
  std::size_t spos = pname.find(" ");
  if(spos!=std::string::npos) pname=pname.substr(0,spos);
  if(pname.length()>0) pname.at(0)=toupper(pname.at(0));
  return pname;
}

// ******************************************************************
// helper: html for linked type name
// ******************************************************************

void TypeHtml(std::ostream* pStream, const std::string& rTypeName) {

  // just text if unknown
  if(!TypeRegistry::G()->Exists(rTypeName)) {
    *pStream << rTypeName;
    return;
  }

  // retrieve definition
  const TypeDefinition& tdef=TypeRegistry::G()->Definition(rTypeName);
  std::string tyname = tdef.Name();
  std::string tyhtml = tdef.HtmlDoc();
  if(tyhtml=="" || tyhtml=="none") {
    *pStream << tyname;
  } else {
    *pStream << "<a href=\"" << tyhtml << "\">" << tyname << "</a>";
  }
}

// ******************************************************************
// helper: html for linked function name
// ******************************************************************

void FunctionHtml(std::ostream* pStream, const std::string& rFunctionName) {

  // just text if unknown
  if(!FunctionRegistry::G()->Exists(rFunctionName)) {
    *pStream << rFunctionName;
    return;
  }

  // retrieve definition
  const FunctionDefinition& fdef=FunctionRegistry::G()->Definition(rFunctionName);
  std::string fname = fdef.Name();
  std::string fhtml = fdef.HtmlDoc();
  if(fhtml=="" || fhtml=="none") {
    *pStream << fname;
  } else {
    *pStream << "<a href=\"" << fhtml << "\">" << fname << "</a>";
  }
}

// ******************************************************************
// type index by keyword (aka plugin)
// ******************************************************************

void TypeIndexHtml(std::ostream* pIndexFile, const std::string& key="") {

  // traverse type registry
  bool head=false;
  for(TypeRegistry::Iterator tit=TypeRegistry::G()->Begin(); 
    tit!=TypeRegistry::G()->End(); tit++) 
  {
    // test for exact key
    if(key!="") {
      std::string match= tit->second->MatchKeyword(key);
      std::transform(match.begin(), match.end(), match.begin(), tolower);
      if(match!=key) continue;
    }
    // test for not in a plugin
    if(key=="") if(tit->second->MatchKeyword("plugin")!="") continue;
    // test for user doc
    if(tit->second->TextDoc()=="") continue;
    // head line
    if(!head) {
      head=true;
      *pIndexFile << "<div class=\"registry_heading\"><strong>Types</strong></div>" << std::endl;
    }
    // index entry for this type
    *pIndexFile << "<div class=\"registry_index\">";
    TypeHtml(pIndexFile,tit->second->Name());
    *pIndexFile << "</div>" << std::endl;
  }
  // done
  if(head) *pIndexFile << "<br>" << std::endl;
}

// ******************************************************************
// function index by keyword (aka plugin)
// ******************************************************************

  void FunctionIndexHtml(std::ostream* pIndexFile, const std::string& key="") {

  // traverse function registry
  bool head=false;
  for(FunctionRegistry::Iterator fit=FunctionRegistry::G()->Begin(); 
    fit!=FunctionRegistry::G()->End(); fit++) 
  {
    // test for key
    if(key!="") {
      std::string match= fit->second->MatchKeyword(key);
      std::transform(match.begin(), match.end(), match.begin(), tolower);
      if(match!=key) continue;
    }
    // test for not in a plugin
    if(key=="") if(fit->second->MatchKeyword("plugin")!="") continue;
    // test for user doc
    if(fit->second->TextDoc()=="") continue;
    // index entry for this function
    std::string fname = fit->second->Name();
    std::string fhtml = fit->second->HtmlDoc();
    if(fhtml=="") continue;
    // head line
    if(!head){
      head=true;
      *pIndexFile << "<div class=\"registry_heading\"><strong>Functions</strong></div>" << std::endl;
    }
    // list entry
    if(fhtml=="none") {
      *pIndexFile << "<div class=\"registry_index\"> "<< fhtml << "</div>" << std::endl;
      continue;
    }
    *pIndexFile << "<div class=\"registry_index\"><a href=\"" << fhtml << "\">" 
      << fname << "</a></div>" << std::endl;
  }

  // done
  if(head) *pIndexFile << "<br>" << std::endl;

}

// ******************************************************************
// html index by keywork (aka plugin)
// ******************************************************************

  void IndexHtml(std::ostream* pIndexFile, const std::string& key="") {

  // prepare list of all plugins
  std::set< std::string > plugins;
  for(TypeRegistry::Iterator tit=TypeRegistry::G()->Begin(); 
    tit!=TypeRegistry::G()->End(); tit++) 
  {
    std::string pkey=tit->second->MatchKeyword("plugin");
    if(pkey!="") plugins.insert(pkey);
  }
  for(FunctionRegistry::Iterator fit=FunctionRegistry::G()->Begin(); 
    fit!=FunctionRegistry::G()->End(); fit++) 
  {
    std::string pkey=fit->second->MatchKeyword("plugin");
    if(pkey!="") plugins.insert(pkey);
  }

  // plugins headline
  if(plugins.size()!=0) 
    *pIndexFile << "<div class=\"registry_heading\"><strong>PlugIns</strong></div>" << std::endl;

  // list plugins
  std::set< std::string >::iterator pit;
  for(pit=plugins.begin(); pit != plugins.end(); pit++) {
    // get nice name
    std::string pname = PluginName(*pit);
    // have link
    std::string phtml = pname+"_index.html";
    std::transform(phtml.begin(),phtml.end(),phtml.begin(),tolower);
    // have entry
    *pIndexFile << "<div class=\"registry_index\"><a href=\"" << phtml << "\">" 
      << pname << "</a></div>" << std::endl;
  }

  // plugins done
  if(plugins.size()!=0) 
    *pIndexFile << "<br>" << std::endl;

  // list types
  TypeIndexHtml(pIndexFile,key);

  // list functions
  FunctionIndexHtml(pIndexFile,key);
}

// ******************************************************************
// signature
// ******************************************************************

void SignatureHtml(std::ostream* pOutFile, std::string function) {

  // bail out on non existence
  if(!FunctionRegistry::G()->Exists(function)) return;

  // function definition
  const FunctionDefinition& fdef=FunctionRegistry::G()->Definition(function);
  
  // bail out if there is no signature
  if(fdef.VariantsSize()==0) return;

  // start signature box
  *pOutFile << "<div class=\"registry_signature\">" << std::endl;
  *pOutFile << "<p><strong>Signature:</strong></p>" << std::endl;
  *pOutFile << "<p>" << std::endl;

  // loop signatures
  for(int i=0; i< fdef.VariantsSize(); i++) {
    const Signature& sigi=fdef.Variant(i);
    *pOutFile << fdef.Name() << "(";
    // loop params
    for(int j=0; j < sigi.Size(); j++) {
      if(j!=0) *pOutFile << ", ";
      const Parameter& parj=sigi.At(j);
      *pOutFile << "+" << Parameter::AStr(parj.mAttr) << "+ ";
      TypeHtml(pOutFile,parj.mTDName);
      *pOutFile << " <i>" << parj.mName << "</i>";
    }
    *pOutFile << ")<br>" << std::endl;
  }
  *pOutFile << "</p>" << std::endl;


  // close signature box
  *pOutFile << std::endl << "</div>" << std::endl;
}

// ******************************************************************
// short doc
// ******************************************************************

void ShortdocHtml(std::ostream* pOutFile, std::string function) {

  // bail out on non existence
  if(!FunctionRegistry::G()->Exists(function)) return;

  // function definition
  const FunctionDefinition& fdef=FunctionRegistry::G()->Definition(function);
  
  // stream doc
  *pOutFile << "<div class=\"registry_signature\">" << std::endl;
  *pOutFile << "<p>" << fdef.TextDoc() << "<p>" << std::endl;
  *pOutFile << "</div>" << std::endl;

  // stream signature
  SignatureHtml(pOutFile,function);
}

// ******************************************************************
// command line ui
// ******************************************************************


int main(int argc, char *argv[]) {

  // min args
  if(argc < 3) Usage();

  // load registry 
  LoadRegistry(argv[1]);

  // output file
  std::ostream* hout= &std::cout;
  std::ofstream fout;
  if(std::string(argv[argc-1]) != "-") {
    fout.open(argv[argc-1], std::ios::out);
    hout = &fout;
  }
  
  // generate global index
  if(std::string(argv[2])=="-index") {
    // bail out
    if(argc != 4) Usage();
    // run and done
    IndexHtml(hout);
    return 0;
  } 

  // generate plugin index
  if(std::string(argv[2])=="-plugin") {
    // bail out
    if(argc != 5) Usage();
    // have lower case plugin pattern
    std::string pattern=std::string(argv[3])+" plugin";
    std::transform(pattern.begin(), pattern.end(), pattern.begin(), tolower);
    // run and done
    IndexHtml(hout,pattern);
    return 0;
  } 

  // generate linked type
  if(std::string(argv[2])=="-function") {
    // bail out
    if(argc != 5) Usage();
    // run and done
    FunctionHtml(hout,argv[3]);
    return 0;
  } 

  // generate linked function
  if(std::string(argv[2])=="-type") {
    // bail out
    if(argc != 5) Usage();
    // run and done
    TypeHtml(hout,argv[3]);
    return 0;
  } 

  // generate signature
  if(std::string(argv[2])=="-signature") {
    // bail out
    if(argc != 5) Usage();
    // run and done
    SignatureHtml(hout,argv[3]);
    return 0;
  } 

  // generate short doc
  if(std::string(argv[2])=="-shortdoc") {
    // bail out
    if(argc != 5) Usage();
    // run and done
    ShortdocHtml(hout,argv[3]);
    return 0;
  } 

  // error
  Usage();
  return(1);
}
