/** @file rtifncts.h Runtime interface, operations on faudes types */

    /* FAU Discrete Event Systems Library (libfaudes)

    Copyright (C) 2009 Ruediger Berndt
    Copyright (C) 2009 Thomas Moor

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef FAUDES_RTIFNCTS_H
#define FAUDES_RTIFNCTS_H

// todo: fix/hide copy constructor

#include "rtitypes.h"

namespace faudes{

/**
 * Structure to model a parameter type within the Signature of a 
 * Function.
 * A Parameter is made of a descriptive name, a faudes type and
 * an io-attribute. The latter specifies whether the parameter 
 * is an input or output of the corresponding function.
 */

class Parameter {

 public:

  /**
   * A function parameter has has one out of four so called io-attrributes;
   */
  enum ParamAttr{
    In,    //// Input parameter, aka argument, remains constant while execution
    Out,   //// Output parameter, aka result, is generated during function execution
    InOut, //// InOut parameter, is interpreted and possibly altered during execution
    UnDef  //// UnDef parameter indicates unconfigured signature
  };

  /** Constructor, default */
  Parameter(void) : mName(""), mTDName(""), mAttr(Parameter::UnDef) {};

  /** Constructor, by member values */
  Parameter(const std::string& rName, const std::string& rTypeName, ParamAttr attr) : 
     mName(rName), mTDName(rTypeName), mAttr(attr) {};

  /** Desctructor */
  ~Parameter(void){};

  /**
   * Convenience method to produce a textual representation of an io attribute.
   *
   * @param attr
   * 	Enum value denoting the attribute.
   * @return
   * 	Parameter IO attribute as std::string
   */
  static std::string AStr(Parameter::ParamAttr attr){
    switch(attr){
    case Parameter::In: return "In";
    case Parameter::Out: return "Out";
    case Parameter::InOut: return "InOut";
    default: break;
    }
    return "UnDef";
  };

  /**
   * Convenience method to produce a textual representation of a parameter.
   *
   */
  std::string Str(void) const {
    std::string res = "+" + AStr(mAttr)+ "+ " + mTDName + " " + mName;
    return res;
  };

  /**
   * Set to "undefined"
   */
  void Clear(){
    mName = "";
    mTDName = "";
    mAttr = Parameter::UnDef;
  };

  /** Name */
  std::string mName;

  /** Faudes type */
  std::string mTDName;

  /** IO-Attribute */
  ParamAttr mAttr;

 protected:

}; // Parameter




/**
 * Signature of a Function.
 * 
 * A Signature describes the faudes types of the positional
 * parameters. Tecnically, a Signature is a vector of Parameters.
 * Each Function may execute serveral variants indicated by setting
 * a particular Signature. A list of valid Signatures is maintained in 
 * the coresponding FunctionDefinition.
 *
 * Core members are
 * - mName: string to identify ths signature
 * - mParameters: vector of Paramters. 
 *
 * The Signature is formally derived from Type to inherit the std token io
 * interface. It is not meant to be registered as a faudes type.
 * The token io format is demonstrated by the following example:
 *
 * @code
 *	<Signature>
 *	"Sum of two integers"
 *	<Parameters>
 *		"arg1"        "Integer"     +InOut+
 *		"arg2"        "Integer"     +In+
 *		"result"      "String"      +Out+
 *	</Parameters>
 *	</Signature>
 * @endcode
 *
 * Technical note: the variable parameter feature offered by FunctionDefinition is a purely
 * cosmetic hack implemented in FunctionDefinition:MergeDocumentation.
 */

class Signature : public Type {

public:

  /** Constructor */
  Signature(void){};

  /** Destructor */
  ~Signature(void){};

  /**
   * Return signature name.
   *
   * @return
   * 	Name
   */
  std::string Name(void) const;

  /**
   * Set signature name.
   *
   * @param
   * 	rName
   */
  void Name(const std::string& rName);

  /**
   * Clear signature
   */
  void Clear(void);

  /**
   * Return number of parameters.
   *
   * @return
   * 	int
   */
  int Size(void) const;

  /**
   * Get parameter type by position.
   *
   * @param n
   * 	Position of patameter.
   *
   * @exception Exception
   * 	- Index out of range
   */
  const Parameter& At(int n) const;

  /**
   * Append positional parameter.
   *
   * @param rParam
   * 	Parameter to append
   */
  void Append(const Parameter& rParam);


 protected:

  /**
   * Read signature from from TokenReader.
   *
   * The section is hardcoded to "Signature", context is ignored.
   *
   * @param rTr
   *   TokenReader to read from
   * @param rLabel
   *   Section to read
   * @param pContext
   *   Read context to provide contextual information (ignored)
   *
   * @exception Exception
   *   - IO error (id 1)
   *   - Token mismatch
   */
  virtual void DoRead(TokenReader& rTr,  const std::string& rLabel = "", const Type* pContext=0);
 
  /**
   * Write configuration data of this object to TokenWriter.
   *
   * The section is hardcoded to "Signature", context is ignored.
   *
   * @param rTw
   *   Reference to TokenWriter
   * @param rLabel
   *   Label of section to write
   * @param pContext
   *   Write context to provide contextual information
   *
   * @exception Exception 
   *   - IO errors (id 2)
   */
  virtual void DoWrite(TokenWriter& rTw, const std::string& rLabel="",const Type* pContext=0) const;

  /** Variable to store name */
  std::string mName;

  /** Vector of Parameter-objects */
  std::vector<Parameter> mParameters;

}; // Signature



// forward 
class Function;


/**
 * A FunctionDefinition defines the interface to a faudes-function.
 * The latter consists of a descriptive name to identify the function and 
 * a list of Signatures the function can operate on. 
 * Technically, the class FunctionDefinition is derived from Documentation and 
 * thereby inherits members for additional
 * documentation-data.
 *
 * Similar to Type and TypeDefinition, a FunctionDefinition uses a prototype object to
 * provide the method NewFunction() which instantiates the corresponding Function object.
 *
 * FunctionDefinition inherits the token io interface from Type, however, the class is
 * not intended to register as a faudes type. The token io format is demonstrated by the
 * following example:
 *
 * @code
 * <FunctionDefinition>
 * "IntegerSum"
 *
 * <TextDoc> "Returns the sum of integer arguments." </TextDoc>
 * <HtmlDoc> "integersum.html" </HtmlDoc>
 * <Keywords> "integer" "elemetary types" </Keywords>
 *
 * <VariantSignatures>
 * <Signature>
 * "Two arguments"
 * "Arg1"  "Integer"   +In+
 * "Arg2"  "Integer"   +In+
 * "Res"   "Integer"   +Out+
 * </Signature>
 *
 * <Signature>          
 * "Three arguments"
 * "Arg1"  "Integer"   +In+
 * "Arg2"  "Integer"   +In+
 * "Arg3"  "Integer"   +In+
 * "Res"   "Integer"   +Out+
 * </Signature>
 * </VariantSignatures>
 *
 * </FunctionDefinition>
 * @endcode
 *
 * For functions that allow for a variable number of arguments of thesame type, there
 * is a purely cosmetic hack: when in a function definition a signature has a double hash "##"
 * in its name, any parameters with a "##" can apprear repeatedly. This short cut is expanded
 * by the token input function MergeDocumentation, that actually inserts the respective signatures. 
 * The compile time option FD_RTIMAXPARAMS defines how many copies of the parameter can appear.
 * Obviously, this only makes sense if the corresponding Function object supports every
 * automatically generated signature. 
 * Example from IntegerSum
 *
 * @code
 * <Signature>
 * "Ires=Sum_(i=1)^(##) Ii"
 * "Arg##" "Integer"   +In+
 * "Res"   "Integer"   +Out+
 * </Signature>
 * @endcode
 *
 *
 * @ingroup RunTimeInterface
 */

class FunctionDefinition : public Documentation {

public:

  /** 
   * Constructor 
   *
   * The default constructor instantiates an invalid function definition
   * without prototype. To construct
   * a valid type definition, use the static Constructor() template
   * function.
   */
  FunctionDefinition(const std::string& name="") : Documentation(), mpFunction(NULL) {Name(name);};

  /**
   * Destructor
   */
  virtual ~FunctionDefinition(){ Prototype(NULL);};

  /**
   * Construct empty FunctionDefinition object.
   * The given template parameter denotes a Function class.
   * Member variable (mpFunction) is set to a new instance of that class
   * whereas the name is set as specified. No further documentation
   * or signatures are recorded.
   *
   * @tparam T
   *   Actual function class, derived from Function
   * @param rFunctName
   *   Name to identify this faudes-function
   * @return
   *   Newly constructed function definition.
   */
  template<class T>
  static FunctionDefinition* Constructor(const std::string& rFunctName="");

  /**
   * Construct FunctionDefinition object and get name and docu from file.
   *
   * The member variable mpFunction is set to a new instance of class T.
   * which must be derived from Function. The function name, any documentation
   * as well as supported signatures are read from the specified file.
   *
   * @tparam T
   *   Actual function class, derived from Function
   * @param rFileName
   *   File to read documentation and signatures from.
   * @return
   *   Newly constructed function definition.
   */
  template<class T>
  static FunctionDefinition* FromFile(const std::string& rFileName);

  /**
   * Clear documentation-data and signature (keep prototype)
   */
  void Clear(void);

  /**
   * Merge documentation and signatures from token stream.
   * This member reads the body of the FunctionDefinition token format
   * and sets the function name, any further documentation and the
   * supported signatures accordingly. An exception is
   * thrown if the current function name differs from the one in the documentation.
   *
   *
   *
   * @param rTr
   * 	TokenReader to read from.
   *
   * @exception Exception
   *  - Type mismatch (id )
   *  - Token mismatch (id 50, 51, 52)
   *  - IO Error (id 1)
   */
  void MergeDocumentationBody(TokenReader& rTr);


  /**
   * Return pointer to function object prototype 
   *
   * Note: this method is meant for inspection only, control over
   * the prototype remains with the FunctionDefinition. Use
   * NewFunction() to instantiate a new function object.
   *
   * @return
   * 	Reference to prototype function.
   */
  const Function* Prototype(void) const;

  /**
   * Construct function on heap.
   * Return pointer to new instance of assigned Function class.
   *
   * Note: If no prototype is installed, NULL is returned.
   *
   * @return
   * 	Pointer to new Function instance.
   */
  Function* NewFunction() const;


  /**
   * Return number of supported Signature instances.
   *
   * @return
   * 	Size of signature vector.
   */
  int VariantsSize(void) const;

  /**
   * Test existence of variant by its name.
   *
   * @return
   * 	True if variant exists
   */
  bool ExistsVariant(const std::string& varname) const;

  /**
   * Return reference to Signature by name.
   *
   * @param rName
   * 	Name of signature to search.
   * @return
   * 	Reference to Signature 
   * @exception Exception
   * 	- No such signature (id 47)
   */
  const Signature& Variant(const std::string& rName) const;

  /**
   * Return reference to Signature by index.
   *
   * @param n
   * 	Index to look up
   * @return
   * 	Reference to Signature 
   * @exception Exception
   * 	- Index out of range (id 47)
   */
  const Signature& Variant(int n) const;

protected:

  /**
   * Read configuration data of this object from TokenReader.
   *
   * The section is hardcode to "FunctionDefinition", context ignored.
   *
   * @param rTr
   *   TokenReader to read from
   * @param rLabel
   *   Section to read
   * @param pContext
   *   Read context to provide contextual information (ignored)
   *
   * @exception Exception
   *   - IO error (id 1)
   */
  virtual void DoRead(TokenReader& rTr,  const std::string& rLabel = "", const Type* pContext=0);
 
  /**
   * Write configuration data of this object to TokenWriter.
   *
   * The section is hardcode to "FunctionDefinition", context ignored.
   *
   * @param rTw
   *   Reference to TokenWriter
   * @param rLabel
   *   Label of section to write
   * @param pContext
   *   Write context to provide contextual information
   *
   * @exception Exception 
   *   - IO errors (id 2)
   */
  virtual void DoWrite(TokenWriter& rTw, const std::string& rLabel="",const Type* pContext=0) const;

  /**
   * Add Signature to function definition.
   *
   * @param pVar
   * 	Signature to insert
   * @exception Exception 
   *   - Signature with same name exists (id 47)
   */
  void AppendVariant(const Signature& pVar);

  /** Disable Copy Constructor */
  FunctionDefinition(const FunctionDefinition& rOther) : Documentation(rOther) {};


  /**
   * Assign prototype object
   *
   * @param pFunc
   * 	Function instance
   *
   */
  void Prototype(Function* pFunc);

  /** Prototype instance */
  Function* mpFunction;

  /** Vector containing all supported Signatures */
  std::vector<faudes::Signature> mVariants;

  /** Variant name to index map */
  std::map<std::string,int> mVariantIndexMap;

}; // FunctionDefinition



/**
 * A faudes-function hosts parameter values of some faudes type and provides 
 * a method to perform an operation on the specified paramters, e.g. the 
 * parallel composition on two generators. The class Function is the base
 * for all faudes-functions, which essenitally differ in the operation method.
 *
 * The base class provides an application interface to
 * - select a variant signature, see Variant(); 
 * - set and get parameter values of any faudes type, see ParamValue();
 * - type check the provided parameter values, see TypeCheck(); 
 * - perform the operation, see Execute().
 *
 * Derivates of the base class, that acually implement an operation,
 * are required to cast the parameter values to the corresponding c types and then
 * call the respective c function. To derive a class from Function, one will
 * reimplment three virtual functions:
 * - DoTypeCheck() to perform the cast
 * - DoExecute() to call the c function.
 * - the constructor New() to replicate the function object.
 * See IntegerSum for a simple example of a derived faudes-function.
 * 
 *
 * Technichal note: In contrast to faudes objects (derivates from Type), a faudes-function
 * must be provided a reference to the corresponding FunctionDefinition. This is because
 * a function should be aware of its signature(s). You must not delete 
 * the FunctionDefinition object during the lifetime of a faudes-function.
 * Future implementations of Function may host a local copy of the
 * corresponding FunctionDefinition ... its a "memory efficiency versus segfaults"  trade off.
 *
 *
 * @ingroup RunTimeInterface
 */  

class Function : public Type {

 public:
  /** 
   * Constructor 
   * For the function to be operational, a valid reference to the corresponding 
   * FunctionDefinition is required. The only exception is the prototype function
   * object used in the FunctionDefinition itself. 
   */
  Function(const FunctionDefinition* fdef);

  /** Destructor */
  ~Function(){}

  /**
   * Construct on heap.
   * Create a new instance of this function class and return pointer.
   * The new instance will use the same function definition as this instance.
   *
   * @return
   * 	Pointer to faudes::Function instance.
   *
   */
  virtual Function* New() const = 0;


  /**
   * Set function definition.
   * Normally, functions are provided with a function definition on construction.
   * The only exception are prototype objects used in function definitions themselfs 
   * and in the function registry.
   *
   * @param fdef
   * 	Function definition to set.
   *
   */
  void Definition(const FunctionDefinition* fdef);


  /**
   * Get function definition.
   *
   * @return 
   * 	Function definition used by this function.
   *
   */
  const FunctionDefinition* Definition(void) const;


  /**
   * Return number of variants.
   *
   * @return
   * 	Size of vector.
   */
  int VariantsSize(void) const;

  /**
   * Set signature from function definition.
   *
   * The index n refers this function's FunctionDefinition.
   * An exception is thrown if no such FunctionDefinition is set, as it
   * is the case for prototype instances.
   *
   * @param n
   * 	Variant index 
   *
   * @exception Exception
   * 	- No function definition available (id 47)
   */
  void Variant(int n);


  /**
   * Set signature from function definition.
   *
   * The name refers to this function's FunctionDefinition.
   * An exception is thrown if no such FunctionDefinition is set, as it
   * is the case for prototype instances.
   *
   * @param rVariantName
   * 	Variant name
   *
   * @exception Exception
   * 	- No function definition available (id 47)
   */
  void Variant(const std::string& rVariantName);


  /**
   * Return pointer to assigned faudes::Signature.
   *
   * @return
   * 	Pointer to faudes::Signature.
   */
  const Signature* Variant() const;


  /**
   * Return number of parameters with current signature.
   *
   * @return
   * 	Size of vector.
   */
  int ParamsSize(void) const;

  /**
   * Set parameter at certain position.
   *
   * Sets the internal refernce of the parameter value
   * at the spcefied index. The ownership of the value remains with
   * the caller, ie it will not be deleted in the destructor of the 
   * function object.
   * You may set the parameter value to any faudes type
   * (classes derived from Type). A type check will be performed before the function
   * is executed.
   *
   * @param n
   * 	Position index of parameter 
   * @param param
   * 	Pointer to faudes object.
   *
   * @exception Exception
   * 	- Index out of range (id 47)
   * 	- No variant set (id 47)
   */
  void ParamValue(int n, Type* param);

  /**
   * Get parameter value.
   * Returns a reference to the parameter value at the specified position.
   *
   * @param n
   * 	Position index of parameter.
   *
   * @exception Exception
   * 	- Index out of range (id 47)
   */
  Type* ParamValue(int n) const;

  /**
   * Construct parameter values.
   *
   * This is a convenience method to allocate faudes objects for 
   * parameter values with the type specified by the current signature.
   * Note that the function does not take ownetship of the parameter values and
   * it is the callers responsibility to delete them when no longer required.
   *
   */
  void AllocateValues(void);
  void AllocateValue(int i);

  /**
   * Destruct parameter values.
   *
   * This is a convenience method to delete the assigned paramer values.
   *
   */
  void FreeValues(void);

  /**
   * Perform a type check on the current parameter values.
   * 
   * The type check is based on c type cast and should accept any types
   * derived from those given in the signature.
   *
   * @exception Exception
   *  - No variant specified (id 48)
   *  - Number of parameters does not match signature (id 48)
   *  - Type mismatch (id 48)
   *
   */
  void TypeCheck(void);

  /** Depreciated / untested */
  bool StrictTypeCheck(void) const;

  /**
   * Perform operation.
   *
   * Runs a type cheack and then the actual function.
   *
   * @exception Exception
   *  - No variant specified (id 48)
   *  - Parameter position out of range (id 48)
   *  - Type mismatch (id 48)
   *  - Any other exception thrown ba the actual function (id xxx)
   */
  void Execute(void);


 protected:

  /**
   * Set signature as specified.
   *
   * This method allows to set any signature. Since the function
   * will only support particular signatures, you should use
   * Variant(int n) whenever possible.
   *
   * @param pVar
   * 	Pointer to faudes::Signature.
   */
  void Variant(const Signature* pVar);

  /**
   * Helper: generate typed reference for parameter.
   *
   * @param n
   *   Parameter position to cast
   * @param rTypedRef
   *   Typed reference to generyte
   * @tparam T
   *   c type to use for cast
   * @exception Exception
   *  - No variant specified (id 48)
   *  - Parameter position out of range (id 48)
   *  - Type mismatch (id 48)
   */  
   template<class T>
   void DoTypeCast(int n, T*& rTypedRef) {
     if(!pVariant) {
       std::stringstream err;
       err << "No variant specified";
       throw Exception("Function::TypeCheck()", err.str(), 48);
     }
    if(n<0 || n >= ParamsSize()) {
       std::stringstream err;
       err << "Parameter position out of range";
       throw Exception("Function::TypeCheck()", err.str(), 48);
    }
     rTypedRef=dynamic_cast<T*>(ParamValue(n));
     if(rTypedRef==NULL) {
        std::stringstream err;
        err << "Cannot cast parameter \"" << pVariant->At(n).mName << "\" to type \"" << pVariant->At(n).mTDName << "\"";
        throw Exception("Function::TypeCheck()", err.str(), 48);
     }
   }

  /**
   * Method to compare types of arguments with arguments
   * of assigned faudes::Signature (i.e. their TypeDefinition label).
   *
   * NOTE:
   * Method is called by Function::Execute().
   *
   * @exception Exception
   * 	- Signature undefined
   * 	- Parameter number unequal
   * 	- Dynamic cast error
   */
  virtual void DoTypeCheck() = 0;

  /**
   * Executes code of reimplemented method of child class(es).
   */
  virtual void DoExecute() = 0;

  /**
   * Write function-data (typeid-name of arguments) to TokenWriter.
   *
   * @param rTw
   * 	Reference to Tokenwriter.
   * @param rLabel
   * 	Label of section to write.
   * @param pContext
   * 	Context pointer, ignored
   *
   * @exception Exception
   * 	- IO Error
   */
  void DoWrite(TokenWriter& rTw, const std::string& rLabel = "", const Type* pContext=0) const;

  /** corresponding function definition */
  const FunctionDefinition* pFuncDef;

  /** current variant aka signature */
  const Signature* pVariant;

  /** Vector of arguments. */
  std::vector<Type*> mParameterValues;

}; // Function



/**********************************************************************************************
***********************************************************************************************
***********************************************************************************************

Implemention of template members functions

***********************************************************************************************
***********************************************************************************************
**********************************************************************************************/


// typedefinition constructor function
template<class T>
FunctionDefinition* FunctionDefinition::Constructor(const std::string& rFunctionName){
  FD_DRTI("FunctionDefinition::Construct<" << typeid(T).name() << ">()");
  // construct void definition
  FunctionDefinition* td = new FunctionDefinition();
  // set its prototype object (note the 0 param in the new constructor)
  td->Prototype(new T(0));
  // set minmum values ie the function name
  std::string name=rFunctionName;
  if(name=="") name=typeid(T).name();
  td->Name(name);
  FD_DRTI("FunctionDefinition::Constructor<" << typeid(T).name() << ">(): done");
  return(td);
}


// function definition constructor function
template<class T>
FunctionDefinition* FunctionDefinition::FromFile(const std::string& rFileName){
  FD_DRTI("FunctionDefinition::FromFile<" << typeid(T).name() << ">()");
  // construct with fallback name
  FunctionDefinition* td = Constructor<T>();
  // read docu, incl actual name
  td->Read(rFileName);
  // done
  FD_DRTI("FunctionDefinition::FromFile<" << typeid(T).name() << ">(): done");
  return(td);
}


} // namespace


#endif /* FAUDES_FUNCTIONS_H */
