/** @file rtiregistry.cpp Runtime interface, registry for faudes types and functions */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2009 Ruediger Berndt
Copyright (C) 2009 Thomas Moor

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */



#include "rtiregistry.h"
#include "corefaudes.h"
#include "allplugins.h"

// auto install types
#ifndef FAUDES_MUTE_RTIAUTOLOAD
#include "rtiautoload.h"
#include "rtiautoload.cpp"
#endif

namespace faudes{



/*
********************************************************************
********************************************************************
********************************************************************

Implemantation of faudes TypeRegistry

********************************************************************
********************************************************************
********************************************************************
*/

// static members: ref to the only one instnace
TypeRegistry* TypeRegistry::mpInstance = 0;

// static member: access to signleton
TypeRegistry* TypeRegistry::G(){
  // lazy initialization
  if(!mpInstance){
    FD_DRTI("TypeRegistry::G(): Constrtuct singleton");
    mpInstance = new TypeRegistry();
  }
  return(mpInstance);
}

// clear all
void TypeRegistry::Clear(){
  // delete all typedefs contained in map
  std::map<std::string, TypeDefinition*>::iterator mit;
  for(mit = mNameToTypeDef.begin(); mit != mNameToTypeDef.end(); mit++){
    if(mit->second != NULL){
      delete mit->second;
      mit->second = NULL;
    }
  }
  // delete maps
  mNameToTypeDef.clear();
  mIdToTypeDef.clear();
}

// query number of entries
int TypeRegistry::Size() const{
  return(mNameToTypeDef.size());
}

// read access on type map
TypeRegistry::Iterator  TypeRegistry::Begin(void) const{
  return(mNameToTypeDef.begin());
}

// read access on type map
TypeRegistry::Iterator TypeRegistry::End(void) const{
  return(mNameToTypeDef.end());
}

// insert new entry
void TypeRegistry::Insert(TypeDefinition* pTypeDef){
  FD_DRTI("TypeRegistration::Insert(): definition for " << pTypeDef->Name());
  // test existence
  if(Exists(pTypeDef->Name())){
    std::stringstream err;
    err << "Cannot overwrite existing entry with type " << pTypeDef->Name() << std::endl;
    throw Exception("TypeRegistry::Insert()", err.str(), 46);
  }
  // record in main map
  mNameToTypeDef[pTypeDef->Name()]=pTypeDef;
  // record in id map
  if(pTypeDef->Prototype())
    mIdToTypeDef[typeid(*(pTypeDef->Prototype())).name()]=pTypeDef;
}


// scan token stream for type definitions
void TypeRegistry::MergeDocumentation(TokenReader& rTr) {
  FD_DRTI("TypeRegistration::MergeMergeDocumentation()");
  // scan file
  Token token;
  while(rTr.Peek(token)) {
    // continue on other tokens
    if(token.Type()!=Token::Begin) { rTr.Get(token); continue; }
    if(token.StringValue()!="TypeDefinition") {  rTr.Get(token); continue; }
    // found type def in file
    rTr.ReadBegin("TypeDefinition");
    rTr.Peek(token);
    std::string ftype=token.StringValue();
    // locate type def in map
    Iterator tit = mNameToTypeDef.find(ftype);
    // case 1: type exists: add documentaion
    if(tit!=mNameToTypeDef.end()) {
      tit->second->MergeDocumentationBody(rTr);
      rTr.ReadEnd("TypeDefinition");
      continue;
    }
    // case 2: type does not exist: insert fake entry
    TypeDefinition* tdef = new TypeDefinition(ftype);
    tdef->MergeDocumentationBody(rTr);
    Insert(tdef);
    rTr.ReadEnd("TypeDefinition");  
  }
}

// scan file for type definitions
void TypeRegistry::MergeDocumentation(const std::string& rFileName) {
  TokenReader tr(rFileName);
  MergeDocumentation(tr);
}


// construct faudes object by typename
Type* TypeRegistry::NewObject(const std::string& rName) const{
  FD_DRTI("TypeRegistry::NewObject(\"" << rName << "\")");
  Iterator mit=mNameToTypeDef.find(rName);
  if(mit == End()) {
    std::stringstream err;
    err << "Unknown Type " << rName << std::endl;
    throw Exception("TypeRegistry::NewObject()", err.str(), 47);
  }
  Type* res=mit->second->NewObject();
  if(!res) {
    std::stringstream err;
    err << "Failed to instatiate new " << rName << std::endl;
    throw Exception("TypeRegistry::NewObject()", err.str(), 47);
  }
  return res;
}

// construct faudes object by typed reference
Type* TypeRegistry::NewObject(const Type& rType) const{
  FD_DRTI("TypeRegistry::NewObject(prototype): typeid " << typeid(rType).name());
  Iterator mit;
  mit=mIdToTypeDef.find(typeid(rType).name());
  if(mit == mIdToTypeDef.end()) {
    std::stringstream err;
    err << "Unknown type by reference" << std::endl;
    throw Exception("TypeRegistry::NewObject()", err.str(), 47);
  }
  return(rType.New());
}

// access type definition by type name
const TypeDefinition& TypeRegistry::Definition(const std::string& rName) const{
  FD_DRTI("TypeRegistry::Definition( " << rName << " )");
  Iterator mit=mNameToTypeDef.find(rName);
  if(mit == End()) {
    std::stringstream err;
    err << "Type not found: " << rName;
    throw Exception("TypeRegistry::Definition()", err.str(), 46);
  }
  return(*(mit->second));
}

// access type definition by typed reference
// todo: test
const TypeDefinition& TypeRegistry::Definition(const Type& rType) const{
  FD_DRTI("TypeRegistry::Definition(): typeid " << typeid(rType).name());
  Iterator mit;
  mit=mIdToTypeDef.find(typeid(rType).name());
  if(mit!=mIdToTypeDef.end()) return *(mit->second);
  /*
  Iterator mit;
  for(mit = Begin(); mit != End(); mit++){
    if(typeid(rType) == typeid(*(mit->second->Prototype())))
      return *(mit->second);
  }
  */
  std::stringstream err;
  err << "Type not found: " << typeid(rType).name();
  throw Exception("TypeRegistry::Definition()", err.str(), 46);
}

// test existsnce by type name
bool TypeRegistry::Exists(const std::string& rName) const{
  return mNameToTypeDef.find(rName) != End();
}

// test existsnce by type name
bool TypeRegistry::Exists(const Type& rType) const{
  return mIdToTypeDef.find(typeid(rType).name()) != mIdToTypeDef.end();
}


// token write (informative/debugging)
void TypeRegistry::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  FD_DRTI("TypeRegistry::DoWrite(): file " << rTw.FileName());
  // ignore label and context
  (void) rLabel;
  (void) pContext;
  // doit
  rTw.WriteBegin("TypeRegistry");
  Iterator tit;
  for(tit=Begin();tit!=End();tit++) {
     // write type definition
     tit->second->Write(rTw);
  }
  rTw.WriteEnd("TypeRegistry");
}


/*
********************************************************************
********************************************************************
********************************************************************

Implemantation of faudes FunctionRegistry

********************************************************************
********************************************************************
********************************************************************
*/

// static members: ref to the only one instnace
FunctionRegistry* FunctionRegistry::mpInstance = 0;

// static member: access to signleton
FunctionRegistry* FunctionRegistry::G(){
  // lazy initialization
  if(!mpInstance){
    FD_DRTI("FunctionRegistry::G(): Construct singleton");
    mpInstance = new FunctionRegistry();
  }
  return(mpInstance);
}

// clear all
void FunctionRegistry::Clear(){
  // delete all functiondefs contained in map
  std::map<std::string, FunctionDefinition*>::iterator mit;
  for(mit = mNameToFunctionDef.begin(); mit != mNameToFunctionDef.end(); mit++){
    if(mit->second != NULL) {
      FD_DRTI("FunctionRegistry::Clear: removing " << mit->second->Name());
      delete mit->second;
      mit->second = NULL;
    }
  }
  // delete maps
  mNameToFunctionDef.clear();
  mIdToFunctionDef.clear();
}

// query number of entries
int FunctionRegistry::Size() const{
  return(mNameToFunctionDef.size());
}

// read access on function map
FunctionRegistry::Iterator  FunctionRegistry::Begin(void) const{
  return(mNameToFunctionDef.begin());
}

// read access on function map
FunctionRegistry::Iterator FunctionRegistry::End(void) const{
  return(mNameToFunctionDef.end());
}

// insert new entry
void FunctionRegistry::Insert(FunctionDefinition* pFunctionDef){
  FD_DRTI("FunctionRegistration::Insert(): definition for " << pFunctionDef->Name());
  // test existence
  if(Exists(pFunctionDef->Name())){
    std::stringstream err;
    err << "Cannot overwrite existing entry with function " << pFunctionDef->Name() << std::endl;
    throw Exception("FunctionRegistry::Insert()", err.str(), 46);
  }
  // record in map
  mNameToFunctionDef[pFunctionDef->Name()]=pFunctionDef;
  // record in id map
  if(pFunctionDef->Prototype())
    mIdToFunctionDef[typeid(*(pFunctionDef->Prototype())).name()]=pFunctionDef;
}


// scan token stream for function definitions
void FunctionRegistry::MergeDocumentation(TokenReader& rTr) {
  FD_DRTI("FunctionRegistration::MergeDocumentation()");
  // scan file
  Token token;
  while(rTr.Peek(token)) {
    // continue on other tokens
    if(token.Type()!=Token::Begin) { rTr.Get(token); continue; }
    if(token.StringValue()!="FunctionDefinition") {  rTr.Get(token); continue; }
    // found function def in file
    rTr.ReadBegin("FunctionDefinition");
    rTr.Peek(token);
    std::string ffunction=token.StringValue();
    // locate function def in map
    Iterator fit = mNameToFunctionDef.find(ffunction);
    // case 1: function exists: add documentaion
    if(fit!=mNameToFunctionDef.end()) {
      fit->second->MergeDocumentationBody(rTr);
      rTr.ReadEnd("FunctionDefinition");
      continue;
    }
    // case 2: function does not exist: insert fake entry
    FunctionDefinition* fdef = new FunctionDefinition(ffunction);
    fdef->MergeDocumentationBody(rTr);
    Insert(fdef);
    rTr.ReadEnd("FunctionDefinition");  
  }
}

// scan file for function definitions
void FunctionRegistry::MergeDocumentation(const std::string& rFileName) {
  TokenReader tr(rFileName);
  MergeDocumentation(tr);
}


// construct faudes object by functionname
Function* FunctionRegistry::NewFunction(const std::string& rName) const{
  FD_DRTI("FunctionRegistry::NewFunction(\"" << rName << "\")");
  Iterator mit=mNameToFunctionDef.find(rName);
  if(mit == End()) {
    std::stringstream err;
    err << "Unknown function " << rName << std::endl;
    throw Exception("FunctionRegistry::NewFunction()", err.str(), 47);
  }
  Function* res=mit->second->NewFunction();
  if(!res) {
    std::stringstream err;
    err << "Failed to instatiate new " << rName << std::endl;
    throw Exception("FunctionRegistry::NewFunction()", err.str(), 47);
  }
  return res;
}

// construct faudes object by function reference
Function* FunctionRegistry::NewFunction(const Function& rFunction) const{
  FD_DRTI("FunctionRegistry::NewFunction(prototype): typeid " << typeid(rFunction).name());
  Iterator mit;
  mit=mIdToFunctionDef.find(typeid(rFunction).name());
  if(mit == mIdToFunctionDef.end()) {
    std::stringstream err;
    err << "Unknown function by reference" << std::endl;
    throw Exception("FunctionRegistry::NewFunction()", err.str(), 47);
  }
  return(rFunction.New());
}

// access function definition by function name
const FunctionDefinition& FunctionRegistry::Definition(const std::string& rName) const{
  FD_DRTI("FunctionRegistry::Definition( " << rName << " )");
  Iterator mit=mNameToFunctionDef.find(rName);
  if(mit == End()) {
    std::stringstream err;
    err << "Function not found: " << rName;
    throw Exception("FunctionRegistry::Definition()", err.str(), 46);
  }
  return(*(mit->second));
}

// access function definition by functiond reference
// todo: have function id map
const FunctionDefinition& FunctionRegistry::Definition(const Function& rFunction) const{
  FD_DRTI("FunctionRegistry::Definition(): typeid " << typeid(rFunction).name());
  Iterator mit;
  mit=mIdToFunctionDef.find(typeid(rFunction).name());
  if(mit!=mIdToFunctionDef.end()) return *(mit->second);
  /*
  Iterator mit;
  for(mit = Begin(); mit != End(); mit++){
    if(typeid(rFunction) == typeid(*(mit->second->Prototype())))
      return *(mit->second);
  }
  */
  std::stringstream err;
  err << "Function not found: " << typeid(rFunction).name();
  throw Exception("FunctionRegistry::Definition()", err.str(), 46);
}

// test existsnce by function name
bool FunctionRegistry::Exists(const std::string& rName) const{
  return mNameToFunctionDef.find(rName) != End();
}

// test existsnce by function name
bool FunctionRegistry::Exists(const Function& rFunction) const{
  return mIdToFunctionDef.find(typeid(rFunction).name()) != mIdToFunctionDef.end();
}


// token write (informative/debugging)
void FunctionRegistry::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  FD_DRTI("FunctionRegistry::DoWrite(): file " << rTw.FileName());
  // ignore label and context
  (void) rLabel;
  (void) pContext;
  // doit
  rTw.WriteBegin("FunctionRegistry");
  Iterator tit;
  for(tit=Begin();tit!=End();tit++) {
     // write function definition
     tit->second->Write(rTw);
  }
  rTw.WriteEnd("FunctionRegistry");
}


/*
********************************************************************
********************************************************************
********************************************************************

Implemantation of LoadRegistry

********************************************************************
********************************************************************
********************************************************************
*/

void LoadRegistry(const std::string& rPath) {
  FD_DRTI("LoadRegistry(" << rPath << ")");
  // default path
  std::string rtipath = rPath;
  if(rtipath=="") rtipath="../doc/registry/registry.rti";  // todo: plattform/configure
  // clear
  TypeRegistry::G()->Clear();
  FunctionRegistry::G()->Clear();

  // auto install types with rti file
#ifndef FAUDES_MUTE_RTIAUTOLOAD
  LoadAutoregisteredTypes();
#endif

  // install programmatic functions
  //FunctionRegistry::G()->Insert<RtiReachability>("Reachability Tests");
  //FunctionRegistry::G()->Insert<RtiIntegerSum>("IntegerSum");
  FunctionRegistry::G()->Insert<TNestedFunction<Integer,IntegerSum> >("IntegerSum");
  FunctionRegistry::G()->Insert<TNestedFunction<vGenerator,LanguageUnion> >("LanguageUnion");
  FunctionRegistry::G()->Insert<TNestedFunction<vGenerator,LanguageIntersection> >("LanguageIntersection");
  //  FunctionRegistry::G()->Insert<TSimpleFunction1<vGenerator,LanguageComplement> >("LanguageComplement");

  // allow build system load registry programmatic contributions defined in plugins
#ifdef FAUDES_PLUGINS_RTILOAD
  FAUDES_PLUGINS_RTILOAD;
#endif

  // auto install functions
#ifndef FAUDES_MUTE_RTIAUTOLOAD
  LoadAutoregisteredFunctions();
#endif

  // merge documentation
  TypeRegistry::G()->MergeDocumentation(rtipath);
  FunctionRegistry::G()->MergeDocumentation(rtipath);

  // test and report status
#ifdef FAUDES_CHECKED
#ifndef FAUDES_MUTE_RTIAUTOLOAD
  TypeRegistry::Iterator tit;
  for(tit=TypeRegistry::G()->Begin(); tit!=TypeRegistry::G()->End(); tit++) {
    // have prototype
    if(tit->second->Prototype()==NULL) 
      FD_WARN("TypeRegistry: " << tit->second->Name() << " has no prototype");
  }
  FunctionRegistry::Iterator fit;
  for(fit=FunctionRegistry::G()->Begin(); fit!=FunctionRegistry::G()->End(); fit++) {
    // have prototype
    if(fit->second->Prototype()==NULL) 
      FD_WARN("FunctionRegistry: " << fit->second->Name() << " has no prototype");
  }
#endif
#endif

  FD_DRTI("LoadRegistry(" << rPath << "): done");
}

void ClearRegistry(void) {
  FD_DRTI("ClearRegistry()");
  // clear
  FunctionRegistry::G()->Clear();
  TypeRegistry::G()->Clear();
}

Type* NewObject(const std::string& rTypeName) { return TypeRegistry::G()->NewObject(rTypeName);}
Function* NewFunction(const std::string& rFunctName) { return FunctionRegistry::G()->NewFunction(rFunctName);}


} // namespace
