/** @file vgenerator.cpp Class vGenerator */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Copyright (C) 2007  Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNES FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "vgenerator.h"


namespace faudes {

// msObjectCount (static) 
Idx vGenerator::msObjectCount = 0;

// msObjectCount (static) 
bool vGenerator::msStateNamesEnabledDefault = true;

// constructor
vGenerator::vGenerator(void) : 
  // my name
  mMyName("Generator"),
  // have std symboltables
  mpStateSymbolTable(&mStateSymbolTable),
  mpEventSymbolTable(GlobalEventSymbolTablep()),
  // other members
  mStateNamesEnabled(msStateNamesEnabledDefault),  
  // initialise heap members
  mpAlphabet(0),
  mpStates(0),
  mpTransRel(0),
  mpGlobalAttribute(0)
{
  FD_DG("vGenerator(" << this << ")::vGenerator()");
  // track generator objects
  msObjectCount++; 
  mId = msObjectCount;
  // allocate core members
  NewCore();
  // fix std names
  mInitStates.Name("InitStates");
  mMarkedStates.Name("MarkedStates");
}

// copy constructor
vGenerator::vGenerator(const vGenerator& rOtherGen) :
  // my name
  mMyName("Generator"),
  // have std symboltables
  mpStateSymbolTable(&mStateSymbolTable),
  mpEventSymbolTable(GlobalEventSymbolTablep()),
  // other members
  mStateNamesEnabled(msStateNamesEnabledDefault),  
  // initialise heap members
  mpAlphabet(0),
  mpStates(0),
  mpTransRel(0),
  mpGlobalAttribute(0)
{
  FD_DG("vGenerator(" << this << ")::vGenerator(" << &rOtherGen << ")");
  // track generator objects
  msObjectCount++; 
  mId = msObjectCount;
  // allocate core members
  NewCore();
  // perform copy
  rOtherGen.Copy(*this);
}

// construct from file
vGenerator::vGenerator(const std::string& rFileName) : 
  // my name
  mMyName("Generator"),
  // have std symboltables
  mpStateSymbolTable(&mStateSymbolTable),
  mpEventSymbolTable(GlobalEventSymbolTablep()),
  // other members
  mStateNamesEnabled(msStateNamesEnabledDefault),  
  // initialise heap members
  mpAlphabet(0),
  mpStates(0),
  mpTransRel(0),
  mpGlobalAttribute(0)
{
  FD_DG("vGenerator(" << this << ")::vGenerator(" << rFileName << ")");
  // track generator objects
  msObjectCount++; 
  mId = msObjectCount;
  // allocate core members
  NewCore();
  // fix std names
  mInitStates.Name("InitStates");
  mMarkedStates.Name("MarkedStates");
  // set defaults for file io
  mStateNamesEnabled=true;
  // do read
  Read(rFileName);
  // restore defaults
  mStateNamesEnabled=msStateNamesEnabledDefault;
}

// construct on heap
vGenerator* vGenerator::New(void) const {
  FD_DG("vGenerator(" << this << ")::New()");
  // allocate
  vGenerator* res = new vGenerator();
  // fix configuration
  res->EventSymbolTablep(mpEventSymbolTable);
  res->mStateNamesEnabled=mStateNamesEnabled;
  return res;
}

// destruct
vGenerator::~vGenerator(void) {
  // free my members
  DeleteCore();
}

// indicate new core
void vGenerator::UpdateCore(void) {
  // fix std names
  if(mpAlphabet) mpAlphabet->Name("Alphabet");
  if(mpStates) mpStates->Name("States");
  if(mpTransRel) mpTransRel->Name("TransRel");
}

// protected helper: allocate core on heap
void vGenerator::NewCore(void) {
  DeleteCore();
  // allocate
  mpAlphabet = new EventSet();
  mpStates = new StateSet();
  mpTransRel = new TransSet();
  mpGlobalAttribute = new AttributeVoid();
  // update callback
  UpdateCore();
}
  
// protected helper: free core from heap
void vGenerator::DeleteCore(void) {
  if(mpAlphabet) delete mpAlphabet; 
  if(mpStates) delete mpStates; 
  if(mpTransRel) delete mpTransRel; 
  if(mpGlobalAttribute) delete mpGlobalAttribute;   
  mpAlphabet=0;
  mpStates=0;
  mpTransRel=0;
  mpGlobalAttribute=0;
  // update callback
  UpdateCore();
}

// copy to other vGenerator (try to convert attributes)
void vGenerator::Copy(vGenerator& rGen) const {
  FD_DG("vGenerator(" << this << ")::Copy(" << &rGen << ")");
  // prepare result (call clear for virtual stuff)
  rGen.Clear();
  // have same event symboltable
  rGen.EventSymbolTablep(mpEventSymbolTable);
  // copy state symboltable
  rGen.StateSymbolTable(mStateSymbolTable);
  // set other members
  rGen.Name(Name());
  rGen.StateNamesEnabled(StateNamesEnabled());
  rGen.InjectInitStates(mInitStates);
  rGen.InjectMarkedStates(mMarkedStates);
  // core members, try attributes
  rGen.InjectStates(*mpStates);
  rGen.InjectAlphabet(*mpAlphabet);
  rGen.InjectTransRel(*mpTransRel);
  rGen.GlobalAttributeTry(*mpGlobalAttribute);
#ifdef FAUDES_CHECKED
  if(!rGen.Valid()) {
    FD_DG("vGenerator()::Copy(): invalid generator");
    rGen.DWrite(); 
    abort();
  }
#endif
} 


// copy to other vGenerator
void vGenerator::CopyWithoutAttributes(vGenerator& rGen) const {
  FD_DG("vGenerator(" << this << ")::CopyWithoutAttributes(" << &rGen << ")");
  // prepare result (call clear for virtual stuff)
  rGen.Clear();
  // have same event symboltable
  rGen.EventSymbolTablep(mpEventSymbolTable);
  // copy state symboltable
  rGen.StateSymbolTable(mStateSymbolTable);
  // set other members
  rGen.Name(Name());
  rGen.StateNamesEnabled(StateNamesEnabled());
  rGen.InjectInitStates(mInitStates);
  rGen.InjectMarkedStates(mMarkedStates);
  // copy core members, discard attributes
  rGen.States().CopyWithoutAttributes(*mpStates);
  rGen.Alphabet().CopyWithoutAttributes(*mpAlphabet);
  rGen.TransRel().CopyWithoutAttributes(*mpTransRel);
#ifdef FAUDES_CHECKED
  if(!rGen.Valid()) {
    FD_DG("vGenerator()::CopyWithoutAttributes(): invalid generator");
    rGen.DWrite(); 
    abort();
  }
#endif
} 



// Move(gen) destructive copy 
void vGenerator::Move(vGenerator& rGen) {
  FD_DG("vGenerator(" << this << ")::Move(" << &rGen << ")");
  // test types
  bool tmm=false;
  if(typeid(rGen.Alphabet())!=typeid(Alphabet())) tmm=true;
  if(typeid(rGen.States())!=typeid(States())) tmm=true;
  if(typeid(rGen.TransRel())!=typeid(TransRel())) tmm=true;
  if(typeid(rGen.GlobalAttribute())!=typeid(GlobalAttribute())) tmm=true;
  if(tmm) { 
    FD_DG("vGenerator(" << this << ")::Move(" << &rGen << "): using std copy");
    Copy(rGen);
    Clear();
    return;
  }
  // prepare result (call clear for virtual stuff)
  rGen.Clear();
  // have same event symboltable
  rGen.EventSymbolTablep(mpEventSymbolTable);
  // copy state symboltable (todo: make this pointer based?)
  rGen.StateSymbolTable(mStateSymbolTable);
  // copy members
  rGen.Name(Name());
  rGen.StateNamesEnabled(StateNamesEnabled());
  rGen.InjectInitStates(InitStates());
  rGen.InjectMarkedStates(MarkedStates());
  // move and invalidate core members
  rGen.mpStates=mpStates;
  rGen.mpAlphabet=mpAlphabet;
  rGen.mpTransRel=mpTransRel;
  rGen.mpGlobalAttribute=mpGlobalAttribute;
  mpStates=0;
  mpAlphabet=0;
  mpTransRel=0;
  mpGlobalAttribute=0;
  // register core update
  rGen.UpdateCore();
  // install new empty core members
  NewCore();
  // clear myself (incl derived classes members)
  Clear();
}

// operator =
vGenerator& vGenerator::operator = (const vGenerator& rOtherGen) {
  FD_DG("vGenerator(" << this << ")::operator = " << &rOtherGen);
  rOtherGen.Copy(*this);
  return *this;
}

// Version(idx)
void vGenerator::Version(Idx version, vGenerator& rResGen) const {
  FD_DG("vGenerator(" << this << ")::Version(" << version << ")");
  std::ostringstream o;
  o << version;
  Version(o.str(),rResGen);
}

// Version(string)
void vGenerator::Version(const std::string& rVersion, vGenerator& rResGen) const {
  FD_DG("vGenerator(" << this << ")::Version(" << rVersion << ")");
  // second arg must not be us if(this==&rResGen) {exception ..}
  if (this==&rResGen) {
    FD_ERR("vGenerator::Version(): rResGen must not be this generator");
    abort();
  }
  // prepare Empty generator
  rResGen.Clear();
  rResGen.GlobalAttribute(GlobalAttribute());
  EventSet::Iterator eit;
  StateSet::Iterator lit;
  TransSet::Iterator tit;
  std::map<Idx,Idx> eventoldnewmap;
  rResGen.Name(Name()+"_"+rVersion);
  // create versioned mAlphabet
  for (eit = AlphabetBegin(); eit != AlphabetEnd(); ++eit) {
    Idx newevent= rResGen.InsEvent(EventName(*eit)+"_"+rVersion); 
    eventoldnewmap[*eit] = newevent;
    rResGen.EventAttribute(newevent,EventAttribute(*eit));
  }
  // create new stateset
  for (lit = StatesBegin(); lit != StatesEnd(); ++lit) {
    rResGen.InsState(*lit);
    rResGen.StateAttribute(*lit,StateAttribute(*lit)); //would be faster if directly copied ?
    if (StateName(*lit) != "") rResGen.StateName(*lit,StateName(*lit));
  }
  // created versioned transrel
  for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
    Transition trans=Transition(tit->X1, eventoldnewmap[tit->Ev], tit->X2);
    rResGen.SetTransition(trans);
    rResGen.TransAttribute(trans, TransAttribute(*tit));
  }
  // set i/m states
  rResGen.mInitStates=mInitStates;
  rResGen.mMarkedStates=mMarkedStates;
  // bahavioural flags
  rResGen.mStateNamesEnabled = mStateNamesEnabled;
}


// Id()
Idx vGenerator::Id(void) const {
  return mId;
}

// Name(rName)
void vGenerator::Name(const std::string& rName) {
  FD_DV("vGenerator(" << this << ")::Name(\"" << rName << "\")");
  mMyName = rName;
}

// Name()
std::string vGenerator::Name(void) const {
  return mMyName;
}

// Valid()
bool vGenerator::Valid(void) {
  FD_DG("vGenerator(" << this << ")::Valid()");
  // core members on heap
  if(mpAlphabet==0) return false;
  if(mpStates==0) return false;
  if(mpTransRel==0) return false;
  if(mpGlobalAttribute==0) return false;
  // transitions to be known
  TransSet::Iterator tit;
  StateSet::Iterator lit;
  for(tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
    if(! ExistsState(tit->X1)) return false;
    if(! ExistsEvent(tit->Ev)) return false;
    if(! ExistsState(tit->X2)) return false;
  }
  // init states to be known
  for(lit = InitStatesBegin(); lit != InitStatesEnd(); ++lit) 
    if(! ExistsState(static_cast<Idx>(*lit))) return false;
  for(lit = MarkedStatesBegin(); lit != MarkedStatesEnd(); ++lit) 
    if(! ExistsState(static_cast<Idx>(*lit))) return false;
  // sets to have proper names
  if(mpAlphabet->Name() != "Alphabet") return false;
  if(mpStates->Name() != "States") return false;
  if(mInitStates.Name() != "InitStates") return false;
  if(mMarkedStates.Name() != "MarkedStates") return false;
  // event symbol table
  NameSet::Iterator eit;
  for(eit = AlphabetBegin(); eit != AlphabetEnd(); eit ++) {
    if(EventName(*eit)=="") return false;
  }
  // done
  return true;
}

// AlphabetSize()
Idx vGenerator::AlphabetSize(void) const {
  return mpAlphabet->Size();
}

// Size()
Idx vGenerator::Size(void) const {
  return mpStates->Size();
}

// Clear()
void vGenerator::Clear(void) {
  FD_DG("vGenerator(" << this << ")::Clear()");
  mpAlphabet->Clear();
  mpStates->Clear();
  mpStateSymbolTable->Clear();
  mpTransRel->Clear();
  mInitStates.Clear();
  mMarkedStates.Clear();
  ClearGlobalAttribute();
}

// ClearGlobalAttribute()
void vGenerator::ClearGlobalAttribute(void) {
  mpGlobalAttribute->Clear(); 
}

// ClearStateAttributes()
void vGenerator::ClearStateAttributes(void) {
  mpStates->ClearAttributes();
}

// ClearEventAttributes()
void vGenerator::ClearEventAttributes(void) {
  mpAlphabet->ClearAttributes();
}

// ClearTransAttributes()
void vGenerator::ClearTransAttributes(void) {
  mpTransRel->ClearAttributes();
}

// ClearAttributes()
void vGenerator::ClearAttributes(void) {
  ClearGlobalAttribute();
  ClearStateAttributes();
  ClearEventAttributes();
  ClearTransAttributes();
}


// ClearStates()
void vGenerator::ClearStates(void) {
  mpStates->Clear();
  mpTransRel->Clear();
  mInitStates.Clear();
  mMarkedStates.Clear();
  mpStateSymbolTable->Clear();
}

// TransRelSize()
Idx vGenerator::TransRelSize(void) const {
  return mpTransRel->Size();
}

// InitStatesSize()
Idx vGenerator::InitStatesSize(void) const {
  return mInitStates.Size();
}

// MarkedStatesSize()
Idx vGenerator::MarkedStatesSize(void) const {
  return mMarkedStates.Size();
}

// AlphabetEmpty()
bool vGenerator::AlphabetEmpty(void) const {
  return mpAlphabet->Empty();
}

// Empty()
bool vGenerator::Empty(void) const {
  return mpStates->Empty();
}

// TransRelEmpty()
bool vGenerator::TransRelEmpty(void) const {
  return mpTransRel->Empty();
}

// InitStatesEmpty()
bool vGenerator::InitStatesEmpty(void) const {
  return mInitStates.Empty();
}

// MarkedStatesEmpty()
bool vGenerator::MarkedStatesEmpty(void) const {
  return mMarkedStates.Empty();
}


// ClearMinStateIndexMap()
void vGenerator::ClearMinStateIndexMap(void) const {
  FD_DG("vGenerator::ClearMinStateIndexMap()");
  // fake const
  vGenerator* fakeconst = const_cast<vGenerator*>(this);
  fakeconst->mMinStateIndexMap.clear();
}

// MinStateIndex(index)
Idx vGenerator::MinStateIndex(Idx index) const {
  std::map<Idx,Idx>::const_iterator minit;
  minit = mMinStateIndexMap.find(index);
  Idx fidx;
  if(minit != mMinStateIndexMap.end()) {
    fidx= minit->second;
  } else {
    fidx=index;
  }
  return fidx;
}

// MinStateIndex()
const std::map<Idx,Idx>& vGenerator::MinStateIndexMap(void) const {
  return mMinStateIndexMap;
}


// SetMinStateIndexMap()
void vGenerator::SetMinStateIndexMap(void) const {
  FD_DG("vGenerator::SetMinStateIndexMap()");
  // fake const
  vGenerator* fakeconst = const_cast<vGenerator*>(this);
  // clear map
  fakeconst->ClearMinStateIndexMap();
  StateSet::Iterator it;
  Idx minindex = 0;
  // if generator states get names
  if(StateNamesEnabled()) {
    // named initial states first
    for(it = InitStatesBegin(); it != InitStatesEnd(); ++it) {
      if(StateName(static_cast<Idx>(*it)) != "") {
	fakeconst->mMinStateIndexMap[*it] = ++minindex;
      }
    }
    // then all other named states
    for(it = StatesBegin(); it != StatesEnd(); ++it) {
      if(mMinStateIndexMap.count(*it) == 0) {
	if(StateName(static_cast<Idx>(*it)) != "") {
	  fakeconst->mMinStateIndexMap[*it] = ++minindex;
	}
      }
    }
    // at last all anonymous states
    for(it = StatesBegin(); it != StatesEnd(); ++it) {
      if(mMinStateIndexMap.count(*it) == 0) {
	fakeconst->mMinStateIndexMap[*it] = ++minindex;
      }
    }
  }
  // if generator states are all anonymous
  else {
    // all initial states first
    for(it = InitStatesBegin(); it != InitStatesEnd(); ++it) {
      fakeconst->mMinStateIndexMap[*it] = ++minindex;
    }
    // then the rest
    for(it = StatesBegin(); it != StatesEnd(); ++it) {
      if(mMinStateIndexMap.count(*it) == 0) {
	fakeconst->mMinStateIndexMap[*it] = ++minindex;
      }
    }
  }
#ifdef FAUDES_DEBUG_CONTAINER
  std::map<Idx,Idx>::const_iterator _it;
  for(_it = mMinStateIndexMap.begin(); _it != mMinStateIndexMap.end(); ++_it) {
    FD_DC("vGenerator::MinStateIndexMap: " << _it->first
	  << " <-- " << SStr(_it->second));
  }
#endif
}

// EventSymbolTablep() const
SymbolTable* vGenerator::EventSymbolTablep(void) const {
  return mpEventSymbolTable;
}

// GlobalEventSymbolTablep() const
SymbolTable* vGenerator::GlobalEventSymbolTablep(void) {
  return SymbolTable::GlobalEventSymbolTablep();
}

// EventSymbolTablep(pSymTab) 
void vGenerator::EventSymbolTablep(SymbolTable* pSymTab) {
  mpEventSymbolTable=pSymTab;
}

// EventSymbolTablep(rOtherGen) 
void vGenerator::EventSymbolTablep(const vGenerator& rOtherGen) {
  EventSymbolTablep(rOtherGen.EventSymbolTablep());
}

// EventIndex(rName)
Idx vGenerator::EventIndex(const std::string& rName) const {
  return mpEventSymbolTable->Index(rName);
}

// EventName(index)
std::string vGenerator::EventName(Idx index) const {
  return mpEventSymbolTable->Symbol(index);
}

// EventName(index, name)
void vGenerator::EventName(Idx index, const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::EventName(" 
      << index << ",\"" << rName << "\")");
#ifdef FAUDES_CHECKED
  if (! ExistsEvent(index)) {
    std::stringstream errstr;
    errstr << "event \"" << index << "\" not found in generator \""
	   << Name() << "\"";
    throw Exception("vGenerator::EventName(name)", errstr.str(), 89);
  }
#endif
  mpEventSymbolTable->SetEntry(index, rName);
}

// UniqueEventName(rName)
std::string vGenerator::UniqueEventName(const std::string& rName) const {
  std::string name=rName;
  if(name=="") name="ev";
  return mpEventSymbolTable->UniqueSymbol(name) ;
}


// EventRename
bool vGenerator::EventRename(Idx event, const std::string& rNewName) {
  // consistency
  FD_DG("vGenerator(" << this << ")::EventRename(" << EStr(event) << ", " << rNewName << ")");
#ifdef FAUDES_CHECKED
  if (! ExistsEvent(event)) {
    std::stringstream errstr;
    errstr << "event \"" << event << "\" not found in generator \""
	   << Name() << "\"";
    throw Exception("vGenerator::EventReame(name)", errstr.str(), 89);
  }
#endif
  // prepare formal result
  bool res=ExistsEvent(rNewName);
  // insert new event
  Idx newidx=InsEvent(rNewName);
  // bail out if events are the same
  if(newidx==event) return true;
  // copy event attribute
  if(!res) EventAttribute(newidx,EventAttribute(event));
  // store new transitions (with their attributes)
  TransSet* newtrans= TransRel().New();
  // iterate over transitions
  TransSet::Iterator tit;
  for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
    if(tit->Ev!=event) continue;
    Transition trans= Transition(tit->X1, newidx, tit->X2);
    newtrans->Insert(trans);
    newtrans->Attribute(trans,TransAttribute(*tit));
    tit=ClrTransition(tit);
  }
  // merge transitions
  for(tit=newtrans->Begin(); tit!=newtrans->End(); tit++) {
    SetTransition(*tit);
    TransAttribute(*tit,newtrans->Attribute(*tit));
  }
  // free temp
  delete newtrans;
  // remore original event
  DelEvent(event);
  // done
  return res;
}



// NewEventSet()
EventSet vGenerator::NewEventSet(void) const {
  EventSet res;
  res.SymbolTablep(mpEventSymbolTable);
  return res;
}

// NewEventSetp()
EventSet* vGenerator::NewEventSetp(void) const {
  EventSet* res = new EventSet();
  res->SymbolTablep(mpEventSymbolTable);
  return res;
}


// StateSymbolTable() const
const SymbolTable& vGenerator::StateSymbolTable(void) const {
  return mStateSymbolTable;
}

// StateSymbolTable(rSymTab) 
void vGenerator::StateSymbolTable(const SymbolTable& rSymTab) {
  mStateSymbolTable=rSymTab;
  mpStateSymbolTable=&mStateSymbolTable;
}

// StateIndex(rName)
Idx vGenerator::StateIndex(const std::string& rName) const {
   return mpStateSymbolTable->Index(rName);
}

// StateName(index)
std::string vGenerator::StateName(Idx index) const {
  return mpStateSymbolTable->Symbol(index);
}

// StateName(index, name)
void vGenerator::StateName(Idx index, const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::StateName(" 
      << index << ",\"" << rName << "\")");
#ifdef FAUDES_CHECKED
  if (! ExistsState(index)) {
    std::stringstream errstr;
    errstr << "state name \"" << rName << "\" not found in generator \""
	   << Name() << "\"";
    throw Exception("vGenerator::StateName(name)", errstr.str(), 90);
  }
#endif
  mpStateSymbolTable->SetEntry(index, rName);
}


// ClearStateNames()
void vGenerator::ClearStateNames(void) {
  FD_DG("vGenerator(" << this << ")::ClearStateNames()");
  mpStateSymbolTable->Clear();
  mStateNamesEnabled = false;
}


// ClrStateName(index)
void vGenerator::ClrStateName(Idx index) {
  FD_DG("Generator(" << this << ")::ClrStateName(\"" << index << "\")");
#ifdef FAUDES_CHECKED
  if (! ExistsState(index)) {
    std::stringstream errstr;
    errstr << "state \"" << index << "\" not found in generator \""
	   << Name() << "\"";
    throw Exception("vGenerator::ClrStateName(name)", errstr.str(), 90);
  }
#endif
  mpStateSymbolTable->ClrEntry(index);
}

// ClrStateName(rName)
void vGenerator::ClrStateName(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::ClrStateName(\"" << rName << "\")");
  Idx index = StateIndex(rName);
  ClrStateName(index);
}


// StateNamesEnabled()
bool vGenerator::StateNamesEnabled(void) const {
  return mStateNamesEnabled;
}

// StateNamesEnabled(flag)
void vGenerator::StateNamesEnabled(bool flag) {
  mStateNamesEnabled = flag;
}

// StateNamesEnabled(flag)
void vGenerator::StateNamesEnabledDefault(bool flag) {
  msStateNamesEnabledDefault = flag;
}

// SetDefaultStateNames()
void vGenerator::SetDefaultStateNames(void) {
  FD_DG("vGenerator(" << this << ")::SetDefaultStateNames()");
  StateSet::Iterator it;
  for (it = StatesBegin(); it != StatesEnd(); ++it) {
    mpStateSymbolTable->SetDefaultSymbol(*it);
  }
}

// EnforceStateNames(rTemplate)
void vGenerator::EnforceStateNames(const std::string& rTemplate) {
  FD_DG("vGenerator(" << this << ")::EnforceStateNames(temp)");
  StateSet::Iterator it;
  for (it = StatesBegin(); it != StatesEnd(); ++it) {
    if(StateName(*it)=="") {
      std::string name=UniqueStateName(rTemplate + "_1");
      StateName(*it,name);
    }
  }
}

// UniqueStateName(rName)
std::string vGenerator::UniqueStateName(const std::string& rName) const {
  std::string name=rName;
  if(name=="") name="st";
  return mpStateSymbolTable->UniqueSymbol(name) ;
}


// iterator AlphabetBegin() const
EventSet::Iterator vGenerator::AlphabetBegin(void) const {
  return mpAlphabet->Begin();
}

// iterator AlphabetEnd() const
EventSet::Iterator vGenerator::AlphabetEnd(void) const {
  return mpAlphabet->End();
}

// iterator StatesBegin() const
StateSet::Iterator vGenerator::StatesBegin(void) const {
  return mpStates->Begin();
}

// iterator StatesEnd() const
StateSet::Iterator vGenerator::StatesEnd(void) const {
  return mpStates->End();
}

// iterator TransRelBegin() const
TransSet::Iterator vGenerator::TransRelBegin(void) const {
  return mpTransRel->Begin();
}

// iterator TransRelEnd() const
TransSet::Iterator vGenerator::TransRelEnd(void) const {
  return mpTransRel->End();
}

// iterator TransRelBegin(x1) const
TransSet::Iterator vGenerator::TransRelBegin(Idx x1) const {
  return mpTransRel->Begin(x1);
}

// iterator TransRelEnd(x1) const
TransSet::Iterator vGenerator::TransRelEnd(Idx x1) const {
  return mpTransRel->End(x1);
}

// iterator TransRelBegin(x1, ev) const
TransSet::Iterator vGenerator::TransRelBegin(Idx x1, Idx ev) const {
  return mpTransRel->Begin(x1, ev);
}

// iterator TransRelEnd(x1, ev) const
TransSet::Iterator vGenerator::TransRelEnd(Idx x1, Idx ev) const {
  return mpTransRel->End(x1, ev);
}

// iterator FindTransition(trans)
TransSet::Iterator vGenerator::FindTransition(const Transition& rTrans) const {
  return mpTransRel->Find(rTrans);
}

// iterator FindTransition(x1, ex x2)
TransSet::Iterator vGenerator::FindTransition(Idx x1, Idx ev, Idx x2) const {
  return mpTransRel->Find(Transition(x1, ev, x2));
}

// iterator FindTransition(x1, ev, x2) 
TransSet::Iterator vGenerator::FindTransition(
   const std::string& rX1, const std::string& rEv, const std::string& rX2) const 
{
  return mpTransRel->Find(StateIndex(rX1), EventIndex(rEv), StateIndex(rX2));
}

// iterator ExistsTransition(trans)
bool vGenerator::ExistsTransition(const Transition& rTrans) const {
  return mpTransRel->Exists(rTrans);
}

// iterator ExistsTransition(x1, ex x2)
bool vGenerator::ExistsTransition(Idx x1, Idx ev, Idx x2) const {
  return mpTransRel->Exists(Transition(x1, ev, x2));
}

// iterator ExistsTransition(x1, ev, x2)
bool vGenerator::ExistsTransition(
  const std::string& rX1, const std::string& rEv, const std::string& rX2) const 
{
  return mpTransRel->Exists(StateIndex(rX1), EventIndex(rEv), StateIndex(rX2));
}


// idx InitState() const
Idx vGenerator::InitState(void) const {
#ifdef FAUDES_CHECKED
  if(mInitStates.Size()!=1) {    
    std::stringstream errstr;
    errstr << "init state does not exist uniquely" << std::endl;
    throw Exception("vGenerator::InitSTate", errstr.str(), 92);
  }
#endif
  return *mInitStates.Begin();
}

// iterator InitStatesBegin() const
StateSet::Iterator vGenerator::InitStatesBegin(void) const {
  return mInitStates.Begin();
}

// iterator InitStatesEnd() const
StateSet::Iterator vGenerator::InitStatesEnd(void) const {
  return mInitStates.End();
}

// iterator MarkedStatesBegin()
StateSet::Iterator vGenerator::MarkedStatesBegin(void) const {
  return mMarkedStates.Begin();
}

// iterator MarkedStatesEnd() const
StateSet::Iterator vGenerator::MarkedStatesEnd(void) const {
  return mMarkedStates.End(); 
}

// InjectAlphabet(newalphabet)
void vGenerator::InjectAlphabet(const EventSet& rNewAlphabet) {
  FD_DG("vGenerator::InjectAlphabet()  " << rNewAlphabet.ToString());
#ifdef FAUDES_CHECKED
  if(rNewAlphabet.SymbolTablep()!=mpEventSymbolTable) {
    std::stringstream errstr;
    errstr << "symboltable mismatch aka not implemented" << std::endl;
    throw Exception("vGenerator::InjectAlphabet", errstr.str(), 88);
  }
#endif
  rNewAlphabet.Copy(*mpAlphabet);
  mpAlphabet->Name("Alphabet");
}

// InsEvent(index)
bool vGenerator::InsEvent(Idx index) {
  FD_DG("vGenerator(" << this << ")::InsEvent(" << index << ")");
  return mpAlphabet->Insert(index);
}

// InsEvent(rName)
Idx vGenerator::InsEvent(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::InsEvent(\"" << rName << "\")");
  return mpAlphabet->Insert(rName);
}

// InsEvents(events)
void vGenerator::InsEvents(const EventSet& events) {
  mpAlphabet->InsertSet(events);
}

// DelEvent(index)
bool vGenerator::DelEvent(Idx index) {
  FD_DG("vGenerator(" << this << ")::DelEvent(" << index << ")");
  mpTransRel->EraseByEv(index);
  return mpAlphabet->Erase(index);
}

// DelEvent(rName)
bool vGenerator::DelEvent(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::DelEvent(\"" << rName << "\")");
  Idx index = mpAlphabet->Index(rName);
  mpTransRel->EraseByEv(index);
  return mpAlphabet->Erase(index);
}

// DelEvents(events)
void vGenerator::DelEvents(const EventSet& rEvents) {
  FD_DG("vGenerator(" << this << ")::DelEvents(\"" 
	<< rEvents.ToString() << "\")");
  EventSet::Iterator it;
  for (it = rEvents.Begin(); it != rEvents.End(); ++it) {
    DelEvent(*it);
  }
}

// DelEventFromAlphabet(index)
bool vGenerator::DelEventFromAlphabet(Idx index) {
  FD_DG("vGenerator(" << this << ")::DelEventFromAlphabet(" 
	<< index << ")");
  return mpAlphabet->Erase(index);
}

// InsState()
Idx vGenerator::InsState(void) {
  FD_DG("vGenerator(" << this << ")::InsState()");
  return mpStates->Insert();
}

// InsState(index)
bool vGenerator::InsState(Idx index) {
  FD_DG("vGenerator(" << this << ")::InsState(" << index << ")");
  return mpStates->Insert(index);
}

// InsState(rName)
Idx vGenerator::InsState(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::InsState(\"" << rName << "\")");
  Idx index=mpStates->Insert();
  StateName(index,rName);
  return index;
}

// InjectState(index)
void vGenerator::InjectState(Idx index) {
  FD_DG("vGenerator(" << this << ")::InjectState(\"" << SStr(index) << "\")");
  mpStates->Insert(index);
}

// InjectStates(rNewStates)
void vGenerator::InjectStates(const StateSet& rNewStates) {
  FD_DG("vGenerator(" << this << ")::InjectStates(" << rNewStates.ToString() << ")");
  rNewStates.Copy(*mpStates);
  mpStates->Name("States");
  mpStateSymbolTable->RestrictDomain(*mpStates);
  FD_DG("vGenerator(" << this << ")::InjectStates(): report " << mpStates->ToString());
}

// InsInitState()
Idx vGenerator::InsInitState(void) {
  FD_DG("vGenerator(" << this << ")::InsInitState()");
  Idx index;
  index = InsState();
  mInitStates.Insert(index);
  return index;
}

// InsInitState(name)
Idx vGenerator::InsInitState(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::InsInitState(\"" << rName << "\")");
  Idx index;
  index = InsState(rName);
  mInitStates.Insert(index);
  return index;
}

// InsInitState(name)
bool vGenerator::InsInitState(Idx index) {
  bool res=InsState(index);
  mInitStates.Insert(index);
  return res;
}

// InsMarkedState()
Idx vGenerator::InsMarkedState(void) {
  FD_DG("vGenerator(" << this << ")::InsMarkedState()");
  Idx index;
  index = InsState();
  mMarkedStates.Insert(index);
  return index;
}

// InsInitState(name)
bool vGenerator::InsMarkedState(Idx index) {
  bool res=InsState(index);
  mMarkedStates.Insert(index);
  return res;
}

// InsMarkedState(rName)
Idx vGenerator::InsMarkedState(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::InsMarkedState(\"" << rName << "\")");
  Idx index;
  index = InsState(rName);
  mMarkedStates.Insert(index);
  return index;
}

// DelState(index)
bool vGenerator::DelState(Idx index) {
  FD_DG("vGenerator(" << this << ")::DelState(" << index << ")");
  // mInitStates
  mInitStates.Erase(index);
  // mstates
  mMarkedStates.Erase(index);
  // transrel 
  mpTransRel->EraseByX1OrX2(index);
  // symbolic name
  mpStateSymbolTable->ClrEntry(index);
  // finally ... remove the state
  return mpStates->Erase(index);
}

// DelState(rName)
bool vGenerator::DelState(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::DelState(\"" << rName << "\")");
  Idx index;
  index = StateIndex(rName);
#ifdef FAUDES_CHECKED
  if (index == 0) {
    std::stringstream errstr;
    errstr << "state name \"" << rName << "\" not found in generator \""
	   << Name() << "\"";
    throw Exception("vGenerator::DelState(name)", errstr.str(), 90);
  }
#endif
  return DelState(index);
}

// DelStates(rDelStates)
void vGenerator::DelStates(const StateSet& rDelStates) {
  FD_DG("vGenerator(" << this << ")::DelStates(" 
	<< rDelStates.ToString() << ")");
  StateSet::Iterator cit;
  StateSet::Iterator cit_end;
  // symbolic state names
  if (StateNamesEnabled()) {
    for (cit = rDelStates.Begin(); cit != rDelStates.End(); ++cit) {
      mpStateSymbolTable->ClrEntry(*cit);
    }
  }
  // statesets
  mpStates->EraseSet(rDelStates);
  mInitStates.EraseSet(rDelStates);
  mMarkedStates.EraseSet(rDelStates);
  // mpTransRel:
  mpTransRel->EraseByX1OrX2(rDelStates);
  // original (slow)
  //for (cit = rDelStates.Begin(); cit != rDelStates.End(); ++cit) {
  //  mpTransRel->EraseByX1OrX2(*cit);
  //}
}

// DelStateFromStates(index)
bool vGenerator::DelStateFromStates(Idx index) {
  FD_DG("vGenerator(" << this << ")::DelStateFromStates(" << index << ")");
  // mStates + global
  return mpStates->Erase(index);
}

// DelStateFromStates(pos)
StateSet::Iterator vGenerator::DelStateFromStates(StateSet::Iterator pos) {
  FD_DG("vGenerator(" << this << ")::DelState(" << *pos << ")");
  return mpStates->Erase(pos);
}

// SetInitState(index)
void vGenerator::SetInitState(Idx index) {
  FD_DG("vGenerator(" << this << ")::SetInitState(" << index << ")");
#ifdef FAUDES_CHECKED
  if (! mpStates->Exists(index)) {
    std::stringstream errstr;
    errstr << "vGenerator::SetMarkedState: index " << index 
	   << " not in stateset";
    throw Exception("vGenerator::SetInitState(..)", errstr.str(), 91);
  }
#endif
  mInitStates.Insert(index);
}

// SetInitState(rName)
void vGenerator::SetInitState(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::SetInitState(\"" << rName << "\")");
  Idx index = StateIndex(rName);
#ifdef FAUDES_CHECKED
  if (index == 0) {
    std::stringstream errstr;
    errstr << "State name \"" << rName << "\" not known in Generator";
    throw Exception("vGenerator::SetInitState(..)", errstr.str(), 90);
  }
#endif
  SetInitState(index);
}

// InjectInitStates(rNewInitStates)
void vGenerator::InjectInitStates(const StateSet& rNewInitStates) {
  FD_DG("vGenerator(" << this << ")::InjectInitStates(" 
	<< rNewInitStates.ToString() << ")");
  mInitStates = rNewInitStates;
  mInitStates.Name("InitStates");
}

// ClrInitState(index)
void vGenerator::ClrInitState(Idx index) {
  FD_DG("vGenerator(" << this << ")::ClrInitState(" << index << ")");
#ifdef FAUDES_CHECKED
  if (! mpStates->Exists(index)) {
    std::stringstream errstr;
    errstr << "vGenerator::SetMarkedState: index " << index 
	   << " not in stateset";
    throw Exception("vGenerator::ClrInitState(..)", errstr.str(), 91);
  }
#endif
  mInitStates.Erase(index);
}

// ClrInitState(rName)
void vGenerator::ClrInitState(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::ClrInitState(\"" << rName << "\")");
  Idx index = StateIndex(rName);
#ifdef FAUDES_CHECKED
  if (index == 0) {
    std::stringstream errstr;
    errstr << "State name \"" << rName << "\" not known in Generator";
    throw Exception("vGenerator::ClrInitState(..)", errstr.str(), 90);
  }
#endif
  ClrInitState(index);
}

// ClrInitState(pos)
StateSet::Iterator vGenerator::ClrInitState(StateSet::Iterator pos) {
  FD_DG("vGenerator(" << this << ")::ClrInitState(" << *pos << ")");
  return mInitStates.Erase(pos);
}

// ClearInitStates()
void vGenerator::ClearInitStates(void) {
  mInitStates.Clear();
}

// SetMarkedState(index)
void vGenerator::SetMarkedState(Idx index) {
  FD_DG("vGenerator(" << this << ")::SetMarkedState(" << index << ")");
#ifdef FAUDES_CHECKED
  if (! mpStates->Exists(index)) {
    std::stringstream errstr;
    errstr << "vGenerator::SetMarkedState: index " << index 
	   << " not in stateset";
    throw Exception("vGenerator::SetMarkedState(..)", errstr.str(), 91);
  }
#endif
  mMarkedStates.Insert(index);
}

// SetMarkedState(rName)
void vGenerator::SetMarkedState(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::SetMarkedState(\"" << rName << "\")");
  Idx index = StateIndex(rName);
#ifdef FAUDES_CHECKED
  if (index == 0) {
    std::stringstream errstr;
    errstr << "State name \"" << rName << "\" not known in Generator";
    throw Exception("vGenerator::SetMarkedState(..)", errstr.str(), 90);
  }
#endif
  SetMarkedState(index);
}

// InjectMarkedStates(rNewMarkedStates)
void vGenerator::InjectMarkedStates(const StateSet& rNewMarkedStates) {
  FD_DG("vGenerator(" << this << ")::InjectMarkedStates(" 
	<< rNewMarkedStates.ToString() << ")");
  mMarkedStates = rNewMarkedStates;
  mMarkedStates.Name("MarkedStates");
}

// ClrMarkedState(index)
void vGenerator::ClrMarkedState(Idx index) {
  FD_DG("vGenerator(" << this << ")::ClrMarkedState(" << index << ")");
#ifdef FAUDES_CHECKED
  if (! mpStates->Exists(index)) {
    std::stringstream errstr;
    errstr << "vGenerator::ClrMarkedState: index " << index 
	   << " not in stateset";
    throw Exception("vGenerator::ClrMarkedState(..)", errstr.str(), 91);
  }
#endif
  mMarkedStates.Erase(index);
}

// ClrMarkedState(rName)
void vGenerator::ClrMarkedState(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::ClrMarkedState(\"" << rName << "\")");
  Idx index = StateIndex(rName);
#ifdef FAUDES_CHECKED
  if (index == 0) {
    std::stringstream errstr;
    errstr << "State name \"" << rName << "\" not known in Generator";
    throw Exception("vGenerator::ClrMarkedState(..)", errstr.str(), 90);
  }
#endif
  ClrMarkedState(index);
}

// ClrMarkedState(pos)
StateSet::Iterator vGenerator::ClrMarkedState(StateSet::Iterator pos) {
  FD_DG("vGenerator(" << this << ")::ClrMarkedState(" << *pos << ")");
  return mMarkedStates.Erase(pos);
}

// ClearMarkedStates()
void vGenerator::ClearMarkedStates(void) {
  mMarkedStates.Clear();
}

// InjectTransition(newtrans)
void vGenerator::InjectTransition(const Transition& rTrans) {
  FD_DG("vGenerator::InjectTransition(" << TStr(rTrans) << ")");
  mpTransRel->Insert(rTrans);
}

// InjectTransRel(newtransrel)
void vGenerator::InjectTransRel(const TransSet& rNewTransrel) {
  FD_DG("vGenerator::InjectTransRel(...)");
  rNewTransrel.Copy(*mpTransRel);
  mpTransRel->Name("TransRel");
}

// SetTransition(rX1, rEv, rX2)
bool vGenerator::SetTransition(const std::string& rX1, const std::string& rEv, const std::string& rX2) {
  FD_DG("vGenerator(" << this << ")::SetTransition(\""
	<< rX1 << "\", \"" << rEv << "\", \"" << rX2 << "\")");
  Idx x1 = StateIndex(rX1);
  Idx x2 = StateIndex(rX2);
#ifdef FAUDES_CHECKED
  if (x1 == 0) {
    FD_ERR("vGenerator::SetTransition: state " << rX1 
	   << " not in stateset");
    std::stringstream errstr;
    errstr << "State name " << rX1 << " not found in Generator";
    throw Exception("vGenerator::SetTransition(..)", errstr.str(), 90);
  }
  if (! mpAlphabet->Exists(rEv)) {
    FD_ERR("vGenerator::SetTransition: event " << rEv << " not in alphabet");
    std::stringstream errstr;
    errstr << "Event name " << rEv << " not found in event domain of Generator";
    throw Exception("vGenerator::SetTransition(..)", errstr.str(), 95);
  }
  if (x2 == 0) {
    FD_ERR("vGenerator::SetTransition: state " << rX2 << " not in stateset");
    std::stringstream errstr;
    errstr << "State name " << rX2 << " not found in Generator";
    throw Exception("vGenerator::SetTransition(..)", errstr.str(), 90);
  }
#endif
  return SetTransition(Transition(x1, EventIndex(rEv), x2));
}


// SetTransition(x1, ev, x2)
bool vGenerator::SetTransition(Idx x1, Idx ev, Idx x2) {
  return SetTransition(Transition(x1,ev,x2));
}

// SetTransition(rTransition)
bool vGenerator::SetTransition(const Transition& rTransition) {
  FD_DG("vGenerator(" << this << ")::SetTransition(" << rTransition.X1 << "," 
	<< rTransition.Ev << "," << rTransition.X2 << ")");
#ifdef FAUDES_CHECKED
  if (! mpStates->Exists(rTransition.X1)) {
    std::stringstream errstr;
    errstr << "vGenerator::SetTransition: state " << rTransition.X1 
	   << " not in stateset";
    throw Exception("vGenerator::SetTransition(..)", errstr.str(), 95);
  }
  if (! mpAlphabet->Exists(rTransition.Ev)) {
    std::stringstream errstr;
    errstr << "vGenerator::SetTransition: event " << rTransition.Ev 
	   << " not in alphabet ";
    throw Exception("vGenerator::SetTransition(..)", errstr.str(), 95);
  }
  if (! mpStates->Exists(rTransition.X2)) {
    std::stringstream errstr;
    errstr << "vGenerator::SetTransition: state " << rTransition.X2 
	   << " not in stateset";
    throw Exception("vGenerator::SetTransition(..)", errstr.str(), 95);
  }
#endif
  return mpTransRel->Insert(rTransition);
}

// ClrTransition.X1, ev, x2)
void vGenerator::ClrTransition(Idx x1, Idx ev, Idx x2) {
  FD_DG("vGenerator(" << this << ")::ClrTransition(" 
	<< x1 << "," << ev << "," << x2 << ")");
  mpTransRel->Erase(x1, ev, x2);
}

// ClrTransition(rTransition)
void vGenerator::ClrTransition(const Transition& rTransition) {
  FD_DG("vGenerator(" << this << ")::ClrTransition(" << TStr(rTransition) << ")");
  mpTransRel->Erase(rTransition);
}

// ClrTransition(it)
TransSet::Iterator vGenerator::ClrTransition(TransSet::Iterator it) {
  FD_DG("vGenerator(" << this << ")::ClrTransition(" << TStr(*it)<< ")" );
  return mpTransRel->Erase(it);
}

// TransAttribute(trans, attr)
void vGenerator::TransAttribute(const Transition& rTrans, const Type& rAttr) {
  FD_DG("vGenerator(" << this << ")::TransAttribute(" 
	<< TStr(rTrans) << ",\"" << rAttr.ToString() << "\")");
  mpTransRel->Attribute(rTrans, rAttr);
}

// TransAttributep(trans)
AttributeVoid* vGenerator::TransAttributep(const Transition& rTrans)  {
  return mpTransRel->Attributep(rTrans);
}


// TransAttribute(trans)
const AttributeVoid& vGenerator::TransAttribute(const Transition& rTrans) const {
  return mpTransRel->Attribute(rTrans);
}

// ClrTransAttribute(trans)
void vGenerator::ClrTransAttribute(const Transition& rTrans)  {
  mpTransRel->ClrAttribute(rTrans);
}


// ClearTransRel()
void vGenerator::ClearTransRel(void) {
  mpTransRel->Clear();
}

// EventAttribute(index, attr)
void vGenerator::EventAttribute(Idx index, const Type& rAttr) {
  FD_DG("vGenerator(" << this << ")::EventAttribute(" 
	<< EStr(index) << ",\"" << rAttr.ToString() << "\")");
  mpAlphabet->Attribute(index, rAttr);
}

// EventAttributes(set)
void vGenerator::EventAttributes(const EventSet& rEventSet) {
  FD_DG("vGenerator(" << this << ")::EventAttributes(" 
	<< rEventSet.ToString() << "\")");
  mpAlphabet->Attributes(rEventSet);
}

// ClrEventAttribute(index)
void vGenerator::ClrEventAttribute(Idx index) {
  FD_DG("vGenerator(" << this << ")::ClrEventAttribute(\"" << EStr(index) << "\")");
  mpAlphabet->ClrAttribute(index);
}

// StateAttribute(index, attr)
void vGenerator::StateAttribute(Idx index, const Type& rAttr) {
  FD_DG("vGenerator(" << this << ")::StateAttribute(" 
	<< SStr(index) << ",\"" << rAttr.ToString() << "\")");
  mpStates->Attribute(index, rAttr);
}

// ClrStateAttribute(index)
void vGenerator::ClrStateAttribute(Idx index) {
  FD_DG("vGenerator(" << this << ")::ClrStateAttribute(\"" << index << "\")");
  mpStates->ClrAttribute(index);
}

// ExistsEvent(index)
bool vGenerator::ExistsEvent(Idx index) const {
  return mpAlphabet->Exists(index);
}

// ExistsEvent(rName)
bool vGenerator::ExistsEvent(const std::string& rName) const {
  return mpAlphabet->Exists(rName);
}

// FindEvent(index) const
EventSet::Iterator vGenerator::FindEvent(Idx index) const {
  return mpAlphabet->Find(index);
}

// FindEvent(rName)
EventSet::Iterator vGenerator::FindEvent(const std::string& rName) const {
  return mpAlphabet->Find(rName);
}

// ExistsState(index)
bool vGenerator::ExistsState(Idx index) const {
  return mpStates->Exists(index);
}

// ExistsName(name)
bool vGenerator::ExistsState(const std::string& rName) const {
  return ExistsState(StateIndex(rName));
}

// FindState(rName) const
StateSet::Iterator vGenerator::FindState(const std::string& rName) const {
  return mpStates->Find(mpStateSymbolTable->Index(rName));
}

// FindState(index) const
StateSet::Iterator vGenerator::FindState(Idx index) const {
  return mpStates->Find(index);
}

// ExistsInitState(index)
bool vGenerator::ExistsInitState(Idx index) const {
  return mInitStates.Exists(index);
}

// FindInitState(index)
StateSet::Iterator vGenerator::FindInitState(Idx index) const {
  return mInitStates.Find(index);
}

// ExistsMarkedState(index)
bool vGenerator::ExistsMarkedState(Idx index) const {
  return mMarkedStates.Exists(index);
}

// FindMarkedState(index)
StateSet::Iterator vGenerator::FindMarkedState(Idx index) const {
  return mMarkedStates.Find(index);
}

// EventAttribute(index)
const AttributeVoid&  vGenerator::EventAttribute(Idx index) const {
  return mpAlphabet->Attribute(index);
}

// EventAttributep(index)
AttributeVoid* vGenerator::EventAttributep(Idx index) {
  return mpAlphabet->Attributep(index);
}

// EventAttribute(rName)
const AttributeVoid&  vGenerator::EventAttribute(const std::string& rName) const {
  return EventAttribute(EventIndex(rName));
}

// EventAttributep(rName)
AttributeVoid* vGenerator::EventAttributep(const std::string& rName) {
  return EventAttributep(EventIndex(rName));
}

// StateAttribute(index)
const AttributeVoid&  vGenerator::StateAttribute(Idx index) const {
  return mpStates->Attribute(index);
}

// StateAttributep(index)
AttributeVoid*  vGenerator::StateAttributep(Idx index) {
  return mpStates->Attributep(index);
}

// GlobalAttribute(attr)
void vGenerator::GlobalAttribute(const Type& rAttr) {
  FD_DG("vGenerator(" << this << ")::GlobalAttribute(" 
	<< rAttr.ToString() << "\")");
  // set to void is ok for silient ignore
  if(typeid(rAttr)==typeid(AttributeVoid)) return;
  // error:
  std::stringstream errstr;
  errstr << "cannot cast global attribute " << rAttr.ToString() << " for generator " << Name();
  throw Exception("vGenerator::GlobalAttribute", errstr.str(), 63);
}

// GlobalAttributeTry(attr)
void vGenerator::GlobalAttributeTry(const Type& rAttr) {
  FD_DG("vGenerator(" << this << ")::GlobalAttributeTry(" 
	<< rAttr.ToString() << "\")");
  // ignore
}

// GlobalAttribute()
const AttributeVoid& vGenerator::GlobalAttribute(void) const {
  FD_DG("vGenerator(" << this << ")::GlobalAttribute()");
  return *mpGlobalAttribute;
}

// GlobalAttributep()
AttributeVoid* vGenerator::GlobalAttributep(void) {
  FD_DG("vGenerator(" << this << ")::GlobalAttributep()");
  return mpGlobalAttribute;
}


// Alphabet()
const EventSet& vGenerator::Alphabet(void) const {
  return *mpAlphabet;
}

// States()
const StateSet& vGenerator::States(void) const {
  return *mpStates;
}

// TransRel()
const TransSet& vGenerator::TransRel(void) const {
  return *mpTransRel;
}


// TransRel(res)
void vGenerator::TransRel(TransSetX1EvX2& res) const { mpTransRel->ReSort(res); }
void vGenerator::TransRel(TransSetEvX1X2& res) const { mpTransRel->ReSort(res); }
void vGenerator::TransRel(TransSetEvX2X1& res) const { mpTransRel->ReSort(res); }
void vGenerator::TransRel(TransSetX2EvX1& res) const { mpTransRel->ReSort(res); }
void vGenerator::TransRel(TransSetX2X1Ev& res) const { mpTransRel->ReSort(res); }
void vGenerator::TransRel(TransSetX1X2Ev& res) const { mpTransRel->ReSort(res); }


// InitStates()
const StateSet& vGenerator::InitStates(void) const {
  return mInitStates;
}

// MarkedStates()
const StateSet& vGenerator::MarkedStates(void) const {
  return mMarkedStates;
}

// MinimizeAlphabet()
void vGenerator::MinimizeAlphabet(void) {
  mpAlphabet->NameSet::EraseSet(UnusedEvents());
}

// UsedEvents()
EventSet vGenerator::UsedEvents(void) const {
  EventSet resultset = NewEventSet();
  TransSet::Iterator it;
  for (it = mpTransRel->Begin(); it != mpTransRel->End(); ++it) {
    resultset.Insert(it->Ev);
  }
  return resultset;
}

// UnusedEvents()
EventSet vGenerator::UnusedEvents(void) const {
  return *mpAlphabet - UsedEvents();
}

// ActiveEventSet(x1)
EventSet vGenerator::ActiveEventSet(Idx x1) const {
  EventSet result = NewEventSet();
  TransSet::Iterator it;
  for (it = TransRelBegin(x1); it != TransRelEnd(x1); ++it) {
    result.Insert(it->Ev);
  }
  return result;
}

// ActiveTransSet(x1)
TransSet vGenerator::ActiveTransSet(Idx x1) const {
  TransSet result;
  TransSet::Iterator it;
  for (it = TransRelBegin(x1); it != TransRelEnd(x1); ++it) {
    result.Insert(*it);
  }
  return result;
}

// TransRelStateSpace()
StateSet vGenerator::TransRelStateSpace(void) const {
  StateSet states = StateSet();
  TransSet::Iterator it;
  for (it=mpTransRel->Begin(); it != mpTransRel->End(); ++it) {
    states.Insert(it->X1);
    states.Insert(it->X2);
  }
  return states;
}

// TransRelStateSpace(x1)
StateSet vGenerator::TransRelStateSpace(Idx x1) const {
  StateSet states = StateSet();
  TransSet::Iterator it = mpTransRel->Begin(x1);
  TransSet::Iterator it_end = mpTransRel->End(x1);
  for (; it != it_end; ++it) {
    states.Insert(it->X2);
  }
  return states;
}


// CheckAccessible(accessibleset, startState)
void vGenerator::CheckAccessible(StateSet& accessibleset, Idx startState) const {
  if (! accessibleset.Exists(startState)) {
    accessibleset.Insert(startState);
    TransSet::Iterator it = TransRelBegin(startState);
    TransSet::Iterator it_end = TransRelEnd(startState);
    for (; it != it_end; ++it) {
      CheckAccessible(accessibleset, it->X2);
    }
  }
}

// CheckCoaccessible(coaccessibleset, rtransrel, startState)
void vGenerator::CheckCoaccessible(StateSet& coaccessibleset, const TransSetX2EvX1& rtrel, Idx startState) const {
  if (! coaccessibleset.Exists(startState)) {
    coaccessibleset.Insert(startState);
    TransSetX2EvX1::Iterator it = rtrel.BeginByX2(startState);
    TransSetX2EvX1::Iterator it_end = rtrel.EndByX2(startState);
    for (; it != it_end; ++it) {
      CheckCoaccessible(coaccessibleset, rtrel, it->X1);
    }
  }
}

// AccessibleSet()
StateSet vGenerator::AccessibleSet(void) const {
  StateSet accessibleset = StateSet();
  StateSet::Iterator it;
  for (it = InitStatesBegin(); it != InitStatesEnd(); ++it) {
    CheckAccessible(accessibleset, *it);
  }
  accessibleset.Name("accessibleset");
  return accessibleset;
}

// Accessible()
bool vGenerator::Accessible(void) {
  StateSet accessibleset = AccessibleSet();
  StateSet not_accessible = *mpStates - accessibleset;
  DelStates(not_accessible);
  // return true if there is an initial state
  if (!mInitStates.Empty()) {
    FD_DF("vGenerator::accessible: generator is accessible");
    return true;
  }
  else {
    FD_DF("vGenerator::accessible: generator is accessible but empty");
    return false;
  }

}

// IsAccessible()
bool vGenerator::IsAccessible(void) const {
  if ((AccessibleSet() == *mpStates) && (! mInitStates.Empty())) {
    FD_DF("vGenerator::accessible: generator is accessible");
    return true;
  }
  else {
    FD_DF("vGenerator::accessible: generator is not accessible");
    return false;
  }
}

// CoaccessibleSet()
StateSet vGenerator::CoaccessibleSet(void) const {
  StateSet coaccessibleset = StateSet();
  StateSet::Iterator it;
  // build reverse transition relation
  TransSetX2EvX1 rtrel;
  TransRel(rtrel);
  for (it = MarkedStatesBegin(); it != MarkedStatesEnd(); ++it) {
    CheckCoaccessible(coaccessibleset, rtrel, *it);
  }
  coaccessibleset.Name("coaccessibleset");
  return coaccessibleset;
}

// Coaccessible()
bool vGenerator::Coaccessible(void) {
  StateSet coaccessibleset = CoaccessibleSet();
  StateSet not_coaccessible = *mpStates - coaccessibleset;
  DelStates(not_coaccessible);
  // return true if there is a marked state
  if (! mMarkedStates.Empty()) {
    FD_DF("vGenerator::coaccessible: generator is coaccessible");
    return true;
  }
  else {
    FD_DF("vGenerator::coaccessible: generator is not coaccessible");
    return false;
  }
}

// IsCoaccessible()
bool vGenerator::IsCoaccessible(void) const {
  if ((CoaccessibleSet() == *mpStates) && (! mMarkedStates.Empty())) {
    FD_DF("vGenerator::coaccessible: generator is coaccessible");
    return true;
  }
  else {
    FD_DF("vGenerator::coaccessible: generator is not coaccessible");
    return false;
  }
}

// TrimSet()
StateSet vGenerator::TrimSet(void) const {
  FD_DF("vGenerator::trimset: trim states: " 
	<< StateSet(AccessibleSet() * CoaccessibleSet()).ToString());
  return AccessibleSet() * CoaccessibleSet();
}

// Trim()
bool vGenerator::Trim(void) {
  FD_DF("vGenerator::trim: generator states: " << mpStates->ToString());
  // better: conpute sets and do one state delete
  bool accessiblebool = Accessible();
  bool coaccessiblebool = Coaccessible();
  FD_DF("vGenerator::trim: trim states: " << mpStates->ToString());
  if (accessiblebool && coaccessiblebool) {
    FD_DF("vGenerator::trim: generator is trim");
    return true;
  }
  else {
    FD_DF("vGenerator::trim: generator is not trim");
    return false;
  }
}

// IsTrim()
bool vGenerator::IsTrim(void) const {
  if (IsAccessible() && IsCoaccessible()) {
    FD_DF("vGenerator::trim: generator is trim");
    return true;
  }
  else {
    FD_DF("vGenerator::trim: generator is not trim");
    return false;
  }
}

// BlockingSet()
StateSet vGenerator::BlockingSet(void) const {
  FD_DF("vGenerator::BlockingSet: blocking states: " 
	<< StateSet(AccessibleSet() - CoaccessibleSet()).ToString());
  return AccessibleSet() - CoaccessibleSet();
}

// IsDeterministic()
bool vGenerator::IsDeterministic(void) const {
  // if there is more than one initial state ... nondet
  if (InitStatesSize() != 1) {
    FD_DG("vGenerator::IsDeterministic: number of initial states not 1");
    return false;
  }
  // if there is a state/event pair with non-unique successor ... nondet
  if (TransRelSize() < 2) return true;
  TransSet::Iterator it1;
  TransSet::Iterator it2;
  for (it1 = TransRelBegin(), it2 = it1++; it1 != TransRelEnd(); it2 = it1++) {
    if ((it1->X1 == it2->X1) && (it1->Ev == it2->Ev)) {
      FD_WP("IsDeterministic: nondeterministic generator. at least one state "
	    << "contains more than on transition with same event: " 
	    << SStr(it1->X1) << ", event " << EStr(it1->Ev));
      return false;
    }
  }
  // in all other cases generator is deterministic
  return true;
}

// DoWrite()
void vGenerator::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) pContext;
  // figure section
  std::string label=rLabel;
  if(label=="") label="Generator"; 
  FD_DG("vGenerator(" << this << ")::DoWrite(): section " << label);
  // reindex for file output
  if(rTw.DestMode()==TokenWriter::File) SetMinStateIndexMap();
  // write generator
  rTw.WriteBegin(label);
  rTw << mMyName;
  rTw << "\n";
  rTw << "\n";
  WriteStatistics(rTw);
  rTw << "\n";
  mpAlphabet->Write(rTw);
  rTw << "\n";
  WriteStateSet(rTw, *mpStates);
  rTw << "\n";
  WriteTransRel(rTw);
  rTw << "\n";
  WriteStateSet(rTw, mInitStates);
  rTw << "\n";
  WriteStateSet(rTw, mMarkedStates);
  rTw << "\n";
  mpGlobalAttribute->Write(rTw);
  rTw << "\n";
  rTw.WriteEnd(label);
  // end of reindex
  ClearMinStateIndexMap();
}

// DoDWrite()
void vGenerator::DoDWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) pContext;
  // figure section
  std::string label=rLabel;
  if(label=="") label="Generator"; 
  FD_DG("vGenerator(" << this << ")::DoDWrite(): section " << label);
  // write generator
  rTw.WriteBegin(label);
  rTw << mMyName;
  rTw << "\n";
  rTw << "\n";
  WriteStatistics(rTw);
  rTw << "\n";
  mpAlphabet->DWrite(rTw);
  rTw << "\n";
  DWriteStateSet(rTw, *mpStates);
  rTw << "\n";
  DWriteTransRel(rTw);
  rTw << "\n";
  DWriteStateSet(rTw, mInitStates);
  rTw << "\n";
  DWriteStateSet(rTw, mMarkedStates);
  rTw << "\n";
  mpGlobalAttribute->Write(rTw);
  rTw << "\n";
  rTw.WriteEnd(label);
  rTw << "\n";
}

// WriteAlphabet() 
void vGenerator::WriteAlphabet(void) const {
  TokenWriter tw(TokenWriter::Stdout);
  WriteAlphabet(tw);
}

// AlphabetToString()
std::string vGenerator::AlphabetToString(void) const {
  TokenWriter tw(TokenWriter::String);
  WriteAlphabet(tw);
  return tw.Str();
}

// WriteAlphabet(rTw&) 
void vGenerator::WriteAlphabet(TokenWriter& rTw) const { 
  mpAlphabet->Write(rTw);
}

// WriteStateSet(rStateSet&) 
void vGenerator::WriteStateSet(const StateSet&  rStateSet) const {
  TokenWriter tw(TokenWriter::Stdout);
  WriteStateSet(tw,rStateSet);
}

// StateSetToString()
std::string vGenerator::StateSetToString(const StateSet&  rStateSet) const {
  TokenWriter tw(TokenWriter::String);
  WriteStateSet(tw,rStateSet);
  return tw.Str();
}


// WriteStateSet(rTw&, rStateSet&) 
void vGenerator::WriteStateSet(TokenWriter& rTw, const StateSet&  rStateSet) const {
  rTw.WriteBegin(rStateSet.Name());
  // build reverse index map of states to write ( fileidx->idx )
  // (this ensures named states to be written first; see SetMinStateIndexMap())
  std::map<Idx,Idx> reversemap;
  std::map<Idx,Idx>::const_iterator minit;
  StateSet::Iterator sit;
  for (sit = rStateSet.Begin(); sit != rStateSet.End(); ++sit) {
    reversemap[MinStateIndex(*sit)] = *sit;
  }
  // iterate states to write 
  for(minit = reversemap.begin(); minit != reversemap.end(); ++minit) {
    // identify anonymous block (consecutive state indices)
    std::map<Idx,Idx>::const_iterator conit=minit;
    Idx start = conit->first;
    Idx anoncount = 0;
    for(; conit != reversemap.end(); ++conit) {
      if(StateName(conit->second) != "") break;
      if(!StateAttribute(conit->second).IsDefault()) break;
      if(conit->first != start+anoncount) break;
      ++anoncount;
    }
    // write anonymous block
    if (anoncount > FD_CONSECUTIVE) {
      rTw.WriteBegin("Consecutive");
      rTw << start;
      rTw << start+anoncount-1;
      rTw.WriteEnd("Consecutive");
      minit=conit;
    } 
    // break loop
    if(minit == reversemap.end() )
      break;
    // write non anonymous state name/idx
    std::string statename = StateName(minit->second);
    if (statename != "") {
      rTw << statename;
    } else {
      rTw << minit->first;
    }
    // write state attribute
    const AttributeVoid& attr=rStateSet.Attribute(minit->second);
    attr.Write(rTw);
  }
  rTw.WriteEnd(rStateSet.Name());
}


// DWriteStateSet(rTw&, rStateSet&) 
void vGenerator::DWriteStateSet(TokenWriter& rTw, const StateSet& rStateSet) const {
  rTw.WriteBegin(rStateSet.Name());
  StateSet::Iterator lit;
  std::map<Idx,Idx>::const_iterator minit;
  for(lit = rStateSet.Begin(); lit != rStateSet.End(); ++lit) {
    rTw << SStr(*lit);
    const AttributeVoid& attr=rStateSet.Attribute(*lit);
    attr.Write(rTw);
  }
  rTw.WriteEnd(rStateSet.Name());
}


// StatesToString()
std::string vGenerator::StatesToString(void) const {
  return StateSetToString(*mpStates);
}

// MarkedStatesToString()
std::string vGenerator::MarkedStatesToString(void) const {
  return StateSetToString(mMarkedStates);
}

// InitStatesToString()
std::string vGenerator::InitStatesToString(void) const {
  return StateSetToString(mInitStates);
}


// WriteTransRel() 
void vGenerator::WriteTransRel(void) const {
  TokenWriter tw(TokenWriter::Stdout);
  WriteTransRel(tw);
}

// TransRelToString()
std::string vGenerator::TransRelToString(void) const {
  TokenWriter tw(TokenWriter::String);
  WriteTransRel(tw);
  return tw.Str();
}

// WriteTransRel(rTw&)
void vGenerator::WriteTransRel(TokenWriter& rTw) const {
  TransSet::Iterator tit;
  int oldcolumns = rTw.Columns();
  rTw.Columns(3);
  rTw.WriteBegin("TransRel");
  bool smalltransrel = (Size() < FD_SMALLTRANSREL);

  // loop all transitions
  for(tit = mpTransRel->Begin(); tit != mpTransRel->End(); ++tit) {

    // write x1
    Idx x1=MinStateIndex(tit->X1);
    if (smalltransrel) {
      std::string x1name = StateName(tit->X1);
      if (x1name != "") {
	rTw << x1name;
      } else {
	rTw << x1;
      }
    } else {
      rTw << x1;
    }

    // write ev
    rTw << EventName(tit->Ev);

    // write x2
    Idx x2=MinStateIndex(tit->X2);
    if (smalltransrel) {
      std::string x2name = StateName(tit->X2);
      if (x2name != "") {
	rTw << x2name;
      } else {
	rTw << x2;
      }
    } else {
      rTw << x2;
    }

    // write attributes
    TransAttribute(*tit).Write(rTw);

  }
  rTw.WriteEnd("TransRel");
  rTw.Columns(oldcolumns);
}

// DWriteTransRel(rTw&)
void vGenerator::DWriteTransRel(TokenWriter& rTw) const {
  TransSet::Iterator tit;
  int oldcolumns = rTw.Columns();
  rTw.Columns(1);
  rTw.WriteBegin("TransRel");
  for (tit = mpTransRel->Begin(); tit != mpTransRel->End(); ++tit) {
    rTw << TStr(*tit);
    mpTransRel->Attribute(*tit).Write(rTw);
  }
  rTw.WriteEnd("TransRel");
  rTw.Columns(oldcolumns);
}



// WriteStatistics() 
void vGenerator::WriteStatistics(void) const  
{
  TokenWriter tw(TokenWriter::Stdout);
  WriteStatistics(tw);
}

// StatisticsToString()
std::string vGenerator::StatisticsToString(void) const {
  TokenWriter tw(TokenWriter::String);
  WriteStatistics(tw);
  return tw.Str();
}

// WriteStatistics(rTw&)
void vGenerator::WriteStatistics(TokenWriter& rTw) const
{
  rTw.WriteComment("");
  rTw.WriteComment(" Statistics for " + Name());
  rTw.WriteComment(" States:        " + ToStringInteger(Size()) );
  rTw.WriteComment(" Init/Marked:   " + ToStringInteger(mInitStates.Size()) 
	      + "/" + ToStringInteger(mMarkedStates.Size()));
  rTw.WriteComment(" Events:        " + ToStringInteger(mpAlphabet->Size()) );
  rTw.WriteComment(" Transitions:   " + ToStringInteger(mpTransRel->Size()) );
  rTw.WriteComment(" StateSymbols:  " + ToStringInteger(mpStateSymbolTable->Size()) );
  rTw.WriteComment(" Attrib. E/S/T: " + ToStringInteger(mpAlphabet->AttributesSize()) 
	      + "/" + ToStringInteger(mpStates->AttributesSize())
	      + "/" + ToStringInteger(mpTransRel->AttributesSize()) );
  rTw.WriteComment("");
}

// DotWrite(rFileName)
void vGenerator::DotWrite(const std::string& rFileName) const {
  FD_DG("vGenerator(" << this << ")::DotWrite(" << rFileName << ")");
  SetMinStateIndexMap();
  StateSet::Iterator lit;
  TransSet::Iterator tit;
  try {
    std::ofstream stream;
    stream.exceptions(std::ios::badbit|std::ios::failbit);
    stream.open(rFileName.c_str());
    stream << "digraph \"" << Name() << "\" {" << std::endl;
    stream << "  rankdir=LR" << std::endl;
    stream << "  node [shape=circle];" << std::endl;
    stream << std::endl;
    stream << "  // initial states" << std::endl;
    int i = 1;
    for (lit = InitStatesBegin(); lit != InitStatesEnd(); ++lit) {
      std::string xname= StateName(*lit);
      if(xname=="") xname=ToStringInteger(MinStateIndex(*lit));
      stream << "  dot_dummyinit_" << i << " [shape=none, label=\"\" ];" << std::endl;
      stream << "  dot_dummyinit_" << i << " -> \"" << xname << "\";" << std::endl; 
      i++;
    }
    stream << std::endl;
    stream << "  // mstates" << std::endl;
    for (lit = MarkedStatesBegin(); lit != MarkedStatesEnd(); ++lit) {
      std::string xname= StateName(*lit);
      if(xname=="") xname=ToStringInteger(MinStateIndex(*lit));
      stream << "  \"" << xname << "\" [shape=doublecircle];" << std::endl;
    }
    stream << std::endl;
    stream << "  // rest of stateset" << std::endl;
    for (lit = StatesBegin(); lit != StatesEnd(); ++lit) {
      if (! (ExistsInitState(*lit) || ExistsMarkedState(*lit)) ) {
	std::string xname= StateName(*lit);
	if(xname=="") xname=ToStringInteger(MinStateIndex(*lit));
	stream << "  \"" << xname << "\";" << std::endl;
      }
    }
    stream << std::endl;
    stream << "  // transition relation" << std::endl;
    for(tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
      std::string x1name= StateName(tit->X1);
      if(x1name=="") x1name=ToStringInteger(MinStateIndex(tit->X1));
      std::string x2name= StateName(tit->X2);
      if(x2name=="") x2name=ToStringInteger(MinStateIndex(tit->X2));
      stream << "  \"" << x1name  << "\" -> \"" << x2name
	     << "\" [label=\"" << EventName(tit->Ev) << "\"];" << std::endl;
    }
    stream << "};" << std::endl;
    stream.close();
  }
  catch (std::ios::failure&) {
    throw Exception("vGenerator::DotWrite", 
		    "Exception opening/writing dotfile \""+rFileName+"\"", 2);
  }
  ClearMinStateIndexMap();
}

// DDotWrite(rFileName)
void vGenerator::DDotWrite(const std::string& rFileName) const {
  FD_DG("vGenerator(" << this << ")::DDotWrite(" << rFileName << ")");
  StateSet::Iterator lit;
  TransSet::Iterator tit;
  try {
    std::ofstream stream;
    stream.exceptions(std::ios::badbit|std::ios::failbit);
    stream.open(rFileName.c_str());
    stream << "digraph \"" << Name() << "\" {" << std::endl;
    stream << "  rankdir=LR" << std::endl;
    stream << "  node [shape=circle];" << std::endl;
    stream << std::endl;
    stream << "  // istates" << std::endl;
    int i = 1;
    for (lit = InitStatesBegin(); lit != InitStatesEnd(); ++lit) {
      stream << "  dot_dummyinit_" << i << " [shape=none, label=\"\" ];" << std::endl;
      stream << "  dot_dummyinit_" << i << " -> \"" 
	     << SStr(*lit)  << "\";" << std::endl; 
      i++;
    }
    stream << std::endl;
    stream << "  // mstates" << std::endl;
    for (lit = MarkedStatesBegin(); lit != MarkedStatesEnd(); ++lit) {
      stream << "  \"" << SStr(*lit) << "\" [shape=doublecircle];" << std::endl;
    }
    stream << std::endl;
    stream << "  // rest of stateset" << std::endl;
    for (lit = StatesBegin(); lit != StatesEnd(); ++lit) {
      // if not in mInitStates or mMarkedStates
      if (! (ExistsInitState(*lit) || ExistsMarkedState(*lit)) ) {
	stream << "  \"" << SStr(*lit) << "\";" << std::endl;
      }
    }
    stream << std::endl;
    stream << "  // transition relation" << std::endl;
    for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
      stream << "  \"" << SStr(tit->X1) 
	     << "\" -> \"" << SStr(tit->X2)
	     << "\" [label=\"" << EventName(tit->Ev) << "\"];" << std::endl;
    }
    stream << "};" << std::endl;
    stream.close();
  }
  catch (std::ios::failure&) {
    throw Exception("vGenerator::DotWrite", 
		    "Exception opening/writing dotfile \""+rFileName+"\"", 2);
  }
}


// XDotWrite(rFileName)
void vGenerator::XDotWrite(const std::string& rFileName) const {
  FD_DG("vGenerator(" << this << ")::XDotWrite(" << rFileName << ")");
  StateSet::Iterator lit;
  TransSet::Iterator tit;
  try {
    std::ofstream stream;
    stream.exceptions(std::ios::badbit|std::ios::failbit);
    stream.open(rFileName.c_str());
    stream << "digraph \"___" << Name() << "___\" {" << std::endl;
    stream << "  rankdir=LR" << std::endl;
    stream << "  node [shape=circle];" << std::endl;
    stream << std::endl;
    stream << "  // stateset" << std::endl;
    for (lit = StatesBegin(); lit != StatesEnd(); ++lit) {
      stream << "  \"s" << *lit << "\";" << std::endl;
    }
    stream << std::endl;
    stream << "  // transition relation" << std::endl;
    for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
      stream << "  \"s" << tit->X1 
	     << "\" -> \"s" << tit->X2
	     << "\" [label=\"e" << tit->Ev << "\" " << "polyline" << "];" << std::endl;
    }
    stream << "};" << std::endl;
    stream.close();
  }
  catch (std::ios::failure&) {
    throw Exception("vGenerator::XDotWrite", 
		    "Exception opening/writing dotfile \""+rFileName+"\"", 2);
  }
}

// DoRead(tr)
void vGenerator::DoRead(TokenReader& rTr,  const std::string& rLabel, const Type* pContext) {
  (void) pContext;
  std::string label=rLabel;
  if(label=="") label="Generator"; 
  FD_DG("vGenerator(" << this << ")::DoRead(): file " << rTr.FileName() << "  section " << label);
  // Clear old stuff
  Clear();
  // find Genertor in any of our current file formats
  rTr.SeekBegin(label);
  // get the Generator "name" token
  ReadGeneratorName(rTr);
  // read mAlphabet
  ReadAlphabet(rTr);
  // read stateset
  ReadStates(rTr);
  // read transrel
  ReadTransRel(rTr);
  // read istates
  ReadStateSet(rTr, "InitStates", mInitStates);
  // read mstates
  ReadStateSet(rTr, "MarkedStates", mMarkedStates);
  // read attribute
  mpGlobalAttribute->Read(rTr);
  // skip unknown global attributes
  AttributeVoid::Skip(rTr);
  // read end 
  rTr.SeekEnd(label);
  FD_DG("vGenerator(" << this << ")::DoRead(): done");
}

// ReadGeneratorName(rFileName)
void vGenerator::ReadGeneratorName(const std::string& rFileName) {
  TokenReader tr(rFileName);
  tr.SeekBegin("Generator");  // mimique old behaviour .. use "ReadBegin" instead   
  ReadGeneratorName(tr);
}

// ReadGeneratorName(tr)
void vGenerator::ReadGeneratorName(TokenReader& rTr) {
  FD_DG("vGenerator(" << this << ")::ReadGeneratorName(\"" 
	<< rTr.FileName() << "\")");
   mMyName=rTr.ReadString();
}

// ReadAlphabet(rFileName)
void vGenerator::ReadAlphabet(const std::string& rFileName) {
  FD_DG("vGenerator(" << this << ")::ReadAlphabet(\"" << rFileName << "\")");
  mpAlphabet->Read(rFileName,"Alphabet");
}

// ReadAlphabet(rTr) 
void vGenerator::ReadAlphabet(TokenReader& rTr) {
  FD_DG("vGenerator(" << this << ")::ReadAlphabet(\"" 
	<< rTr.FileName() << "\")");
  mpAlphabet->Read(rTr,"Alphabet");
}

// ReadStates(rFileName) 
void vGenerator::ReadStates(const std::string& rFileName) {
  TokenReader tr(rFileName);
  ReadStates(tr);
}

// ReadStates(tr) 
void vGenerator::ReadStates(TokenReader& rTr) {
  FD_DG("vGenerator(" << this << ")::ReadStates(\"" << rTr.FileName() << "\")");
  // HELPERS:
  Token token;
  AttributeVoid* attrp = mpStates->Attribute().New();
  FD_DG("vGenerator(" << this << ")::ReadStates(..): attribute type " << typeid(*attrp).name());
  // ALGORITHM:
  mpStates->Clear();
  mpStates->Name("States");
  mStateSymbolTable.Clear();
  rTr.SeekBegin("States");  
  while(rTr.Peek(token)) {
    // break end of section
    if (token.Type() == Token::End) {
      break;
    }
    // read state by index
    if (token.Type() == Token::Integer) {
      rTr.Get(token);
      Idx index = token.IntegerValue();
      FD_DG("vGenerator(" << this << ")::ReadStates(\"" << rTr.FileName() << "\"): by index " << index);
      if(mpStates->Exists(index)) {
	std::stringstream errstr;
	errstr << "Token " << token.IntegerValue() << " appears twice in stateset"
	       << rTr.FileLine();
	throw Exception("vGenerator::ReadStates", errstr.str(), 80);
      }
      // read attribute
      attrp->Read(rTr,"",this);
      // skip unknown attributes
      AttributeVoid::Skip(rTr);
      // insert element with attribute
      InsState(index); 
      StateAttribute(index,*attrp);
      continue;
    } 
    // read state by name
    if (token.Type() == Token::String) {
      rTr.Get(token);
      FD_DG("vGenerator(" << this << ")::ReadStates(\"" << rTr.FileName() << "\"): by name " << token.StringValue());
      if(ExistsState(token.StringValue())) {
	std::stringstream errstr;
	errstr << "Token " << token.StringValue() << " appears twice in stateset"
	       << rTr.FileLine();
	throw Exception("vGenerator::ReadStates", errstr.str(), 80);
      }
      // read attribute
      attrp->Read(rTr,"",this);
      // skip unknown attributes
      AttributeVoid::Skip(rTr);
      // insert element with attribute
      Idx index=InsState(token.StringValue()); 
      StateAttribute(index,*attrp);
      continue;
    } 
    // read consecutive block of anonymous states
    if (token.Type() == Token::Begin && token.StringValue() == "Consecutive") {
      rTr.ReadBegin("Consecutive");
      Token token1,token2;
      rTr.Get(token1);
      rTr.Get(token2);
      FD_DG("vGenerator(" << this << ")::ReadStates(\"" << rTr.FileName() << "\"): consecutive range");
      if ((token1.Type() != Token::Integer) || (token2.Type() != Token::Integer)) {
	std::stringstream errstr;
	errstr << "Invalid range of consecutive states"  << rTr.FileLine();
	throw Exception("vGenerator::ReadStates", errstr.str(), 80);
      }
      for(Idx index = (Idx) token1.IntegerValue(); index <= (Idx) token2.IntegerValue(); ++index) {
	if(mpStates->Exists(index)) {
	  std::stringstream errstr;
	  errstr << "Index " << index << " appears twice in stateset"
		 << rTr.FileLine();
	  throw Exception("vGenerator::ReadStates", errstr.str(), 80);
	}
	InsState(index);
      }
      rTr.ReadEnd("Consecutive");
      continue;
    }
    // cannot process token
    std::stringstream errstr;
    errstr << "Invalid token" << rTr.FileLine();
    throw Exception("vGenerator::ReadStates", errstr.str(), 80);
  }
  rTr.SeekEnd("States");
  // dispose attribute
  delete attrp;
  FD_DG("vGenerator(" << this << ")::ReadStates(\"" << rTr.FileName() << "\"): done");
}


// ReadStateSet(tr, rLabel, rStateSet) 
void vGenerator::ReadStateSet(TokenReader& rTr, const std::string& rLabel, StateSet& rStateSet)  const {
  FD_DG("vGenerator(" << this << ")::ReadStateSet(\"" << rLabel<< "\")");
  // HELPERS:
  Token token;
  AttributeVoid* attrp = rStateSet.Attribute().New();
  FD_DG("vGenerator(" << this << ")::ReadStateSet(..): attribute type " << typeid(*attrp).name());
  // ALGORITHM:
  rStateSet.Clear();
  rTr.SeekBegin(rLabel);  
  rStateSet.Name(rLabel);
  while(rTr.Peek(token)) {
    // break end of section
    if (token.Type() == Token::End) {
      break;
    }
    // read state by index
    if (token.Type() == Token::Integer) {
      rTr.Get(token);
      Idx index = token.IntegerValue();
      if(!ExistsState(index)) {
	std::stringstream errstr;
	errstr << "Token " << token.IntegerValue() << " not in stateset"
	       << rTr.FileLine();
	throw Exception("vGenerator::ReadStateSet", errstr.str(), 80);
      }
      // read attribute
      attrp->Read(rTr,"",this);
      // skip unknown attributes
      AttributeVoid::Skip(rTr);
      // insert element with attribute
      rStateSet.Insert(index); 
      rStateSet.Attribute(index,*attrp);
      continue;
    } 
    // read state by name
    if (token.Type() == Token::String) {
      rTr.Get(token);
      Idx index =StateIndex(token.StringValue());
      if(index==0) {
	std::stringstream errstr;
	errstr << "Token " << token.StringValue() << " not in stateset"
	       << rTr.FileLine();
	throw Exception("vGenerator::ReadStateSet", errstr.str(), 80);
      }
      // read attribute
      attrp->Read(rTr,"",this);
      // skip unknown attributes
      AttributeVoid::Skip(rTr);
      // insert element with attribute
      rStateSet.Insert(index); 
      rStateSet.Attribute(index,*attrp);
      continue;
    } 
    // read consecutve block of anonymous states
    if (token.Type() == Token::Begin && token.StringValue() == "Consecutive") {
      rTr.ReadBegin("Consecutive");
      Token token1,token2;
      rTr.Get(token1);
      rTr.Get(token2);
      if ((token1.Type() != Token::Integer) || (token2.Type() != Token::Integer)) {
	std::stringstream errstr;
	errstr << "Invalid range of consecutive states"  << rTr.FileLine();
	throw Exception("vGenerator::ReadStateSet", errstr.str(), 80);
      }
      for(Idx index = (Idx) token1.IntegerValue(); index <= (Idx) token2.IntegerValue(); ++index) {
	if(!ExistsState(index)) {
	  std::stringstream errstr;
	  errstr << "Token " << token.IntegerValue() << " not in stateset"
		 << rTr.FileLine();
	  throw Exception("vGenerator::ReadStateSet", errstr.str(), 80);
	}
	rStateSet.Insert(index);
      }
      rTr.ReadEnd("Consecutive");
      continue;
    }
    // cannot process token
    std::stringstream errstr;
    errstr << "Invalid token" << rTr.FileLine();
    throw Exception("vGenerator::ReadStateSet", errstr.str(), 50);
  }
  rTr.SeekEnd(rLabel);
  // dispose attribute
  delete attrp;
}

// ReadTransRel(rFileName)
void vGenerator::ReadTransRel(const std::string& rFileName) {
  TokenReader tr(rFileName);
  ReadTransRel(tr);
}

// ReadTransRel(tr) 
void vGenerator::ReadTransRel(TokenReader& rTr) {
  FD_DG("vGenerator(" << this << ")::ReadTransRel(\"" << rTr.FileName() << "\")");
  AttributeVoid* attrp = mpTransRel->Attribute().New();
  FD_DG("vGenerator(" << this << ")::ReadTransRel(..): attribute type " << typeid(*attrp).name());
  try {
    Token token;
    Idx x1 = 0, ev = 0, x2 = 0;
    rTr.SeekBegin("TransRel");
    // Read tokens
    while (rTr.Peek(token)) {
      // 0: end of transition relation
      if ((token.Type() == Token::End)) break;

      // 1: the x1 token
      rTr >> token;
      if (token.Type() == Token::Integer) {
	x1=token.IntegerValue();
      } else if (token.Type() == Token::String) {
	token.StringValue();
	x1=StateIndex(token.StringValue());
      } else break;

      // 2: the event token
      rTr >> token;
      if (token.Type() == Token::Integer) {
	ev=token.IntegerValue();
      } else if (token.Type() == Token::String) {
	token.StringValue();
	ev=EventIndex(token.StringValue());
      } else break;

      // 3: the x1 token
      rTr >> token;
      if (token.Type() == Token::Integer) {
	x2=token.IntegerValue();
      } else if (token.Type() == Token::String) {
	token.StringValue();
	x2=StateIndex(token.StringValue());
      } else break;

      // 4: attributes
      attrp->Read(rTr,"",this);

      // 5: skip unknown attributes
      AttributeVoid::Skip(rTr);

      // check values
      if( (!ExistsState(x1)) || (!ExistsState(x2)) ){
	std::stringstream errstr;
	errstr << "invalid state" << rTr.FileLine();
	throw Exception("vGenerator::ReadTransRel", errstr.str(), 85);
      } 
      if(!ExistsEvent(ev)) {
	std::stringstream errstr;
	errstr << "invalid event" << rTr.FileLine();
	throw Exception("vGenerator::ReadTransRel", errstr.str(), 85);
      } 

      // insert transition
      Transition trans=Transition(x1,ev,x2);
      SetTransition(trans);
      TransAttribute(trans,*attrp);

    } // end while
    rTr.SeekEnd("TransRel");   
  }
  catch (faudes::Exception&) {
    std::stringstream errstr;
    errstr << "Reading TransRel failed in " << rTr.FileLine();
    throw Exception("vGenerator::ReadTransRel", errstr.str(), 50); 
  }
  FD_DG("vGenerator(" << this << ")::ReadTransRel(\"" << rTr.FileName() << "\"): sone");
  // dispose attribute
  delete attrp;
}


// EStr(index)
std::string vGenerator::EStr(Idx index) const {
  return EventName(index);
}

// SStr(index)
std::string vGenerator::SStr(Idx index) const {
  std::string name = StateName(index);
  if(name == "") name=ToStringInteger(index);
  return name+"["+faudes::ToStringInteger(index)+"]";
}

// TStr(index)
std::string vGenerator::TStr(const Transition& trans) const {
  return SStr(trans.X1) + "--(" + EStr(trans.Ev) + ")-->" + SStr(trans.X2);
}


// GraphWrite(rFileName,rOutFormat)
void vGenerator::GraphWrite(const std::string& rFileName, const std::string& rOutFormat,
   const std::string& rDotExec) const {
  FD_DG("vGenerator::GraphWrite(...)");
  // generate temp dot ibput file
  std::string dotin = CreateTempFile();
  if(dotin == "") {
    std::stringstream errstr;
    errstr << "Exception opening temp file";
    throw Exception("vGenerator::GraphWrite", errstr.str(), 2);
  }
  try{
    DotWrite(dotin);
  } 
  catch (faudes::Exception& exception) {  
    RemoveFile(dotin);
    std::stringstream errstr;
    errstr << "Exception writing dot input file";
    throw Exception("vGenerator::GraphWrite", errstr.str(), 2);
  }
  try{
    faudes::ProcessDot(dotin,rFileName,rOutFormat,rDotExec);
  } 
  catch (faudes::Exception& exception) {  
    RemoveFile(dotin);
    std::stringstream errstr;
    errstr << "Exception processing dot file";
    throw Exception("vGenerator::GraphWrite", errstr.str(), 3);
  }
  RemoveFile(dotin);
}

// rti wrapper
void IsAccessible(const vGenerator& rGen, bool& rRes) {
  rRes=rGen.IsAccessible();
}

// rti wrapper
void IsCoaccessible(const vGenerator& rGen, bool& rRes) {
  rRes=rGen.IsCoaccessible();
}

// rti wrapper
void IsTrim(const vGenerator& rGen, bool& rRes) {
  rRes=rGen.IsTrim();
}

// rti wrapper
void IsDeterministic(const vGenerator& rGen, bool& rRes) {
  rRes=rGen.IsDeterministic();
}


// rti wrapper
void Accessible(vGenerator& rGen) {
  rGen.Accessible();
}

// rti wrapper
void Accessible(const vGenerator& rGen, vGenerator& rRes) {
  rRes=rGen;
  rRes.Accessible();
}  

// rti wrapper
void Coaccessible(vGenerator& rGen) {
  rGen.Coaccessible();
}

// rti wrapper
void Coaccessible(const vGenerator& rGen, vGenerator& rRes) {
  rRes=rGen;
  rRes.Coaccessible();
}  

// rti wrapper
void Trim(vGenerator& rGen) {
  rGen.Trim();
}

// rti wrapper
void Trim(const vGenerator& rGen, vGenerator& rRes) {
  rRes=rGen;
  rRes.Trim();
}  

} // name space


